package de.cnc;

import java.lang.reflect.Method;
import java.util.HashMap;

import javax.servlet.http.*;

import de.cnc.util.*;

/**
 * Cache fr die Klassen mit den Action-Methoden . <br/>
 * Dieser Cache wird vom zentralen Prozessor des CnC-Systems {@link FlowControl}
 * zum Aufruf der Action-Methoden verwendet. <br/>
 * <br/>
 * Ihre applikations-spezifischen Action-Methoden teilen Sie dem CnC-System ber
 * ihr eigenes vom {@link CncServlet} abgeleitetes Servlet mit:<br/>
 * Implementieren Sie dazu die Methode {@link CncServlet#getActionPackagePrefix}<br/>
 *
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner Kcker</a><br/>
 * <br/>
 *
 * @author Heiner Kcker
 * @version $Id:$
 */
public final class ActionMethodClassCache
{

   /**
    * Merker fr Package-String der Action-Klassen . <br/><br/>
    */
   private static String actionPackage;

   /**
    * Merker fr Package-String der impliziten Action-Klassen . <br/><br/>
    */
   private static String implActionPackage = "de.cnc.implicit.";

   /**
    * Merker fr Package-String der impliziten Action-Klassen der Cnc-Anwendung . <br/><br/>
    */
   private static String appImplActionPackage;

   /**
    * Merker fr den Namen der Default-Action-Klasse . <br/><br/>
    * <pre>
    * Fr Action-Aufrufe ohne enthaltenes Nummernsysmbol # (Trenner Klasse-Methode) wird eine
    * Default-Action-Klasse verwendet.
    * Der Name der Default-Action-Klasse wird dem CnC-System durch die Methode
    *
    *     String getDefaultActionClassName() ;
    *
    * der Implementationsklasse des CncServlet bekannt gemacht.
    * </pre>
    */
   private static String defaultActionClassName;

   /**
    * Hashtable zum Cachen der ActionMethodClass-Objekte
    */
   private static HashMap actionMethodClassObjectCache = new HashMap();

   /**
    * bekannt machen des Package der Action-Methods-Klassen . <br/>
    * wird aufgerufen von {@link FlowControl#init}
    */
   public static void setActionPackage(String newActionPackage)
   {
      actionPackage = newActionPackage;
   } // end method setPackage

   /**
    * bekannt machen des Package der Action-Methods-Klassen f&uuml;r implizite Actions . <br/>
    * wird aufgerufen von {@link FlowControl#init}
    */
   public static void setAppImplActionPackage(String newAppImplActionPackage)
   {
      appImplActionPackage = newAppImplActionPackage;
   } // end method setPackage

   /**
    * bekannt machen der Default(Standard)-Klasse f&uuml;r die CnC-Anwendung . <br/>
    * Dieser Klassenname wird angewendet, wenn keine Klasse in der Action angegeben wird
    * (nur Methode statt Klasse#Methode).<br/>
    *
    * wird aufgerufen von {@link FlowControl#init}
    */
   public static void setDefaultActionClassName(String newDefaultActionClassName)
   {
      defaultActionClassName = newDefaultActionClassName;
   } // end method setDefaultActionClassName

   /**
    * Aufruf der Action-Methode fr die entsprechende Action-Klasse. <br/>
    * <pre>
    * Diese Methode trennt die Action-Klasse von der Action-Methode (Trennzeichen #)
    * oder setzt die Default-Action-Methods-Klasse ein.
    *
    * z.B.:  Bestellung#listeGet
    *           |         |
    *           |         +----  Action-Methode  ->  listeGet
    *           |
    *           +--------------- Action-Klasse   ->  Bestellung
    *
    * </pre>
    *
    * @param actionClassAndMethodName Name der Action mit Action-Klasse
    * @param iPaIndex Index f&uuml;r ein evtl. involviertes Listenelement
    * @param request
    * @param response
    * @return String fr Ergebnis-Auswertungen
    */
   public final static Object invokeActionMethod(
      String actionClassAndMethodName,
      int iPaIndex,
      HttpServletRequest request,
      HttpServletResponse response,
      CncContext cncContext)
   {
      String actionClassName = null;
      String actionMethodName = null;

      actionClassAndMethodName = StringUtil.cutLeft(actionClassAndMethodName, "#");
      // eventuelles beginnendes # entfernen

      if (StringUtil.contents(actionClassAndMethodName, '#'))
      {
         actionClassName = StringUtil.preNumSign(actionClassAndMethodName);
         actionMethodName = StringUtil.postNumSign(actionClassAndMethodName);
      }
      else
      { // kein #-Trenner fr Klasse und Methode
         actionClassName = defaultActionClassName; // default
         actionMethodName = actionClassAndMethodName;
      } // end if

      // Achtung Achtung: hier wird der Package vor den className gesetzt
      Object retObj =
         invokeActionMethod(actionPackage + actionClassName, actionMethodName, iPaIndex, request, response, cncContext);

      return retObj;
   } // end method invokeActionMethod

   /**
    * Aufruf der Action-Methode fr die entsprechende Action-Klasse. <br/>
    * <pre>
    * Diese Methode trennt die Action-Klasse von der Action-Methode (Trennzeichen #)
    * oder setzt die Default-Action-Methods-Klasse ein.
    *
    * z.B.:  Bestellung#listeGet
    *           |         |
    *           |         +----  Action-Methode  ->  listeGet
    *           |
    *           +--------------- Action-Klasse   ->  Bestellung
    *
    * </pre>
    *
    * @param actionClassAndMethodName Name der Action mit Action-Klasse
    * @param iPaIndex Index f&uuml;r ein evtl. involviertes Listenelement
    * @param request
    * @param response
    * @return String fr Ergebnis-Auswertungen
    */
   public final static Object invokeImplActionMethod(
      String actionClassAndMethodName,
      int iPaIndex,
      HttpServletRequest request,
      HttpServletResponse response,
      CncContext cncContext)
   {
      // throws ServletException, java.io.IOException {

      String actionClassName = null;
      String actionMethodName = null;

      actionClassName = StringUtil.preNumSign(actionClassAndMethodName);
      actionMethodName = StringUtil.postNumSign(actionClassAndMethodName);

      // Achtung Achtung: hier wird der implActionPackage vor den className gesetzt
      Object retObj =
         invokeActionMethod(
            implActionPackage + actionClassName,
            actionMethodName,
            iPaIndex,
            request,
            response,
            cncContext);

      return retObj;
   } // end method invokeImplActionMethod

   /**
    * Aufruf der Action-Methode fr die entsprechende Action-Klasse. <br/>
    * <pre>
    * Diese Methode trennt die Action-Klasse von der Action-Methode (Trennzeichen #)
    * oder setzt die Default-Action-Methods-Klasse ein.
    *
    * z.B.:  Bestellung#listeGet
    *           |         |
    *           |         +----  Action-Methode  ->  listeGet
    *           |
    *           +--------------- Action-Klasse   ->  Bestellung
    *
    * </pre>
    *
    * @param actionClassAndMethodName Name der Action mit Action-Klasse
    * @param iPaIndex Index f&uuml;r ein evtl. involviertes Listenelement
    * @param request
    * @param response
    * @return String fr Ergebnis-Auswertungen
    */
   public final static Object invokeAppImplActionMethod(
      String actionClassAndMethodName,
      int iPaIndex,
      HttpServletRequest request,
      HttpServletResponse response,
      CncContext cncContext)
   {
      // throws ServletException, java.io.IOException {

      String actionClassName = null;
      String actionMethodName = null;

      actionClassName = StringUtil.preNumSign(actionClassAndMethodName);
      actionMethodName = StringUtil.postNumSign(actionClassAndMethodName);

      // Achtung Achtung: hier wird der appImplActionPackage vor den className gesetzt
      Object retObj =
         invokeActionMethod(
            appImplActionPackage + actionClassName,
            actionMethodName,
            iPaIndex,
            request,
            response,
            cncContext);

      return retObj;
   } // end method invokeAppImplActionMethod

   /**
    * Aufruf der Action-Methode fr die entsprechende Action-Klasse. <br/>
    * Name der Action-Klasse und Action-Methode schon getrennt.<br/>
    * <pre>
    * z.B.:  package.Bestellung#listeGet
    *           |         |         |
    *           |         |         +----  Action-Methode  ->  listeGet
    *           |         |
    *           |         +--------------- Action-Klasse   ->  Bestellung
    *           |
    *           +------------------------- Package-Angabe
    *
    * </pre>
    *
    * @param actionClassName Name der Action-Klasse
    * @param actionMethodName Name der Action-Methode
    * @param iPaIndex Index f&uuml;r ein evtl. involviertes Listenelement
    * @param request
    * @param response
    * @return String fr Ergebnis-Auswertungen
    */
   private final static Object invokeActionMethod(
      String actionClassName,
      String actionMethodName,
      int iPaIndex,
      HttpServletRequest request,
      HttpServletResponse response,
      CncContext cncContext)
   {
      // throws ServletException, java.io.IOException {

      Object actionMethodsObject = getActionMethodsObject(actionClassName);
      // Log.debug( "invokeActionMethod " + actionMethodsObject.getClass().getName() + "#" + actionMethodName ) ;
      Object retObj =
         invokeActionMethod(actionMethodsObject, actionMethodName, iPaIndex, request, response, cncContext);

      // Log.debug( "retStr " + retStr ) ;
      return retObj;
   } // end method invokeActionMethod

   /**
    * Holen des Objektes mit den Action-Methoden aus dem Cache oder
    * Initialisierung, wenn notwendig.
    *
    * @param actionBereich Name des Action-Bereichs kurz
    * @return das xxxActionMethods-Objekt
    */
   private static final Object getActionMethodsObject(String actionBereich)
   {
      // throws ServletException {
      Object retObject = actionMethodClassObjectCache.get(actionBereich);

      if (retObject == null)
      {
         synchronized (actionMethodClassObjectCache)
         {
            if (retObject == null)
            {
               retObject = loadClass(actionBereich);

               actionMethodClassObjectCache.put(actionBereich, retObject);
            } // end if
         } // end synchronized
      } // end if

      return retObject;
   } // end method getActionMethodsObject

   /**
    * Laden der Klasse fr die einzelnen ActionBereiche. <br/><br/>
    * wird benutzt in {@link #getActionMethodsObject} . <br/><br/>
    * Achtung Achtung: hier wird der Package vor und den actionBereich gesetzt . <br/>
    *
    * @param actionBereich Name des Action-Bereichs kurz
    * @return die xxx-Action-Methods-Klasse
    */
   private synchronized static final Object loadClass(String className)
   {
      // throws ServletException {
      Object returnObject = null;

      //Log.debug("ActionMethodClassCache loadClass: " + className);
      try
      {
         Class myClass = Class.forName(className);

         returnObject = myClass.newInstance();
      }
      catch (Exception e)
      {
          Log.fatal(e);
         // throw new ServletException("cant load Class " + className);
         throw new RuntimeException("cant load Class " + className);
      }
      return returnObject;
   } // end method loadClass

   /**
    * Aufrufen der Action-Methode anhand des Namens der Action. <br/>
    * <pre>
    * z.B.: listeGet -> listeGetAction
    * </pre>
    *
    * @param actionMethodsObject das von {@link #getActionMethodsObject} bereitgestellte Objekt
    * @param methodName Action-Name kurz
    * @param request
    * @param response
    * @return String fr Ergebnis-Auswertungen
    */
   private static final Object invokeActionMethod(
      Object actionMethodsObject,
      String methodName,
      int iPaIndex,
      HttpServletRequest request,
      HttpServletResponse response,
      CncContext cncContext)
   {
      // throws ServletException, java.io.IOException {
      Object retObj = null;

      try
      {

         //Achtung: Funktioniert nicht korrekt, wenn es mehrere Methoden mit dem gleichen Namen gibt
         Method method = MethodCache.getMethod(actionMethodsObject, methodName);

Log.debug( "method" , method );

         Object[] paramArr = { new Integer(iPaIndex), request, response, cncContext };

         retObj = method.invoke(/*actionMethodsObject, bei Aufruf von static-Methoden ist keine Instanz ntig*/ null, paramArr);
      }
      catch (Exception e)
      {
         Log.fatal(e);
         // throw new ServletException("cant invoke Method " + actionMethodsObject.getClass().getName() + "#" + methodName + " cause " + e.getMessage());
         throw new RuntimeException(
            "cant invoke Method "
               + actionMethodsObject.getClass().getName()
               + "#"
               + methodName
               + " cause "
               + e.getMessage());
      }
      return retObj;
   } // end method invokeActionMethod

} // end class ActionMethodClassCache
