package de.cnc;

import java.util.*;

import javax.servlet.ServletException;
import javax.servlet.http.*;

import de.cnc.util.*;
import de.cnc.util.config.ConfigManager;
import de.cnc.validator.FormFieldErrorMap;

/**
 * zentrales und einziges Servlet zum Verarbeiten aller Actions in der CnC-Applikation . <br/>
 * Dieses Servlet ist in der web.xml auf das URL-Pattern "*.cnc" gemappt . <br/>
 * <pre>
 *      +--&gt; Browser: Link oder Form-Action *.cnc
 *      |      |
 *      |      +--&gt; Servlet-Engine mappt *.cnc auf CncServlet
 *      |             |
 *      |             +--&gt; CncServlet startet zentralen Prozessor {@link FlowControl}
 *      |                     |
 *      |                     +--&gt; {@link FlowControl} arbeitet die Befehle aus <b>flow.cnc</b> ab
 *      |                            |
 *      |                            +--&gt; {@link FlowControl} gibt String mit der nchsten forward-JSP zur&uuml;ck
 *      |                                   |
 *      +-----------------------------------+
 *
 * </pre>
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner K&uuml;cker</a><br/>
 * <br/>
 *
 * @author Heiner K&uuml;cker
 * @version $Id:$
 */
abstract public class CncServlet extends HttpServlet
{

    /**
     * Diese Methode gibt den Namen der im CnC-Framework laufenden Applikation zurck .<br/>
     *
     * @return String mit Applikations-Name
     */
    abstract public String getApplicationName();

    /**
     * Diese Methode gibt das Package der Action-Methoden-Klassen zur&uuml;ck . <br/>
     * Sie muss vom zentralen Servlet der Cnc-Applikation implementiert werden.<br/>
     * Siehe Konzept_Actions.txt .<br/>
     *
     * @return String mit Package-Angabe
     */
    abstract public String getActionPackage();

    /**
     * Diese Methode gibt den Namen der Default-Action-Methoden-Klasse zur&uuml;ck . <br/>
     * Sie muss vom zentralen Servlet der Cnc-Applikation implementiert werden.<br/>
     * Siehe Konzept_Actions.txt .<br/>
     *
     * @return String mit Package-Angabe
     */
    abstract public String getDefaultActionClassName();

    /**
     * Diese Methode gibt den Package-Prefix der Action-Methoden-Klassen f&uuml;r implizite Actions zur&uuml;ck . <br/>
     * Sie muss vom zentralen Servlet der Cnc-Applikation implementiert werden.<br/>
     * Siehe Konzept_Actions.txt .<br/>
     *
     * @return String mit Package-Angabe
     */
    abstract public String getAppImplActionPackage();

    /**
     * Diese Methode gibt den Package-Prefix der Action-Methoden-Klassen f&uuml;r implizite Actions zur&uuml;ck . <br/>
     * Sie muss vom zentralen Servlet der Cnc-Applikation implementiert werden.<br/>
     * Siehe Konzept_Actions.txt .<br/>
     *
     * @return String mit Prefix-Angabe
     */
    abstract public String getAppImplActionPrefix();

    /**
     * Pr&uuml;fen der Berechtigung des aktuellen Users zum Ausf&uuml;hren der Cnc-Prozedur . <br/>
     *
     * @param request hier&uuml;ber hat die Methode Zugriff auf die Session, aus der sich die Methode das User-Objekt holen kann
     * @param procName Name der flow.cnc-procedure, f&uuml;r die das Zugriffsrecht gepr&uuml;ft werden soll
     *
     * @return ob berechtigt
     */
    abstract public boolean hasUserPermissionForProcedure(
        HttpServletRequest request,
        String procName);

    /**
     * Pr&uuml;fen der Berechtigung des aktuellen Users zum Ausf&uuml;hren der Action-Methode . <br/>
     *
     * @param request hier&uuml;ber hat die Methode Zugriff auf die Session, aus der sich die Methode das User-Objekt holen kann
     * @param procName Name der Action-Methode, f&uuml;r die das Zugriffsrecht gepr&uuml;ft werden soll
     *
     * @return ob berechtigt
     */
    abstract public boolean hasUserPermissionForAction(
        HttpServletRequest request,
        String actionName);

    /**
     * Pr&uuml;fen der Berechtigung des aktuellen Users zum Ausf&uuml;hren der impliziten Action . <br/>
     *
     * @param request hier&uuml;ber hat die Methode Zugriff auf die Session, aus der sich die Methode das User-Objekt holen kann
     * @param procName Name der Action-Methode, f&uuml;r die das Zugriffsrecht gepr&uuml;ft werden soll
     *
     * @return ob berechtigt
     */
    abstract public boolean hasUserPermissionForImplAction(
        HttpServletRequest request,
        String actionName);

    /**
     * Pr&uuml;fen der Berechtigung des aktuellen Users zum Ausf&uuml;hren der JSP . <br/>
     *
     * @param request hier&uuml;ber hat die Methode Zugriff auf die Session, aus der sich die Methode das User-Objekt holen kann
     * @param procName Name der JSP, f&uuml;r die das Zugriffsrecht gepr&uuml;ft werden soll
     *
     * @return ob berechtigt
     */
    abstract public boolean hasUserPermissionForJsp(HttpServletRequest request, String jspName);

    /**
     * die Implementationsklasse von CncServlet liefert einen alternativen
     * forward, falls keine Berechtigung f&uuml;r die aktuelle Prozedur, JSP oder Action-Methode vorliegt . <br/>
     * Dies sollte die login.jsp oder mainmenu.jsp sein.<br/>
     *
     * @param request hier&uuml;ber hat die Methode Zugriff auf die Session, aus der sich die Methode das User-Objekt holen kann
     * @param strPaSection String mit Namen der Prozedur, JSP oder Action-Methode, f&uuml;r welche die Berechtigung verweigert wurde.<br/>
     *                     Durch Vergleich mit String.startsWith() kann festgestellt werden, um welchen Applikationsteil es sich handelt.
     *
     * @return String mit alternativem forward
     */
    abstract public String getAlternateForward(HttpServletRequest request, String strPaSection);

    /**
     * Prefix aller URIs, die {@link #CncServlet} aufrufen. <br/>
     * <b> /cncdemo/ </b> <br/>
     * Ist durch die Festlegung des Applikations-Namens bestimmt. <br/>
     * Muss zum Auswerten der Action von der Request-URI abgeschnitten werden. <br/>
     */
    private static String webappPrefix;

    /**
     * Postfix aller Actions. <br/>
     * <b> .cnc </b> <br/>
     * Wird in web.xml gemappt, so dass alle URIs mit der Endung .cnc zum Aufruf von {@link #CncServlet} f&uuml;hren. <br/>
     * Muss zum Auswerten der Action von der Request-URI abgeschnitten werden. <br/>
     */
    // Default ist ".cnc"
    public static final String actionPostfix = RequestKeys.CNC_URL_POSTFIX;

    /**
     * String mit Package fuer implizite Actions der CnC-Applikation
     */
    // private static String strImplicitActionsPackage ;

    /**
     * String mit Prefix fuer implizite Actions der CnC-Applikation
     */
    // private static String strImplicitActionsPrefix ;

    /**
     * ob die Fluss-Steuer-Datei bei Aktualisierung (nderung Datum Zeit) neu gelesen werden soll,
     * alternativ nur bei Server-Start.
     */
    private boolean isFlowHotReload = false;

    /** 
     * Schalter fuer die Synchronisation der doRequest-Methode des CnCServlet,
     * sollte bei der Arbeit mit Frames auf 'true' stehen, sonst 'false' aus 
     * Performancegruenden.
     */
    private boolean isSynchronized = false;

    /**
     * Initializes the Servlet. <br/>
     */
    public void init() throws ServletException
    {
        try
        {
            // double checked synchronize block
            synchronized (this.getClass())
            {
                super.init();

                // Laden der Konfiguration aus den Properties-Dateien aus dem Config-Pfad entsprechend web.xml (param-name:configroot) und
                // Initialisierung des {@link de.cnc.util.config.ConfigManager}. <br/>
                synchronized (this.getClass())
                {
                    super.init();

                    ConfigManager.init(this);

                    Log.init();

                    // Log.debug( "os.name: " + System.getProperty( "os.name" ) );

                    // PersistensHelper.init();

                    // DatabaseCache.init();

                    // Uebergabe this-Parameter damit FlowControl.init auf die Initialization-Parameter zugreifen kann
                    FlowControl.init(this);
                    // Bekannt machen einer Referenz auf dieses zentrale Servlet zur allgemeinen Verwendung
                    CncServletAnchor.setCncServlet(this);

                    // Abfragen ob die Fluss-Steuer-Datei bei Aktualisierung (nderung Datum Zeit) neu gelesen werden soll
                    this.isFlowHotReload = StringUtil.str2bool(this.getInitParameter("flowhotreload"));

                    this.isSynchronized = StringUtil.str2bool(this.getInitParameter("synchronized"));

                    // Speichern Package und Prefix fuer implizite Actions der CnC-Applikation
                    // strImplicitActionsPackage = getServletConfig().getInitParameter( "implicit-actions-package" ) ;
                    // strImplicitActionsPrefix  = getServletConfig().getInitParameter( "implicit-actions-prefix" ) ;

                    // Ausgabe der Steuertabelle (Low-Level-Anweisungen) auf Datei zum Debugging
                    // StringUtil.str2file( "" + FlowControl.getFlowControlTable() , "C:\\Programme\\eclipse-2.1.3\\eclipse\\workspace\\cnc\\debug\\flow.out" ) ;
                    // StringUtil.str2file( FlowControl.getFlowControlTable().toString() , "C:\\debug\\flow.out" );

                    Log.debug(
                        "**********************************************************************");
                    Log.debug("  CnC application " + getApplicationName() + " initialized");
                    Log.debug(
                        "**********************************************************************");
                    // Log.debug("getAppImplActionPrefix() " + getAppImplActionPrefix());
                }
            }
        }
        catch (Exception e)
        {
            Log.fatal(e);
            throw new ServletException(e.getMessage());
        }
    } // end method init

    /**
     * Free any resources the servlet has acquired that will not be garbage collected.
     */
    public void destroy()
    {
        super.destroy();

        // insert user code from here
    }

    /**
     * Forwards Request to doRequest to avoid double implementation in doGet/doPost. <br/>
     *
     * @param req Servlet Request
     * @param resp Servlet Response
     */
    public void doGet(HttpServletRequest req, HttpServletResponse resp)
        throws ServletException, java.io.IOException
    {
        doRequest(req, resp);
    } // end doGet

    /**
     * Forwards Request to doRequest to avoid double implementation in doGet/doPost. <br>
     *
     * @param req Servlet Request
     * @param resp Servlet Response
     */
    public void doPost(HttpServletRequest req, HttpServletResponse resp)
        throws ServletException, java.io.IOException
    {
        doRequest(req, resp);
    } // end doPost

    /**
     * Diese Methode fasst die HTTP-Methoden Get und Post zusammen. <br/>
     * Wird von {@link #doGet doGet} und {@link #doPost doPost} aufgerufen. <br/>
     *
     * @param request
     * @param response
     *
     * @throws ServletException
     * @throws java.io.IOException
     */
    protected void doRequest(final HttpServletRequest request, final HttpServletResponse response)
        throws ServletException, java.io.IOException
    {
        if (this.isSynchronized)
        {
            //synchronisiertes Abarbeiten wegen Frames
            synchronized( request.getSession() )
            {
                innerDoRequest(request,response);
            }
        }
        else
        {
            //nicht synchronisiertes Arbeiten fr bessere Performance
            innerDoRequest(request,response);
        }
    }

    /**
     * Inneres doRequest fr synchronized-Aufruf.
     * 
     * @param request
     * @param response
     *
     * @throws ServletException
     * @throws java.io.IOException
     */
    private void innerDoRequest(final HttpServletRequest request, final HttpServletResponse response)
        throws ServletException, java.io.IOException
    {
        try
        {
            // Log.debug( request.getContextPath() ) ;
            // Setzen Name der Web-Applikation
            if (webappPrefix == null)
            {
                webappPrefix = request.getContextPath() + "/";
            }

            HttpSession session = request.getSession();

            if (isFlowHotReload)
            {
                // Neu Einlesen der Fluss-Steuer-Datei, falls diese sich gendert hat
                FlowControl.reload(this);
            }

            final HttpServletRequest multiPartRequest = request;
            // MultipartHttpServletRequest.handleMultipartRequest(request);

            if (Log.currentLogMode <= Log.DEBUG)
            {
                Log.debug(
                    "+------------------------------------------------------------------------------+");
                Log.debug(" " + getClass().getName() + "#doRequest");
                Log.debug(
                    "+------------------------------------------------------------------------------+");
                Log.dump(multiPartRequest);
                Log.debug(""); // Leerzeile
                Log.dump(session);
                Log.debug("action", CncUtil.getAction(multiPartRequest));
            }

            // pro Request einmal Aufruf zum Reorganisieren des Datenbank-Caches
            // DatabaseCache.reorg();

            if (session.getAttribute(SessionKeys.CNC_SINGLE_STEP_DEBUG) != null)
            { // Einzelschritt-Debug
                // Vermerken der Original-Action, damit diese beim weiteren Aufruf von singlestepdebug.jsp erhalten bleibt
                // wird durch das Tag debugrqfields weitergeschleust
                CncUtil.setAttribute(
                    session,
                    SessionKeys.CNC_SINGLE_STEP_DEBUG_ACTION,
                    CncUtil.getTrueAction(multiPartRequest));
            } // end if Einzelschritt-Debug

            // Ermitteln der vorherigen JSP
            // hidden-Field srcjsp= ... wird von FormTag.java und HrefTag mitgegeben
            CncUtil.setSrcJsp(multiPartRequest);

            //request.getParameter( RequestKeys.CNC_JAVASCRIPT );

            //Auswerten einer GOTO-Action (Navigations-Link)
            String actionStr = CncUtil.getAction(request);

            //Abfrage, ob ein Button mit der Option validate=false" bettigt wurde
            boolean isNoValidate = CncUtil.isNoValidate(request);

            // implizites Ausschalten Single-Step-Debug
            if (multiPartRequest.getParameter("cnc_debug_off") != null)
            { // if Button "Single-Step-Debug off" singlestepdebug.jsp
                session.removeAttribute(SessionKeys.CNC_SINGLE_STEP_DEBUG);
                // Single-Step-Debug ausschalten
            } // if Button "Single-Step-Debug off" singlestepdebug.jsp

            // vermeiden einparsen der Parameter, wenn die vorherige jsp singlestepdebug.jsp war, weil sonst fehler
            if (multiPartRequest.getParameter("cnc_debug_continue")
                == null // if ! vorher war singlestepdebug.jsp
            // vermeiden einparsen wenn ein Button cnc_no_parse (Abbruch) bettigt wurde
                && multiPartRequest.getParameter(RequestKeys.CNC_BUTTON_NO_PARSE) == null
                //bei GOTO nicht einparsen und nicht validieren
                && !actionStr.startsWith(RequestKeys.CNC_GOTO_ACTION_PREFIX))
            { // if ! button cnc_no_parse

                // Log.dump( session ) ;
                Object form = session.getAttribute(SessionKeys.CNC_FORM_KEY);
                // fester Key f&uuml;r alle Forms

                if (form != null)
                {
                    //    if ( form instanceof FormValidatorInterface )
                    //    {
                    //        boolean isValid = ( (FormValidatorInterface) form ).isRequestValid(request);
                    //        if ( ! isValid )
                    //        {
                    //            String forward = CncUtil.getSrcJsp(session);
                    //            Log.debug("forward to", forward);
                    //            request.getRequestDispatcher(response.encodeURL(forward)).forward(
                    //                request,
                    //                response);
                    //            return;
                    //        }
                    //    }

                    // diese folgende Mimik ist notwendig,
                    // weil der Browser bei nicht gecheckten Checkboxen( Checkbox-Value == false )
                    // einfach die Checkboxen nicht in den Request mitsendet
                    // Dadurch werden einmal angeklickte Checkboxen nicht mehr ausgeschaltet
                    // Im InputTag.java werden den checkboxen Hidden-Fields mit dem Prefix RequestKeys.CNC_CHECKBOX_DEFAULT_PREFIX
                    // mitgegeben.
                    // Die folgende Schleife sucht den Request nach Parametern, die mit RequestKeys.CNC_CHECKBOX_DEFAULT_PREFIX
                    // beginnen, ab und setzt alle entsprechenden Form-Properties auf false.
                    // Gecheckte Checkboxen liefern name=true und setzen die Form-Properties wieder.
                    Enumeration en = multiPartRequest.getParameterNames();

                    while (en.hasMoreElements())
                    {
                        String s = (String) en.nextElement();

                        // Object  o = multiPartRequest.getParameter(s);
                        if (s.startsWith(RequestKeys.CNC_CHECKBOX_DEFAULT_PREFIX))
                        {
                            s = s.substring(RequestKeys.CNC_CHECKBOX_DEFAULT_PREFIX.length());
                            // Prefix abschneiden

                            // Log.debug(RequestKeys.CNC_CHECKBOX_DEFAULT_PREFIX + s);
                            BeanUtil.setIndexed(form, s, "false");
                            // f&uuml;r den String-Setter
                            BeanUtil.setIndexed(form, s, Boolean.FALSE);
                            // f&uuml;r den Boolean-Setter
                        }
                    }

                    // wenn Form existiert einparsen des Requests
                    // @todo: unterschiedliche methoden zum parsen form und list (hat [ xx ] ) zum performancetuning
                    CncUtil.parseAndSet(multiPartRequest, form);
                }

                // Einparsen der Formulardaten in eine evtl. vorhandene Form-Liste (indiziertes Formular, Tabelle)
                Object formlistObject = session.getAttribute(SessionKeys.CNC_LIST_KEY);

                if (formlistObject != null)
                {
                    if (formlistObject instanceof ArrayList)
                    {
                        // es werden indizierte Daten (key[i]) einer Tabelle (DataGrid) erwartet
                        // @todo: unterschiedliche methoden zum parsen form und list (hat [ xx ] ) zum performancetuning
                        CncUtil.parseAndSet(multiPartRequest, formlistObject);
                    }
                    else
                    {
                        throw new ServletException(
                            "type "
                                + ArrayList.class.getName()
                                + " expected on session key "
                                + SessionKeys.CNC_LIST_KEY);
                    }
                }

                // Einlesen implziter Session-Values,
                // das heisst Request-Keys mit dem Prefix RequestKeys.CNC_SESSION_VALUE_PREFIX.
                // Der Prefix wird abgeschnitten und der Value unter dem Rest-Key in der Session gespeichert
                // Das cnc:input-Tag setzt den Prefix bei gesetztem Attribut sessionvalue="true" vor den Request-Key
                // Noch Lsung fr Checkboxen machen !!!
                CncUtil.parseAndSetImplSessionValues(multiPartRequest);

                if (!isNoValidate)
                {
                    FormFieldErrorMap formErrMap = new FormFieldErrorMap();
                    boolean isValid = validate(request, form, formlistObject, formErrMap);
                    if (!isValid)
                    {
                        String forward = CncUtil.getSrcJsp(session);
                        Log.debug("forward to", forward);

                        //bergeben der Fehlermeldungen an die JSP zum Rendern
                        session.setAttribute(SessionKeys.CNC_FORM_FIELD_ERROR_MAP, formErrMap);

                        request.getRequestDispatcher(response.encodeURL(forward)).forward(
                            request,
                            response);

                        //messages consumed
                        session.removeAttribute(SessionKeys.CNC_FORM_FIELD_ERROR_MAP);
                        return;
                    }
                } //end if ! isNoValidate
            } // end if ! vorher war singlestepdebug.jsp

            String forward;
            // = null; // "/logon.jsp"; // JspNames.LOGON_JSP ; // default-forward

            // Log.debug("getRequestURI: " + multiPartRequest.getRequestURI());
            // String  forward = "/logon.jsp" ; // JspNames.LOGON_JSP ; // default-forward
            // String  requestAction = "";

            // Abschneiden des Prefixes /app-name/ von der requestUri weil sonst weitere Verarbeitung nicht klappt
            // requestAction = StringUtil.cutLeft(multiPartRequest.getRequestURI(), webappPrefix);  // webappPrefix abschneiden

            // actionPostfix "*.cnc" abschneiden
            // String action = StringUtil.cutRight(requestAction, actionPostfix);

            // Log.debug("getAppImplActionPrefix() " + getAppImplActionPrefix());

            // implizite Action des CnC-Kern-Systems
            String strImplAction =
                CncUtil.scanForNameStartsWith(multiPartRequest, RequestKeys.CNC_IMPL_ACTION_PREFIX);

            // allgemeinen Kontext erzeugen
            CncContext cncContext =
                new CncContext(
                    multiPartRequest,
                    response,
                    session,
                    this.getServletContext(),
                    actionStr,
                    isNoValidate);

            if (strImplAction != null)
            {
                Log.debug("implizite Action", strImplAction);
                strImplAction =
                    StringUtil.cutLeft(strImplAction, RequestKeys.CNC_IMPL_ACTION_PREFIX);
                Log.debug("implizite Action", strImplAction);
                // doImplAction( strImplAction , multiPartRequest , response ) ;
                ActionMethodClassCache.invokeImplActionMethod(// -
                strImplAction, // -
                CncUtil.getCnCButtonIndex(multiPartRequest), // -
                multiPartRequest, // -
                response, // -
                cncContext);

                forward = CncUtil.getSrcJsp(session);
            } // -
            // implizite Action der CnC-Anwendung (cncdemo)
            // else if ( strImplicitActionsPrefix != null && ( strImplAction = CncUtil.scanForNameStartsWith( multiPartRequest , strImplicitActionsPrefix ) ) != null ) {
            else if (
                getAppImplActionPrefix() != null
                    && (strImplAction =
                        CncUtil.scanForNameStartsWith(multiPartRequest, getAppImplActionPrefix()))
                        != null)
            {
                Log.debug("implizite Anwendungs-Action", strImplAction);
                strImplAction = StringUtil.cutLeft(strImplAction, getAppImplActionPrefix());
                Log.debug("implizite Anwendungs-Action", strImplAction);
                // doAppImplAction( strImplAction , multiPartRequest , response ) ;
                ActionMethodClassCache.invokeAppImplActionMethod(// -
                strImplAction, // -
                CncUtil.getCnCButtonIndex(multiPartRequest), // -
                multiPartRequest, // -
                response, // -
                cncContext);

                forward = CncUtil.getSrcJsp(session);
            } // end implizite App-Action
            else
            { // Start CnC-Prozessor
                    forward = FlowControl.execute(// -
        // multiPartRequest, //-
        // response, //-
        // getServletContext(), //-
    cncContext);
            } // end if
            Log.debug("forward to", forward);
            request.getRequestDispatcher(response.encodeURL(forward)).forward(request, response);
            return;
        }
        catch (Exception e)
        {
            Log.fatal(e);
            // throw new ServletException(Log.getMessageAndStackTrace(e));
            // weiterleitung auf fehlerseite, vermerken Exception in Session fr Anzeige auf der Fehlerseite
            final HttpSession session = request.getSession();
            session.setAttribute("CNC_EXCEPTION", e);
            // @todo: Konstante machen
            request.getRequestDispatcher(response.encodeURL("/errorpge.jsp")).forward(
                request,
                response);
            // @todo: jsp-name aus Konstante oder Config-File
        }
    }

    /**
     * Ausf&uuml;hren implizite Action des CnC-Kern-Systems . <br/>
     *
     * @param request  Servlet Request
     * @param response Servlet Response
     */

    /*
     private String doImplAction ( String strPaActionKey , HttpServletRequest request , HttpServletResponse response ) {
     if ( ! hasUserPermissionForImplAction( request , strPaActionKey ) ) {
     CncUtil.setMessage( request , "no permission for action " + strPaActionKey ) ;
     return "" ;
     }
     String strActionClass = StringUtil.cutLeft( strPaActionKey , RequestKeys.CNC_IMPL_ACTION_PREFIX ) ;
     Log.debug( strActionClass ) ;
     String strActionClassWithPackage = "de.cnc.implicit." + strActionClass ;
     Log.debug( strActionClass ) ;
     Object objImplAction = null ;
     try {
     Class myClass = Class.forName( strActionClassWithPackage );
     objImplAction = myClass.newInstance();
     }
     catch ( Exception e ) {
     //zweiter Versuch mit Actions der abgeleiteten Applikation
     //if ( strImplicitActionsPackage != null ) {
     //  strActionClassWithPackage = strImplicitActionsPackage + "." + strActionClass ;
     //  try {
     //    Class myClass = Class.forName( strActionClassWithPackage );
     //    objImplAction = myClass.newInstance();
     //  }
     //  catch ( Exception e1 ) {
     //    throw new RuntimeException("cant load Class " + strActionClassWithPackage);
     //  }
     //}
     //else {
     throw new RuntimeException("cant load Class " + strActionClassWithPackage);
     //}
     }
     ImplicitActionInterface implicitActionInterface = (ImplicitActionInterface) objImplAction ;
     return implicitActionInterface.doAction( CncUtil.getCnCButtonIndex( request ) , request , response ) ;
     } // end method doImplAction
     */

    /**
     * Ausf&uuml;hren implizite Action der CnC-Anwendung (z.B. cncdemo) . <br/>
     *
     * @param request  Servlet Request
     * @param response Servlet Response
     */

    /*
     private String doAppImplAction ( String strPaActionKey , HttpServletRequest request , HttpServletResponse response ) {
     if ( ! hasUserPermissionForImplAction( request , strPaActionKey ) ) {
     CncUtil.setMessage( request , "no permission for action " + strPaActionKey ) ;
     return "" ;
     }
     String strActionClass = StringUtil.cutLeft( strPaActionKey , strImplicitActionsPrefix ) ;
     Log.debug( strActionClass ) ;
     String strActionClassWithPackage = strImplicitActionsPackage + "." + strActionClass ;
     Log.debug( strActionClass ) ;
     Object objImplAction = null ;
     try {
     Class myClass = Class.forName( strActionClassWithPackage );
     objImplAction = myClass.newInstance();
     }
     catch ( Exception e ) {
     throw new RuntimeException("cant load Class " + strActionClassWithPackage);
     }
     ImplicitActionInterface implicitActionInterface = (ImplicitActionInterface) objImplAction ;
     return implicitActionInterface.doAction( CncUtil.getCnCButtonIndex( request ) , request , response ) ;
     } // end method doAppImplAction
     */

    /**
     * Ausfhren Validierung
     * 
     * @param request
     * @param form
     * @param formlistObject
     * @param formErrMap
     * @return
     */
    private boolean validate(
        final HttpServletRequest request,
        Object form,
        Object formlistObject,
        FormFieldErrorMap formErrMap)
    {
        boolean isValid = true;
        if (form instanceof FormValidatorInterface)
        {
            isValid = ((FormValidatorInterface) form).isBeanValid(request, formErrMap);
//System.out.println( "form isValid" + isValid );
            //                        if ( ! isValid )
            //                        {
            //                            String forward = CncUtil.getSrcJsp(session);
            //                            Log.debug("forward to", forward);
            //                        
            //                            //bergeben der Fehlermeldungen an die JSP zum Rendern
            //                            session.setAttribute(SessionKeys.CNC_FORM_FIELD_ERROR_MAP, formErrMap);
            //                        
            //                            request.getRequestDispatcher(response.encodeURL(forward)).forward(
            //                                request,
            //                                response);
            //
            //                            //messages consumed
            //                            session.removeAttribute(SessionKeys.CNC_FORM_FIELD_ERROR_MAP);
            //
            //                            return;
            //                        }
        }

        if (formlistObject != null)
        {
            //Validieren
            ArrayList formArrList = (ArrayList) formlistObject;

            for (int i = 0; i < formArrList.size(); i++)
            {
                Object runObj = formArrList.get(i);
                //System.out.println("runObj: " + runObj);
                if (runObj instanceof FormValidatorInterface)
                {
                    FormFieldErrorMap oneListLineFormErrMap = new FormFieldErrorMap();
                    isValid =
                        isValid
                            && ((FormValidatorInterface) runObj).isBeanValid(
                                request,
                                oneListLineFormErrMap);
//System.out.println("oneListLineFormErrMap: " + oneListLineFormErrMap);
//System.out.println("listForm isValid: " + isValid);
                    if (!oneListLineFormErrMap.isEmpty())
                    {
                        //auf map mit [index] addieren, damit der Key in der Fehler-Map identisch ist mit dem Key im Request
                        formErrMap.mapForIndex(i, oneListLineFormErrMap);
                    }
                }
            }
        }
        return isValid;
    }

}
