package de.cnc;

import java.io.File;

import javax.servlet.http.*;

import de.cnc.domainparser.DomainParseException;
import de.cnc.expression.JspRuntimeEnvironment;
import de.cnc.flowcompiler.FlowControlCompiler;
import de.cnc.flowprogram.*;
import de.cnc.util.*;

/**
 * zentrales Singleton zum Steuern des Ablaufes der Web-Applikation. <br/>
 * Diese Klasse stellt den eigentlichen Prozessor (Control) im CnC-System dar.<br/>
 * Aufgaben:<br/>
 * &nbsp;Auslesen des Programm-Status aus dem Servlet-Request<br/>
 * &nbsp;Auswerten der request- und Session-Parameter<br/>
 * &nbsp;Aufruf der notwendigen Action-Methoden<br/>
 * &nbsp;Springen (forward) zu den entsprechenden JSPs<br/>
 *
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner Kcker</a>
 * <br/>
 *
 * @author Heiner Kcker
 * @version $Id:$
 */
public final class FlowControl
{
    /**
     * Tabelle mit den Steuer-Informationen
     */
    private static FlowControlTable flowControlTable;

    /**
     * {@link FlowSrcFileLine}-Array mit den Zeilen der Steuerdatei <b>flow.cnc</b>
     * und den Prozedur-Namen. <br/>
     */
    private static FlowSrcFileLine[] srcFileLineArr;

    /**
     * Referenz auf die konkrete Implementation des CncServlet der CnC-Applikation
     * zum Zugriff auf die permission-Methoden und den Action-Methods-Class-Package-Prefix
     */
    protected static CncServlet appServlet;

    /**
     * Konstruktor private, instanziieren verboten
     */
    private FlowControl()
    {
        ;
    }

    /**
     * ob der Fluss Browser-gesteuert (Back-Button und new-Window erlaubt) sein soll (true),<br/>
     * oder Session-gesteuert (Back-Button und new-Window nicht erlaubt) sein soll (false).<br/>
     */
    private static boolean browserLeading = false;

    /**
     * Dateinamen und Pfad der Fluss-Steuer-Datei
     */
    private static String flowCtrlFileNameWithPath;

    /**
     * Zeitpunkt der letzten nderung der Fluss-Steuer-Datei
     * fr das Hot-Reload . <br>
     */
    private static long flowCtrlFileLastModified;

    /**
     * get browserLeading . <br/>
     * ob der Fluss Browser-gesteuert (Back-Button und new-Window erlaubt) sein soll (true),<br/>
     * oder Session-gesteuert (Back-Button und new-Window nicht erlaubt) sein soll (false).<br/>
     *
     * @return browserLeading
     */
    public static boolean getBrowserLeading()
    {
        return browserLeading;
    } // end method getBrowserLeading

    /**
     * set browserLeading . <br/>
     * ob der Fluss Browser-gesteuert (Back-Button und new-Window erlaubt) sein soll (true),<br/>
     * oder Session-gesteuert (Back-Button und new-Window nicht erlaubt) sein soll (false).<br/>
     *
     * @param browserLeading new Value
     */
    public static void setBrowserLeading(boolean newBrowserLeading)
    {
        browserLeading = newBrowserLeading;
    } // end method setBrowserLeading

    /**
     * Initialisierung der Fluss-Steuerung . <br/>
     *
     * @param servlet Referenz auf aufrufendes Servlet, wird bentigt um an die Configurationseinstellungen aus <b>web.xml</b> zu kommen
     */
    static public void init(CncServlet servlet)
    {
        if (flowControlTable != null)
        {
            return;
        } // end if

        appServlet = servlet; // Anker auf Implementationsklasse CncServlet

        Log.debug("FlowControl", "init");
        synchronized (FlowControl.class)
        {
            if (flowControlTable == null)
            {
                // Abfrage des Namens des Fluss-Control-Files aus web.xml (Default: flow.cnc)
                String fileName =
                    servlet.getServletConfig().getInitParameter("flowfile");

                if (StringUtil.empty(fileName))
                {
                    fileName = "flow.cnc"; // "flow.cnc";
                } // end if
                Log.debug("Flow-File", fileName);

                // Abfrage des Pfades der Properties-Files aus web.xml
                // String  filePath = servlet.getServletConfig().getInitParameter("configroot");
                String filePath = "/WEB-INF";
                filePath =
                    servlet.getServletConfig().getServletContext().getRealPath(
                        filePath);
                filePath = filePath.trim();
                /*String*/
                flowCtrlFileNameWithPath = filePath + File.separator + fileName;

                Log.debug("Flow-File with Path", flowCtrlFileNameWithPath);

                //Ermittlung zeitpunkt letzte nderung der Fluss-Steuer-Datei fr Hot-Reload
                File tmpFile = new File(flowCtrlFileNameWithPath);
                flowCtrlFileLastModified = tmpFile.lastModified();

                try
                {
                    init(flowCtrlFileNameWithPath);
                }
                catch (DomainParseException e)
                {
                    Log.fatal(e);
                    throw new RuntimeException(e.getMessage());
                }

                // bergabe PackagePrefix, DefaultActionClassName und PackagePrefix fuer implzite Klassen
                ActionMethodClassCache.setActionPackage(
                    servlet.getActionPackage());
                ActionMethodClassCache.setDefaultActionClassName(
                    servlet.getDefaultActionClassName());
                // ActionMethodClassCache.setImplActionPackagePrefix( servlet.getImplActionPackagePrefix() ) ;
                ActionMethodClassCache.setAppImplActionPackage(
                    servlet.getAppImplActionPackage());

                // Leading-Method
                String strLeadingMethod =
                    servlet.getServletConfig().getInitParameter("flowleading");

                strLeadingMethod = StringUtil.lower(strLeadingMethod);
                setBrowserLeading(!"session".equals(strLeadingMethod));
                if (!StringUtil.empty(strLeadingMethod)
                    && !"browser".equals(strLeadingMethod)
                    && !"session".equals(strLeadingMethod))
                {
                    Log.debug(
                        "warning: Leading-Method ",
                        strLeadingMethod,
                        " unknown");
                } // end if
                Log.debug(
                    "CnC-Leading-Method",
                    (getBrowserLeading() ? "browser" : "session"));
            }
        }
    } // end method init

    /**
     * Fluss-Steuer-Datei neu einlesen, wenn gendert.
     *
     * @param servlet
     */
    public static void reload(CncServlet servlet)
    {
        File file = new File(flowCtrlFileNameWithPath);
        long tmpFlowCtrlFileLastModified = file.lastModified();
        if (tmpFlowCtrlFileLastModified > flowCtrlFileLastModified)
        {
            Log.debug("FlowControl", "reload");
            flowControlTable = null;
            init(servlet);
        }
    } // end method

    /**
     * Initialisierung anhand Steuerdatei flow.cnc
     *
     * @param fileNameWithPath
     * @throws xxxException wenn Formatfehler in Datei flow.cnc
     */
    public static void init(String fileNameWithPath /*,String pEncoding*/
    ) throws DomainParseException
    {
        flowControlTable =
            FlowControlCompiler.parseFile(fileNameWithPath /*,pEncoding*/
        );
        srcFileLineArr = FlowControlCompiler.getSrcFileLineArr();

        // Zuordnen Low-Level-Adressen (evtl. noch mal bessere Stelle dafr im Source suchen)
        // String fr Ausgabe vorbereiten
        for (int i = 0; i < srcFileLineArr.length; i++)
        {
            srcFileLineArr[i].adr = flowControlTable.getAdrForSrcLine(i + 1);
            // Basis 1
            StringBuffer sb = new StringBuffer(500);
            sb.append(StringUtil.str(i + 1, 4));
            sb.append(":");
            if (srcFileLineArr[i].adr >= 0)
            { // -1 heisst nicht gefunden
                sb.append(StringUtil.str(srcFileLineArr[i].adr, 4));
            }
            else
            {
                sb.append("    "); // 4 Spaces
            } // end if
            sb.append(":");
            sb.append(srcFileLineArr[i].lineEncodet);
            srcFileLineArr[i].lineOut = sb.toString();
        } // end for
    } // end method init

    /**
     * Ausfhren eines Befehls
     *
     * @return String mit forward-Direction
     */
    public static String execute(//--
    //final HttpServletRequest request, //-
    //final HttpServletResponse response, //-
    //final ServletContext application, //-
    final CncContext cncContext)
    {
        // throws IOException, ServletException
        Log.debug("FlowControl", "execute");

        HttpServletRequest request = cncContext.getRequest();

        // Lesen requestKey cnc_status
        String strReqCncStatus =
            request.getParameter(SessionKeys.CNC_STATUS);

        Log.debug("CNC_STATUS ", strReqCncStatus);

        // FlowStack als lokale Variable, weil fr jede Applikations-Session eine eigene Instanz existiert
        FlowStack flowStack =
            new FlowStack(strReqCncStatus, flowControlTable, request);

        String forwardStr =
            execute(flowStack /*, request, response, application*/
        , cncContext);

        // Speichern cnc_status in Session fr encode-Tags
        // Bei Methode sessionleading ausserdem speichern des FlowStack in der Session
        final String newStatus = flowStack.getEncodedCnCStatus(request);
        CncUtil.setAttribute(
            request.getSession(),
            SessionKeys.CNC_STATUS,
            newStatus);
        Log.debug("neuer Status ", newStatus);

        return forwardStr;
    } // end method execute(request,response)

    /**
     * Ausfhren eines Befehls
     *
     * @return String mit forward-Direction
     */
    private static String execute(//-
    final FlowStack flowStack, //-
    //final HttpServletRequest request, //-
    //final HttpServletResponse response, //-
    //final ServletContext application, //-
    final CncContext cncContext)
    {
        // throws IOException, ServletException
        // throws ServletException {

        HttpServletRequest request = cncContext.getRequest();
        HttpServletResponse response = cncContext.getResponse();

        HttpSession session = request.getSession();

        // Auswerten einer GOTO-Action (Navigations-Link)
        //String actionStr = CncUtil.getAction(request);

        if (cncContext.getAction().startsWith(RequestKeys.CNC_GOTO_ACTION_PREFIX))
        {
            // eine GOTO-Action ist aufgetreten (Prefix cnc_goto)
            String labelStr =
                    StringUtil.cutLeft(
                    cncContext.getAction(),
                    RequestKeys.CNC_GOTO_ACTION_PREFIX);

            Log.debug("GOTO-Action", labelStr);
            int iGotoAdr = getAdrForLabel(labelStr);

            flowStack.stackGoto(iGotoAdr);
        } // end if

        // Auswerten eines GOTO-Button (Navigations-Button)
        final String gotoButtonStr =
            CncUtil.scanForNameStartsWith(
                request,
                RequestKeys.CNC_GOTO_ACTION_PREFIX);

        if (gotoButtonStr != null)
        { // ein GOTO-Button wurde bettigt (Prefix cnc_goto)
            String labelStr =
                StringUtil.cutLeft(
                    gotoButtonStr,
                    RequestKeys.CNC_GOTO_ACTION_PREFIX);

            // .x oder .y am Ende entfernen fr Image-Button
            if (labelStr.endsWith(".x"))
            {
                labelStr = StringUtil.cutRight(labelStr, ".x");
            } // end if
            else if (labelStr.endsWith(".y"))
            {
                labelStr = StringUtil.cutRight(labelStr, ".y");
            } // end if
            Log.debug("GOTO-Button", labelStr);
            int iGotoAdr = getAdrForLabel(labelStr);

            flowStack.stackGoto(iGotoAdr);
        } // end if

        String forwardStr = null; //"";
        boolean bContinue = true;

        final JspRuntimeEnvironment jspRuntimeEnv =
            new JspRuntimeEnvironment(
                request,
                response,
                cncContext.getApplication(),
                cncContext.getNoValidate());

        while (bContinue)
        {

            int prgLine = flowStack.getAdr();

            // Abfangung gegen falsche Nummer erfolgt in getFlowElement
            FlowControlInstruction tableElement =
                flowControlTable.getFlowElement(prgLine);

            String srcStr;
            String procName = tableElement.procName; //"";

            try
            { // gegen arrayindexoutofboundsexception
                srcStr =
                    srcFileLineArr[tableElement.sourcePosition.getRow()
                        - 1].lineNotEncodet;
                // umrechnen auf Basis 0
                //procName = xmlFileLineArr[ tableElement.getSourcePosition().getRow() - 1 ].procName; // umrechnen auf Basis 0
            }
            catch (Exception e)
            {
                Log.fatal(e);
                srcStr = "";
            }

            //Log.debug( "tableElement" , tableElement );
            if (!tableElement.isNop() && !tableElement.noUserValidation)
            {
                // user-validierung
                if (!appServlet
                    .hasUserPermissionForProcedure(
                        request,
                        tableElement
                            .procName //staPaStack.getCurrentProcName(flowControlTable)
                ))
                {
                    flowStack.init();
                    // beim alternativen forward muss der vorhandene Fluss-Status zurckgestzt werden, sonst ist der benutzer in der alternativem JSP aber noch an der vorherigen Adresse
                    return appServlet.getAlternateForward(request, procName);
                    // die Implementationsklasse von CncServlet liefert einen alternativen forward, falls keine Berechtigung fr die aktuelle prozedur vorliegt
                } // end if
            } // end if

            Log.debug(
                "Adr: " + prgLine,
                " Src: " + tableElement.getLineNumber(),
                ":" + srcStr);

            //final ControlCode controlCode = tableElement.getControlCode();

            switch (tableElement.controlCode)
            {
                //if (tableElement.isNop())
                case ControlCode.NOP :
                case ControlCode.PROC_START :
                case ControlCode.FUNC_START :
                case ControlCode.LABEL_FOR_GOTO :
                case ControlCode.START_DO :
                    {
                        Log.debug("NOP", ""); // "leeres Kommando");
                        flowStack.skip(); // ein Kommando weiter schalten
                    }
                    break;
                    //else if (controlCode == ControlCode.JUMP)
                case ControlCode.JUMP :
                    {
                        Log.debug("JUMP", "");
                        flowStack.jumpTo(tableElement.jumpDestination);
                    }
                    break;
                    //else if (controlCode == ControlCode.END_WHILE)
                case ControlCode.END_WHILE :
                    {
                        Log.debug("END_WHILE", "");
                        flowStack.jumpTo(tableElement.jumpDestination);
                    }
                    break;
                    //else if (controlCode == ControlCode.JUMP_IF)
                case ControlCode.JUMP_IF :
                    {
                        Log.debug("JUMP_IF or END_DO", "");
                        if (tableElement.executeCondition(jspRuntimeEnv))
                        {
                            flowStack.jumpTo(tableElement.jumpDestination);
                        }
                        else
                        {
                            flowStack.skip(); // ein Kommando weiter schalten
                        } // end if
                    }
                    break;
                    //else if (controlCode == ControlCode.END_DO)
                case ControlCode.END_DO :
                    {
                        Log.debug("END_DO", "");
                        if (tableElement.executeCondition(jspRuntimeEnv))
                        {
                            flowStack.jumpTo(tableElement.jumpDestination);
                        }
                        else
                        {
                            flowStack.skip(); // ein Kommando weiter schalten
                        } // end if
                    }
                    break;
                    //else if (controlCode == ControlCode.JUMP_IF_NOT
                    //        || controlCode == ControlCode.START_WHILE)
                case ControlCode.JUMP_IF_NOT : // durchfallen
                case ControlCode.START_WHILE :
                    { // START_WHILE entspricht JUMP_IF_NOT
                        Log.debug("START_WHILE", tableElement.debugInfo);
                        // Log.debug("jetzt wird tableElement.excuteCondition aufgerufen "+tableElement ) ;
                        // Log.debug(" tableElement.excuteCondition="+tableElement.excuteCondition( jspRuntimeEnv ));
                        if (!tableElement.executeCondition(jspRuntimeEnv))
                        {
                            Log.debug("sprung", "end while");
                            flowStack.jumpTo(tableElement.jumpDestination);
                        }
                        else
                        {
                            Log.debug("entry ", "while");
                            flowStack.skip(); // ein Kommando weiter schalten
                        } // end if
                    }
                    break;
                    //else if (controlCode == ControlCode.LOOP_CONTINUE)
                case ControlCode.LOOP_CONTINUE :
                    { // entspricht JUMP_IF
                        Log.debug("LOOP_CONTINUE", "");
                        if (tableElement.executeCondition(jspRuntimeEnv))
                        {
                            flowStack.jumpTo(tableElement.jumpDestination);
                        }
                        else
                        {
                            flowStack.skip(); // ein Kommando weiter schalten
                        } // end if
                    }
                    break;
                    //else if (controlCode == ControlCode.LOOP_BREAK)
                case ControlCode.LOOP_BREAK :
                    { // entspricht JUMP_IF
                        Log.debug("LOOP_BREAK", "");
                        if (tableElement.executeCondition(jspRuntimeEnv))
                        {
                            flowStack.jumpTo(tableElement.jumpDestination);
                        }
                        else
                        {
                            flowStack.skip(); // ein Kommando weiter schalten
                        } // end if
                    }
                    break;
                    //else if (controlCode == ControlCode.EVAL_EXPRESSSION)
                case ControlCode.EVAL_EXPRESSSION :
                    {
                        tableElement.evalExpression(jspRuntimeEnv);
                        flowStack.skip(); // ein Kommando weiter schalten
                    }
                    break;
                    //else if (controlCode == ControlCode.ACTION)
                case ControlCode.ACTION :
                    {
                        final String actionNameStr =
                            tableElement.getJspOrActionName(jspRuntimeEnv);
                        // komplizierte Mimik bei indizierten Buttons
                        // bei der Abfrage der Bedingung wird der Index in dieser lokalen Variable vermerkt
                        // und bei Aufruf einer Action mitgegeben
                        final int iIndex = CncUtil.getCnCButtonIndex(request);

                        Log.debug("ACTION", actionNameStr, "index: " + iIndex);

                        //if (tableElement.executeCondition( jspRuntimeEnv ))
                        //{
                        if (!tableElement.noUserValidation)
                        {
                            // user-validierung
                            if (!appServlet
                                .hasUserPermissionForAction(
                                    request,
                                    actionNameStr))
                            {
                                flowStack.init();
                                // beim alternativen forward muss der vorhandene Fluss-Status zurckgestzt werden, sonst ist der benutzer in der alternativem JSP aber noch an der vorherigen Adresse
                                return appServlet.getAlternateForward(
                                    request,
                                    actionNameStr);
                                // die Implementationsklasse von CncServlet liefert einen alternativen forward, falls keine Berechtigung fr die aktuelle prozedur vorliegt
                            } // end if
                        } // end if

                        Object retObj =
                            ActionMethodClassCache.invokeActionMethod(
                                actionNameStr,
                                iIndex,
                                request,
                                response,
                                cncContext);

                        if (tableElement.returnSessionKey != null)
                        {
                            //return-Wert in Session vermerken
                            CncUtil.setAttribute(
                                session,
                                tableElement.returnSessionKey,
                                retObj);
                        }
                        //}// end if
                        flowStack.skip(); // ein Kommando weiter schalten
                    }
                    break;
                    //else if (controlCode == ControlCode.JSP)
                case ControlCode.JSP :
                    {
                        final String jspNameStr =
                            tableElement.getJspOrActionName(jspRuntimeEnv);
                        Log.debug("JSP", jspNameStr);
                        if (tableElement.executeCondition(jspRuntimeEnv))
                        {
                            if (!tableElement.noUserValidation)
                            {
                                // user-validierung
                                if (!appServlet
                                    .hasUserPermissionForJsp(
                                        request,
                                        jspNameStr))
                                {
                                    Log.debug(
                                        "no permission for jsp",
                                        jspNameStr);
                                    flowStack.init();
                                    // beim alternativen forward muss der vorhandene Fluss-Status zurckgestzt werden, sonst ist der benutzer in der alternativem JSP aber noch an der vorherigen Adresse
                                    return appServlet.getAlternateForward(
                                        request,
                                        jspNameStr);
                                    // die Implementationsklasse von CncServlet liefert einen alternativen forward, falls keine Berechtigung fr die aktuelle prozedur vorliegt
                                } // end if
                            } // end if
                            forwardStr = jspNameStr;
                            // Log.debug( "forward to " + forwardStr ) ;
                        } // end if
                        flowStack.skip(); // ein Kommando weiter schalten
                        bContinue = false;
                    }
                    break;
                    //else if (controlCode == ControlCode.GOTO)
                case ControlCode.GOTO :
                    {
                        Log.debug("GOTO", "" + tableElement.jumpDestination);
                        //if (tableElement.executeCondition( jspRuntimeEnv ))
                        //{
                        flowStack.stackGoto(tableElement.jumpDestination);
                        //}
                        //else
                        //{
                        //    staPaStack.skip(); // ein Kommando weiter schalten
                        //}// end if
                    }
                    break;
                    //else if (controlCode == ControlCode.CALL)
                case ControlCode.CALL :
                    {
                        Log.debug("CALL", tableElement.procOrLabelNameToCall);
                        //if (tableElement.executeCondition(request, response))
                        //{

                        // Parameter des calls sammeln
                        FlowProcParamList tmpFlowProcParamList = null;

                        if (tableElement.callParamArrList != null)
                        {
                            tmpFlowProcParamList = new FlowProcParamList();
                            final int callParamCount =
                                tableElement.callParamArrList.size();
                            for (int i = 0; i < callParamCount; i++)
                            {

                                // Parameter-Ifo aus call-Statement klonen
                                final FlowProcCallParam callFlowProcParam =
                                    ( tableElement
                                        .callParamArrList
                                        .get(i));
                                final FlowProcCallParam tmpFlowProcParam =
                                    callFlowProcParam
                                        .getClone();
                                //Log.debug( "\nFlowControl: tmpFlowProcedureParameter: " + tmpFlowProcedureParameter );

                                // Wert als Object vermerken
                                //if ( ! StringUtil.startsWith( tmpFlowProcedureParameter.outsideKey , "session." ) )
                                //{
                                //FlowControlCompiler.complain( "parameter prefix 'session.' expected" , tmpFlowProcedureParameter.sourcePosition );
                                //}
                                //String sessKey = StringUtil.cutLeft( tmpFlowProcedureParameter.outsideKey , "session." );

                                //                            //alten Wert sichern
                                //                            if ( tmpFlowProcedureParameter.insideKey != null )
                                //                            {
                                //                                tmpFlowProcedureParameter.backupValue = session.getAttribute(
                                //                                        FlowControlUtil.checkAndCutPrefixSession(
                                //                                            tmpFlowProcedureParameter.insideKey , tmpFlowProcedureParameter.sourcePosition ) );
                                //                            }
                                //
                                //                            //bei value-parameter (nicht per reference) wird der ehemalige
                                //                            tmpFlowProcedureParameter.outsideKey = tmpFlowProcedureParameter.insideKey;

                                //                            //ab 16.08.2004 umgestellt auf Speichern je auf einer lokalen Variable
                                //                            staPaStack.addLocalVar(
                                //                                request,
                                //                                tmpFlowProcedureParameter.insideKey,
                                //                                tmpFlowProcedureParameter.sourcePosition );

                                tmpFlowProcParamList.add(tmpFlowProcParam
                                /*, tableElement.getSourcePosition()*/
                                );

                            } // end for
                        } // end if
                        Log.debug("Parameter-values", tmpFlowProcParamList);

                        flowStack.call(tableElement.jumpDestination);

                        // Parameter an procedure bergeben
                        FlowControlInstruction procBeginInstr =
                            flowControlTable.getFlowElement(
                                tableElement.jumpDestination);

                        if (tmpFlowProcParamList != null
                            /* sollte stimmen, prft der Parser: && procBeginInstr.procParamList != null*/
                            )
                        {
                            // Sessionwerte fr Parameter und lokale Variable in der Prozedur setzen
                            final int tmpFlowProcParamListSize =
                                tmpFlowProcParamList.size();
                                for (int i = 0;
                                    i < tmpFlowProcParamListSize
                                /* die Grsse muss gleich sein, wird durch den Parser abgesichert: && i < procBeginInstr.procParamList.size()*/;
                                    i++)
                            {
                                // Session-Key intern
                                String prmInternName =
                                    procBeginInstr.procParamList.get(
                                        i).insideKey;
                                //System.out.println( "******* prmInternName " + prmInternName ) ;
                                FlowProcCallParam tmpFlowProcedureParameter =
                                    tmpFlowProcParamList.get(i);

                                tmpFlowProcedureParameter.insideKey =
                                    prmInternName;
                                //tmpFlowProcedureParameter.outsideKey = prmInternName;

                                // Parameter-Info zum zurckschreiben vermerken
                                flowStack
                                    .addParam(request, tmpFlowProcedureParameter
                                /*, tmpFlowProcedureParameter.sourcePosition*/
                                );

                                // bergebenen Wert in Session umkopieren
                                // value oder reference parameter
                                //System.out.println( "parameter expression: " + tmpFlowProcedureParameterList.get(i).paramExpression );

                                Object valueObject =
                                    tmpFlowProcParamList.get(
                                        i).evalParamExpression(
                                        jspRuntimeEnv);

                                //System.out.println( "parameter value: " + valueObject);
                                //if ( tmpFlowProcedureParameter.insideKey != null )
                                //{
                                //                                CncUtil.setAttribute(session,
                                //                                    FlowControlUtil.checkAndCutPrefixSession(
                                //                                            tmpFlowProcedureParameter.insideKey , tmpFlowProcedureParameter.sourcePosition ) , valueObject);

                                //System.out.println( "param key: " + tmpFlowProcedureParameter.insideKey );
                                CncUtil.setAttribute(
                                    session,
                                    tmpFlowProcedureParameter.insideKey,
                                    valueObject);

                                //}

                                // Vermerken, ob es sich um einen Referenz-Parameter handelt (ist in der proc-declaration hinterlegt
                                if (procBeginInstr
                                    .procParamList
                                    .get(i)
                                    .isRefParam)
                                {
                                    //System.out.println( "FlowControl: isRefParam");

                                    tmpFlowProcedureParameter.isRefParam = true;

                                    tmpFlowProcedureParameter.outsideKey =
                                        FlowControlUtil
                                            .checkAndCutPrefixSession(
                                            tmpFlowProcedureParameter
                                                .paramExpression
                                                .getSource(),
                                            tmpFlowProcedureParameter
                                                .sourcePosition);
                                    //System.out.println( "FlowControl: outsideKey: " + tmpFlowProcedureParameter.outsideKey );

                                }

                            } // end for
                        } // end if

                        // returnSessionKey-Attribute des call-Statements fr Funktionen
                        if (tableElement.returnSessionKey != null)
                        {
                            flowStack.addReturnOutsideKey(
                                tableElement.returnSessionKey);
                            //   System.out.println( "returnSessionKey: " + tableElement.returnSessionKey);
                        }

                        //}
                        //else
                        //{
                        //    staPaStack.skip(); // ein Kommando weiter schalten
                        //}// end if
                    }
                    break;
                    //else if (controlCode == ControlCode.PROC_RETURN)
                case ControlCode.PROC_RETURN :
                    {
                        Log.debug("PROC_RETURN", tableElement.procName);
                        flowStack.procReturn(request);
                    }
                    break;
                case ControlCode.FUNC_RETURN :
                    {
                        Log.debug("FUNC_RETURN", tableElement.procName);
                        String returnSessionKey =
                            flowStack.getReturnOutsideKey();

                        //System.out.println( "FlowControl#PROC_RETURN returnExpression: " + tableElement.returnExpression );
                        //return-Expression ausfhren
                        Object retObj =
                            tableElement.evalReturnExpression(jspRuntimeEnv);

                        flowStack.procReturn(request);

                        //System.out.println( "FlowControl#FUNC_RETURN retObj: " + retObj );
                        //System.out.println( "FlowControl#FUNC_RETURN returnSessionKey: " + returnSessionKey );

                        if ( returnSessionKey != null ) // ein Funktionsaufruf ist auch ohne Speichern des Rckgabewertes mglich
                        {
                            //return-Wert in Session vermerken
                            CncUtil.setAttribute(session, returnSessionKey, retObj);
                        }
                    }
                    break;
                    //            else if (tableElement.getControlCode() == ControlCode.SET)
                    //            {
                    //                Log.debug("SET " + tableElement.getSessionKey() + "value:" + tableElement.getValue()
                    //                        + "class:" + tableElement.getSetClass() + "fromSessionKey:"
                    //                        + tableElement.getFromSessionKey() + "fromRequestKey:"
                    //                        + tableElement.getFromRequestKey() + "fromAction:"
                    //                        + tableElement.getFromAction());
                    //
                    //                if (!StringUtil.empty(tableElement.getSetClass()))
                    //                { // keinen String sondern Klasse laut Name erzeugen
                    //                    String className = tableElement.getSetClass();
                    //
                    //                    try
                    //                    {
                    //                        Class myClass = Class.forName(className);
                    //                        Object obj = myClass.newInstance();
                    //
                    //                        CncUtil.setAttribute(session, tableElement.getSessionKey(), obj);
                    //                    }
                    //                    catch (Exception e)
                    //                    {
                    //                        throw new RuntimeException("cant load Class " + className);
                    //                    }
                    //                }
                    //                else if (!StringUtil.empty(tableElement.getValue()))
                    //                {  // String in Session setzen
                    //                    CncUtil.setAttribute(session, tableElement.getSessionKey(), tableElement.getValue());
                    //                }// end if
                    //                else if (!StringUtil.empty(tableElement.getFromSessionKey()))
                    //                {  // String in Session setzen
                    //                    Object objFromSessionKey = session.getAttribute(tableElement.getFromSessionKey());
                    //
                    //                    if (objFromSessionKey != null)
                    //                    {
                    //                        CncUtil.setAttribute(session, tableElement.getSessionKey(), objFromSessionKey);
                    //                    }
                    //                    else
                    //                    { // Quelle war null
                    //                        session.removeAttribute(tableElement.getFromSessionKey());
                    //                    }// end if
                    //                }// end if
                    //                else if (!StringUtil.empty(tableElement.getFromRequestKey()))
                    //                { // fromRequestKey
                    //                    Object objFromRequestKey = request.getParameter(tableElement.getFromRequestKey());
                    //
                    //                    if (objFromRequestKey != null)
                    //                    {
                    //                        CncUtil.setAttribute(session, tableElement.getSessionKey(), objFromRequestKey);
                    //                    }
                    //                    else
                    //                    { // Quelle war null
                    //                        session.removeAttribute(tableElement.getFromRequestKey());
                    //                    }// end if
                    //                }// end if
                    //                else if (!StringUtil.empty(tableElement.getFromAction()))
                    //                { // fromAction
                    //                    if (actionStr != null)
                    //                    {
                    //                        CncUtil.setAttribute(session, tableElement.getSessionKey(), actionStr);
                    //                    }
                    //                    else
                    //                    { // Quelle war null
                    //                        session.removeAttribute(tableElement.getFromRequestKey());
                    //                    }// end if
                    //                }// end if
                    //                else
                    //                {
                    //                    throw new RuntimeException("undefined set instruction" + tableElement);
                    //                }// end if
                    //                staPaStack.skip(); // ein Kommando weiter schalten
                    //            }
                    //else if (controlCode == ControlCode.RESET)
                case ControlCode.RESET :
                    {
                        final String resetNameStr =
                            tableElement.getJspOrActionName(jspRuntimeEnv);
                        //Log.debug("RESET " + resetNameStr);
                        if (!StringUtil.empty(resetNameStr))
                        {
                            Log.debug("RESET", resetNameStr);
                            session.removeAttribute(resetNameStr);
                        }
                        else
                        { // kein SessionKey bergeben, also gesamte Session lschen
                            Log.debug("RESET", "all");
                            // CncUtil.invalidate( request.getSession( true ) ) ;
                            // session.invalidate() ;
                            CncUtil.removeAll(session);
                        } // end if
                        flowStack.skip(); // ein Kommando weiter schalten
                    }
                    //            else if (tableElement.getControlCode() == ControlCode.SESSION_INVALIDATE)
                    //            {
                    //                Log.debug("SESSION_INVALIDATE");
                    //                if (tableElement.executeCondition(request, response))
                    //                {
                    //                    session.invalidate();
                    //                }
                    //                else
                    //                {
                    //                    staPaStack.skip(); // ein Kommando weiter schalten
                    //                }// end if
                    //                staPaStack.skip(); // ein Kommando weiter schalten
                    //            }
                    //            else if (tableElement.getControlCode() == ControlCode.INC)
                    //            {
                    //                String valueStr = tableElement.getValue();
                    //                int iValue = 1;
                    //
                    //                if (!StringUtil.empty(valueStr))
                    //                {
                    //                    iValue = StringUtil.str2int(valueStr);
                    //                }
                    //                Object sessionObject = session.getAttribute(tableElement.getSessionKey());
                    //                String sessionString = "" + sessionObject;
                    //
                    //                sessionString = "" + (StringUtil.str2double(sessionString) + iValue);
                    //                sessionString = StringUtil.cutRight(sessionString, ".0"); // auf ganze Zahl konvertieren, wenn mglich
                    //                CncUtil.setAttribute(session, tableElement.getSessionKey(), sessionString);
                    //                staPaStack.skip(); // ein Kommando weiter schalten
                    //            }
                    //            else if (tableElement.getControlCode() == ControlCode.DEC)
                    //            {
                    //                String valueStr = tableElement.getValue();
                    //                int iValue = 1;
                    //
                    //                if (!StringUtil.empty(valueStr))
                    //                {
                    //                    iValue = StringUtil.str2int(valueStr);
                    //                }
                    //                Object sessionObject = session.getAttribute(tableElement.getSessionKey());
                    //                String sessionString = "" + sessionObject;
                    //
                    //                sessionString = "" + (StringUtil.str2double(sessionString) - iValue);
                    //                sessionString = StringUtil.cutRight(sessionString, ".0"); // auf ganze Zahl konvertieren, wenn mglich
                    //                CncUtil.setAttribute(session, tableElement.getSessionKey(), sessionString);
                    //                staPaStack.skip(); // ein Kommando weiter schalten
                    //            }
                    break;
                    //else if (controlCode == ControlCode.LOCAL_VAR)
                case ControlCode.LOCAL_VAR :
                    {
                        flowStack.addLocalVar(
                            request,
                            tableElement.getLocalVarNameWithoutPrefix(),
                            tableElement.sourcePosition);
                        flowStack.skip(); // ein Kommando weiter schalten
                    }
                    break;
                    //else
                default :
                    {
                        throw new RuntimeException(
                            "unknown Command " + tableElement.controlCode);
                    }
            }

            if (bContinue
                && session.getAttribute(SessionKeys.CNC_SINGLE_STEP_DEBUG)
                    != null)
            {
                // Einzelschritt-Debug
                // nach jedem Schritt wird zu einer speziellen Debug-JSP verzweigt
                forwardStr = "/cncdebug/singlestepdebug.jsp";
                // noch konfigurierber machen
                // staPaStack.skip() ; // ein Kommando weiter schalten
                bContinue = false;

                // Vermerken der aktuellen Action, damit die singlestepdebug.jsp die
                // indentische Action weitergibt
                CncUtil.setAttribute(
                    session,
                    SessionKeys.CNC_SINGLE_STEP_DEBUG_ACTION,
                    cncContext.getAction());
                CncUtil.saveRequestInSession(request);
            } // end if Einzelschritt-Debug
        } // end while

        Log.debug("forward to", forwardStr);
        return forwardStr;
    } // end method execute(Stack,request,response)

    /**
     * Ausgabe Debug-Information mit gesamter Flussdatei flow.cnc . <br/>
     * Der Pfeil in der Darstellung der XML-Datei zeigt jeweils auf den nchsten auszufhrenden Befehl . <br/>
     *
     * @param request fr Debug-Ausgabe Request und Session
     * @return String mit HTML-formatierter Debug-Ausgabe
     */
    public static String getHtmlDebugWithEntireFlow(HttpServletRequest request)
    {
        return getHtmlDebug(request, true);
    } // end method getHtmlDebugWithEntireFlow

    /**
     * Ausgabe Debug-Information ohne gesamte Flussdatei flow.cnc . <br/>
     * Der Pfeil in der Darstellung der XML-Datei zeigt jeweils auf den nchsten auszufhrenden Befehl . <br/>
     *
     * @param request fr Debug-Ausgabe Request und Session
     * @return String mit HTML-formatierter Debug-Ausgabe
     */
    public static String getHtmlDebugWithoutEntireFlow(HttpServletRequest request)
    {
        return getHtmlDebug(request, false);
    } // end method getHtmlDebugWithoutEntireFlow

    /**
     * Ausgabe Debug-Information mit Optionsschalter fr gesamte Flussdatei flow.cnc . <br/>
     * Der Pfeil in der Darstellung der XML-Datei zeigt jeweils auf den nchsten auszufhrenden Befehl . <br/>
     *
     * @param request fr Debug-Ausgabe Request und Session
     * @return String mit HTML-formatierter Debug-Ausgabe
     */
    private static String getHtmlDebug(
        HttpServletRequest request,
        boolean pPaEntireFlow)
    {
        HttpSession session = request.getSession(true);
        String cnc_status =
            (String) CncUtil.getAttribute(
                session,
                SessionKeys.CNC_STATUS,
                String.class);
        FlowStack tmpFlowStack =
            new FlowStack(cnc_status, flowControlTable, request);

        int prgLine = tmpFlowStack.getPreAdr();

        FlowControlInstruction tableElement =
            flowControlTable.getFlowElement(prgLine);
        int srcLineNumber = tableElement.getLineNumber();

        StringBuffer sb = new StringBuffer();

        sb.append("<pre>\n");
        // sb.append( "Flow-Debug\n" ) ;
        // sb.append( "----------\n" ) ;
        // sb.append( "\n" ) ;
        if (session.getAttribute(SessionKeys.CNC_SINGLE_STEP_DEBUG) != null)
        {
            // Einzelschritt-Debug

            // sb.append( "Single-Step-Debug is on\n" ) ;
            // sb.append( "\n" ) ;

            // sb.append( "gltige Action: " + session.getAttribute( SessionKeys.CNC_SINGLE_STEP_DEBUG_ACTION ) + "\n" ) ;
            sb.append("gltige Action: " + CncUtil.getAction(request) + "\n");
            sb.append("\n");
        } // end if

        sb.append(tmpFlowStack.toHtmlString(flowControlTable, srcFileLineArr));
        // sb.append( "\n" ) ;
        // sb.append( "flow.cnc\n" ) ;
        // sb.append( "\n" ) ;

        // Ausgabe augenblicklich (beim letzten Befehl) laufende Prozedure
        String strCurrProcName = tmpFlowStack.getPreProcName(flowControlTable);

        if (!pPaEntireFlow || !"main".equals(strCurrProcName))
        { // procedure main nicht doppelt ausgeben
            sb.append("\n");

            // sb.append( "current procedure: " + strCurrProcName + "\n" ) ;
            // sb.append( "-----------------------------------------------------\n" ) ;
            sb.append("<u>current procedure: " + strCurrProcName + "</u>\n\n");

            sb.append("    Src: Adr:\n");
            for (int i = 0; i < srcFileLineArr.length; i++)
            {
                if (StringUtil
                    .stringEqual(srcFileLineArr[i].procName, strCurrProcName))
                {
                    // nur Zeilen der aktuellen procedure ausgeben
                    final int decSrcLineNumber = srcLineNumber - 1;
                    final boolean isCurrentSrcLine = (decSrcLineNumber == i);

                    sb.append((isCurrentSrcLine ? "<b>-> " : "   "));
                    sb.append(srcFileLineArr[i].lineOut);
                    sb.append((isCurrentSrcLine ? "</b>" : ""));
                    sb.append("\n");
                } // end if
            } // end for
        } // end if

        // Ausgabe gesamte flow.cnc
        if (pPaEntireFlow)
        {
            // sb.append( "\n" ) ;
            sb.append("\n");

            // sb.append( "flow.cnc\n" ) ;
            // sb.append( "--------\n" ) ;
            sb.append("<u>flow.cnc</u>\n\n");

            sb.append("   Src: Adr:\n");
            for (int i = 0; i < srcFileLineArr.length; i++)
            {
                int decXmlLineNumber = srcLineNumber - 1;

                sb.append((decXmlLineNumber == i ? "<b>-> " : "   "));
                sb.append(srcFileLineArr[i].lineOut);
                sb.append((decXmlLineNumber == i ? "</b>" : ""));
                sb.append("\n");
            } // end for
        }

        sb.append("</pre>\n");
        return sb.toString();
    } // end method getHtmlDebug

    /**
     * Suchen eines Labels oder einer Prozedur mit Namen entsprechend dem String-Parameter
     *
     * @param strPaLabel Name des Labels oder der Prozedur
     * @return Adresse des Labels oder der Prozedur
     * @exception wirft eine RuntimeException, wenn Name ungltig
     */
    private static int getAdrForLabel(String strPaLabel)
    {
        for (int i = 0; i < flowControlTable.size(); i++)
        {
            FlowControlInstruction suchElement =
                (FlowControlInstruction) flowControlTable.get(i);

            if (suchElement.controlCode == ControlCode.LABEL_FOR_GOTO
                && strPaLabel.equals(suchElement.procOrLabelNameToCall))
            {
                return i;
            } // end if
            if ( ( suchElement.controlCode == ControlCode.PROC_START
                  || suchElement.controlCode == ControlCode.FUNC_START )
                && strPaLabel.equals(suchElement.procName))
            {
                return i;
            } // end if
        } // end for
        // nichts gefunden
        throw new RuntimeException(
            "procedure or label " + strPaLabel + " not found");
    } // end method getAdrForLabel

    /**
     * herausreichen der flowControlTable zum Debugging . <br/>
     * siehe {@link CncServlet#init} . <br/>
     */
    public static FlowControlTable getFlowControlTable()
    {
        return flowControlTable;
    }

} // end class FlowControl
