package de.cnc;

import java.util.Stack;

import javax.servlet.http.*;

import de.cnc.domainparser.util.SourcePosition;
import de.cnc.flowprogram.*;
import de.cnc.util.*;

/**
 * Programm-Status- und Variablen-Stapel des Flusses der Web-Applikation {@link FlowControl} . <br/>
 * Benutzt {@link FlowStackElement}.<br/>
 *
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner Kcker</a>
 * <br/>
 *
 * @author Heiner Kcker
 * @version $Id:$
 */
public final class FlowStack
{

   /**
    * Tabelle mit den Steuer-Informationen
    */
   private Stack flowStack;

   /**
    * parameterloser Konstruktor, Verwendung verboten
    */
   private FlowStack()
   {
      ;
   }

   /**
    * Konstruktor mit String-Parameter, der den CnC-Status aus href oder action (encodete Tags) enthlt
    *
    * @param strPaCnCStatus vom Browser als codierter String gesendeter CnC-Status
    * @param flowControlTablePa zur Kontrolle Einhaltung Tabellengrenzen (wenn Adresse grsser als mglich, dann abweisen Request)
    * @param request zum Setzen Fehlermeldung: CNC-Status corrupted !!! Dontuse Back-Button or new window.
    *
    * @return Fluss-Stack entsprechend geparstem Status-String oder leerer Stack, wenn Status-String invalid
    */
   public FlowStack(String strPaCnCStatus, FlowControlTable flowControlTablePa, HttpServletRequest request)
   {
      flowStack = parseCncStatus(strPaCnCStatus, flowControlTablePa, request);
   } // end constructor

   /**
    * Stack mit Start an Adresse 0 neu aufbauen
    */
   public void init()
   {
      flowStack = new Stack();
      flowStack.push(new FlowStackElement(0, 0));
   } // end method init

   /**
    * Parsen des CnC-Status. <br/>
    * Codierung als Adresse_AdresseInRufenderProzedur_AdresseInRufenderProzedur.<br/>
    * dabei wird geprft, ob eine Adresse ausserhalb des gltigen Bereichs auftritt.<br/>
    *
    * @param strPaCoded vom Browser als codierter String gesendeter CnC-Status
    * @param flowControlTablePa zur Kontrolle Einhaltung Tabellengrenzen (wenn Adresse grsser als mglich, dann abweisen Request)
    * @param request zum Setzen Fehlermeldung: CNC-Status corrupted !!! Dontuse Back-Button or new window.
    *
    * @return Fluss-Stack entsprechend geparstem Status-String oder leerer Stack, wenn Status-String invalid
    */
   private Stack parseCncStatus( //--
      String pCodedStatusStr, //--
      final FlowControlTable pFlowControlTable, //--
      final HttpServletRequest request)
   {
      if (FlowControl.getBrowserLeading())
      {
         // BROWSER_LEADING
         // Status wird zum Browser gesendet und vom Browser-Request wieder aufgenommen
         // Back-Button und new Window erlaubt
         // hier knnte noch ein entschlsseler fr den cnc_status aufgerufen werden
         Stack retStack = new Stack();

         if (pCodedStatusStr == null)
         {
            retStack.push(new FlowStackElement(0, 0));
            return retStack;
         } // end if

         while (!StringUtil.empty(pCodedStatusStr))
         {
            Log.debug("parseCncStatus" , pCodedStatusStr);
            String partStr = null;

            if (pCodedStatusStr.indexOf('_') > -1)
            {
               // Abschneiden des Teiles vor dem UnderScore
               partStr = pCodedStatusStr.substring(0, pCodedStatusStr.indexOf('_'));
               pCodedStatusStr = StringUtil.cutLeft(pCodedStatusStr, partStr);
               pCodedStatusStr = StringUtil.cutLeft(pCodedStatusStr, "_");
            }
            else
            {
               partStr = pCodedStatusStr;
               pCodedStatusStr = null;
            } // end if
            // Log.debug( "partStr " + partStr ) ;
            // Log.debug( "strPaCoded " + strPaCoded ) ;

            int iAdr = StringUtil.str2int(partStr);

            // Plausibilitt des bergebenen Status prfen
            if (iAdr < 0 || iAdr >= pFlowControlTable.size())
            {
               Log.debug("****************************");
               Log.debug("* CNC-Status corrupted !!! *");
               Log.debug("****************************");
               CncUtil.setErrMessage(
                  request,
                  "CNC-Status corrupted !!! Dontuse Back-Button or new window or browser cache.");
               
               //Brutal raus-schmeissen wegen instabilen Verhaltens
               request.getSession().invalidate();
               
               return new Stack();
            }
            else if (iAdr == 0)
            {
               // genau 0 kann eigentlich nur nach logon.jsp vorkommen
               Log.debug("adress 0 is raised","stack initialized");
               return new Stack();
            } // end if

            retStack.push(new FlowStackElement(iAdr, iAdr - 1 /* @todo: das ist nicht korrekt, alte adresse codiert mit zum browser senden, debug-schalter einbauen und beachten */ ) );
         } // end while

         return retStack;
      }
      else
      {
         // SESSION_LEADING
         // Status ist in der Session gespeichert
         // Zum Browser wird eine Prfnummer gesendet, welche die Einhaltung des Flusses sichert
         // bei Abweichung wird zur Adresse 0 gesprungen

         Log.debug("parseCncStatus" , pCodedStatusStr);

         HttpSession session = request.getSession(true);
         String strSessionCncStatus = (String) session.getAttribute(SessionKeys.CNC_STATUS);

         if (StringUtil.empty(strSessionCncStatus))
         {
            // Status ist leer, wahrscheinlich Applikations-Start
            Log.debug("parseCncStatus" , "Status ist leer, wahrscheinlich Applikations-Start");
            Stack retStack = new Stack();
            return retStack;
         }
         else if (
            CompareUtil.notEquals(
               pCodedStatusStr,
               "cncStatusNoCheck") // Check nicht ausgeschalten  @todo: noch Config-Schalter machen, ob "cncStatusNoCheck" berhaupt erlaubt ist, @todo: Konstante machen
               && CompareUtil.notEquals(pCodedStatusStr, strSessionCncStatus))
         {
            // Status nicht ok
            Log.debug("****************************");
            Log.debug("* CNC-Status corrupted !!! *");
            Log.debug("****************************");
            CncUtil.setErrMessage(request, "CNC-Status corrupted !!! Dontuse Back-Button or new window or browser cache.");
            Stack retStack = new Stack();

            return retStack;
         }
         else
         {
            // alles ok
            Stack retStack = (Stack) session.getAttribute(SessionKeys.CNC_FLOWSTACK_KEY);
            Log.debug("parseCncStatus alles ok" , retStack);

            return retStack;
         } // end if
      } // end if
   } // end method parseCncStatus

   /**
    * Abfragen String mit encodetem CnC-Status zur Weitergabe an href oder action (form) . <br/>
    * Codieren des CnC-Status in einen String zur Mitgabe an Links und Actions. <br/>
    * Bei Methode browserleading Codierung als Adresse_AdresseInRufenderProzedur_AdresseInRufenderProzedur.<br/>
    * Bei Methode sessionleading, Mitgabe Prf-Key und speichern des FlowStack in der Session.<br/>
    *
    * @param request zum Zugriff auf Session zum Speichern Stack
    *
    * @return codierter (encodeter) Status-String
    */
   public String getEncodedCnCStatus(HttpServletRequest request)
   {
      // behandeln browser- oder session-leading
      if (FlowControl.getBrowserLeading())
      {
         // BROWSER_LEADING
         // Status wird zum Browser gesendet und vom Browser-Request wieder aufgenommen
         // Back-Button und new Window erlaubt
         if (flowStack == null)
         {
            return "";
         } // end if

         String retString = "";

         while (flowStack.size() > 0)
         {
            if (!StringUtil.empty(retString))
            {
               retString = "_" + retString;
            } // end if

            retString = "" + ((FlowStackElement) flowStack.pop()).adress + retString;
         } // end while
         // hier knnte noch ein Verschlsseler fr den cnc_status aufgerufen werden
         return retString;
      }
      else
      {
         // SESSION_LEADING
         // Status ist in der Session gespeichert
         // Zum Browser wird eine Prfnummer gesendet, welche die Einhaltung des Flusses sichert
         // bei Abweichung wird zur Adresse 0 gesprungen

         // Speichern Stack in Session
         HttpSession session = request.getSession();

         CncUtil.setAttribute(session, SessionKeys.CNC_FLOWSTACK_KEY, flowStack);

         String retString = "" + Math.round(Math.random() * Long.MAX_VALUE);

         return retString;
      } // end if
   } // end method getEncodedCnCStatus

   /**
    * Einen Befehl weiter schalten
    */
   public void skip()
   {
      // WaitUtil.wait( 1000 ) ; // fr langsamen Ablauf beim Debugging
      // Fehlerabfangungen
      if (flowStack == null)
      {
         init();
         Log.debug("flowStack was null");
         return; // erst mal neu anfangen
      } // end if
      if (flowStack.size() < 1)
      {
         init();
         Log.debug("flowStack was empty");
         return; // erst mal neu anfangen
      } // end if

      FlowStackElement flowStackElement = (FlowStackElement) flowStack.peek();

      flowStackElement.preAdress = flowStackElement.adress; // vorherige Adresse fr Anzeig in Debug vermerken
      flowStackElement.adress++;
      Log.debug("skip to" , flowStackElement.adress);
   } // end method skip

   /**
    * einen bestimmten Befehl anspringen, Low-Level-Jump innerhalb einer procedure, kein GOTO
    *
    * @param iPaAdr anzuspringende Adresse
    */
   public void jumpTo(int iPaAdr)
   {
      // WaitUtil.wait( 1000 ) ;
      // Fehlerabfangungen
      if (flowStack == null)
      {
         init();
         Log.debug("flowStack was null");
         return; // erst mal neu anfangen
      } // end if
      if (flowStack.size() < 1)
      {
         init();
         Log.debug("flowStack was empty");
         return; // erst mal neu anfangen
      } // end if

      if (flowStack.size() < 1)
      {
         init();
         return; // erst mal neu anfangen
      } // end if

      // flowStack.pop() ;
      // Log.debug( "jumpTo " + iPaAdr ) ;
      // flowStack.push( new FlowStackElement( iPaAdr ) ) ;

      FlowStackElement flowStackElement = (FlowStackElement) flowStack.peek();
      flowStackElement.preAdress = flowStackElement.adress; // vorherige Adresse fr Anzeig in Debug vermerken
      flowStackElement.adress = iPaAdr;
      Log.debug("jumpTo to" , flowStackElement.adress);
   } // end method jumpTo

   /**
    * Unterprogramm aufrufen
    *
    * @param iPaAdr anzuspringende(aufzurufende) Adresse
    */
   public void call(int iPaAdr)
   {
      // WaitUtil.wait( 1000 ) ;
      // Fehlerabfangungen
      if (flowStack == null)
      {
         init();
         Log.debug("flowStack was null");
         return; // erst mal neu anfangen
      } // end if
      if (flowStack.size() < 1)
      {
         init();
         Log.debug("flowStack was empty");
         return; // erst mal neu anfangen
      } // end if

      int preAdr = getAdr(); // vermerken Absprungadresse fr Anzeige in Debug
      skip(); // damit nach dem RETURN kein skip noetig ist
      Log.debug("call" , iPaAdr);
      flowStack.push(new FlowStackElement(iPaAdr, preAdr));
   } // end method call

   /**
    * aus Unterprogramm zurckkehren
    *
    * @param request zum Zugriff auf Session zum Zurck-Kopieren Backups fr lokale Variable
    */
   public void procReturn(HttpServletRequest request)
   {
      // WaitUtil.wait( 1000 ) ;
      Log.debug("FlowStack" , "procReturn");
      // Fehlerabfangung
      if (flowStack == null)
      {
         init();
         Log.debug("flowStack was null");
         return; // erst mal neu anfangen
      } // end if

      // wenn der Stack keine tiefere Schachtelungsebene hat wird zur Adresse Null gesprungen
      // kann vorkommen (und ist erlaubt), wenn die Prozedur ber Goto (Navigationslinks) angesprungen wurde
      if (flowStack.size() < 1)
      {
         init();
         Log.debug("flowStack was empty");
         return; // erst mal neu anfangen
      } // end if

      if (flowStack.size() > 1)
      {
         FlowStackElement flowStackElement = (FlowStackElement) flowStack.pop();

         Log.debug("" + flowStackElement);
         flowStackElement.restore(request); // Zurck-Kopieren Backups fr lokale Variable und Parameter
         // flowStack.pop() ;
      }
      else
      {
         // Stack hat nur eine Ebene und kann gar kein return ausfhren
         // Das kann passieren, wenn die procedure ber ein GOTO gestartet wurde (ist erlaubt um Navigationslinks zu untersttzen)
         init();
      } // end if

   } // end method proc_return

   /**
    * Stack abrumen und einen bestimmten Befehl anspringen
    *
    * @param iPaAdr anzuspringende Adresse
    */
   public void stackGoto(int iPaAdr)
   {
      // WaitUtil.wait( 1000 ) ;
      Log.debug("stackGoto " + iPaAdr);
      // Fehlerabfangungen
//      if (flowStack == null)
//      {
//         init();
//         Log.debug("flowStack was null");
//         return; // erst mal neu anfangen
//      } // end if
//      if (flowStack.size() < 1)
//      {
//         init();
//         Log.debug("flowStack was empty");
//         return; // erst mal neu anfangen
//      } // end if

      int preAdr = getAdr(); // vermerken Absprungadresse fr Anzeige in Debug
      flowStack = new Stack();
      flowStack.push(new FlowStackElement(iPaAdr, preAdr));
   } // end method stackGoto

   /**
    * Abfrage aktuelle Adresse
    *
    * @return aktuelle Adresse als int-Wert
    */
   public int getAdr()
   {
      // Fehlerabfangungen
      if (flowStack == null)
      {
         init();
         Log.debug("flowStack was null");
         return 0; // erst mal neu anfangen
      } // end if
      if (flowStack.size() < 1)
      {
         init();
         Log.debug("flowStack was empty");
         return 0; // erst mal neu anfangen
      } // end if

      return ((FlowStackElement) flowStack.peek()).adress;
   } // end method getAdr

   /**
    * Abfrage vorherige Adresse fr Debug
    *
    * @return aktuelle Adresse als int-Wert
    */
   public int getPreAdr()
   {
      // Fehlerabfangungen
      if (flowStack == null)
      {
         init();
         Log.debug("flowStack was null");
         return 0; // erst mal neu anfangen
      } // end if
      if (flowStack.size() < 1)
      {
         init();
         Log.debug("flowStack was empty");
         return 0; // erst mal neu anfangen
      } // end if

      return ((FlowStackElement) flowStack.peek()).preAdress;
   } // end method getAdr

   /**
    * Debug-Ausgabe der einzelnen Adressen im Stack
    */
   public String toString()
   {
      StringBuffer sb = new StringBuffer(1024 * 2);

      sb.append("FlowStack");
      if (flowStack == null)
      {
         sb.append(" is null");
      }
      else
      {
         sb.append("\n");
         for (int i = flowStack.size() - 1; i >= 0; i--)
         {
            sb.append(
               "Adr "
                  + StringUtil.leftPad("" + i, 3)
                  + ":"
                  + StringUtil.leftPad("" + ((FlowStackElement) flowStack.get(i)).adress, 5));
            // + "\n" ) ;
         } // end for
      } // end if
      return sb.toString();
   } // end method toString

   /**
    * Debug-Ausgabe der einzelnen Adressen im Stack in Tabellenform . <br/>
    * Aufruf in {@link FlowControl#getHtmlDebug}
    *
    * @param flowControlTable aus der Table kann diese Methode anhand der Adresse die XML-Linenumber und den proc-namen ermitteln
    * @param paXmlFileLineArr fr die Ausgabe der jeweiligen XML-Zeile
    *
    * @return String mit Stack-Status in Tabellenform
    */
   public String toHtmlString(FlowControlTable flowControlTablePa, FlowSrcFileLine[] paXmlFileLineArr)
   {
      StringBuffer sb = new StringBuffer(1024 * 4);

      sb.append("<u>FlowStack</u>");
      if (flowStack == null)
      {
         sb.append(" is null");
         sb.append("\n");
      }
      else if (flowStack.size() < 1)
      {
         sb.append(" is empty");
         sb.append("\n");
      }
      else
      {
         sb.append("\n");
         sb.append("\n");
         sb.append("Adr.|Src-Line|   Procedure        |\n");
         sb.append("----+--------+--------------------+-----------------------------------------------------------------------\n");
         for (int i = flowStack.size() - 1; i >= 0; i--)
         {

            int iAdr = ((FlowStackElement) flowStack.get(i)).preAdress;

            if (iAdr >= 0 && iAdr < flowControlTablePa.size())
            {
               FlowControlInstruction element = (FlowControlInstruction) flowControlTablePa.get(iAdr);
               int iSrcLine = element.getLineNumber();
//               if (i < flowStack.size() - 1)
//               {
//                  iSrcLine--; // auf dem Stack die Rckkehr-Adresse, hier Anzeige der Absprungadresse (-1)
//               }

               sb.append(StringUtil.str(iAdr, 4) + "|" + StringUtil.str(iSrcLine, 8) + "|");
               sb.append(StringUtil.pad(element.procName, 20) + "|");
               if (iSrcLine > 0 && iSrcLine <= paXmlFileLineArr.length)
               {
                  sb.append(paXmlFileLineArr[iSrcLine - 1].lineEncodet);
               } // end if
               sb.append("\n");
            } // end if

            FlowStackElement flowStackElement = (FlowStackElement) flowStack.get(i);

            if (flowStackElement.procParamList != null)
            {
               // Ausgabe Backups fr lokale Variable
               sb.append("    |        |                    | *** local vars ***\n");
               for (int j = 0; j < flowStackElement.procParamList.size(); j++)
               {
                  FlowProcCallParam flowProcParam =
                     flowStackElement.procParamList.get(j);

                  if (!flowProcParam.isRefParam)
                  { // value-param
                     sb.append(
                        "    |        |                    | value   "
                           + flowProcParam.insideKey
                           + " -> "
                           + Log.objToStrListCompact(flowProcParam.backupValue)
                           + "\n");
                  }
                  else
                  { // ref-param
                     sb.append(
                        "    |        |                    | refer   "
                           + flowProcParam.outsideKey
                           + " -> "
                           + flowProcParam.insideKey
                           + "\n");
                  } // end if
               } // end for
            } // end if
         } // end for
      } // end if
      return sb.toString();
   } // end method toString

   /**
    * Ausgeben des Namens der gerade laufenden Prozedur
    *
    * @return CnC-Procedure-Name
    */
   public String getCurrentProcName(FlowControlTable flowControlTablePa)
   {
      if (flowStack != null && flowStack.size() > 0)
      {
         // int iAdr = ( (FlowStackElement) flowStack.get( flowStack.size() - 1 ) ).adress ;
         int iAdr = ((FlowStackElement) flowStack.peek()).adress;
         FlowControlInstruction element = (FlowControlInstruction) flowControlTablePa.get(iAdr);

         return element.procName;
      } // end if
      return null;
   } // end method getCurrentProcName

   /**
    * Ausgeben des Namens der im letzten ausgefhrten
    * Befehl laufenden Prozedur zur Ausgabe beim Debug
    *
    * @return CnC-Procedure-Name
    */
   public String getPreProcName(FlowControlTable flowControlTablePa)
   {
      if (flowStack != null && flowStack.size() > 0)
      {
         // int iAdr = ( (FlowStackElement) flowStack.get( flowStack.size() - 1 ) ).adress ;
         int iAdr = ((FlowStackElement) flowStack.peek()).preAdress;
         FlowControlInstruction element = (FlowControlInstruction) flowControlTablePa.get(iAdr);

         return element.procName;
      } // end if
      return null;
   } // end method getCurrentProcName

   /**
    * Sichern Backup fr lokale Variable
    *
    * @param request zum Zugriff auf Session zum Backup fr lokale Variable
    */
   public void addLocalVar(
      HttpServletRequest request,
      String pVarNameWithoutPrefix,
      SourcePosition pSourcePosition)
   {
      Log.debug("FlowStack#addLocalVar" , pVarNameWithoutPrefix);
      if (flowStack != null && flowStack.size() > 0)
      {
         FlowStackElement flowStackElement = (FlowStackElement) flowStack.peek();

         flowStackElement.addLocalVar(request, pVarNameWithoutPrefix, pSourcePosition);
         Log.debug("flowStackElement" , flowStackElement);
      } // end if
   } // end method addLocalVar

   /**
    * Vermerken Key-Mapping: Key-Aussen -> Key-Innen fr Referenz-Parameter
    *
    */
   public void addParam(//-
   HttpServletRequest request, //-
   FlowProcCallParam pFlowProcParam) //-
   //SourcePosition pSourcePosition)
   {
      Log.debug("FlowStack#addParam " , pFlowProcParam);
      if (flowStack != null && flowStack.size() > 0)
      {
         FlowStackElement flowStackElement = (FlowStackElement) flowStack.peek();

         flowStackElement.addParam(request, pFlowProcParam /*, pSourcePosition*/
         );
         Log.debug("flowStackElement: " , flowStackElement);
      } // end if
   } // end method addRefParm

   /**
    * Vermerken Key-Aussen fr Return-Value
    */
   public void addReturnOutsideKey(String pOutsideKey)
   {
      Log.debug("FlowStack#addReturnOutsideKey " , pOutsideKey);
      if (pOutsideKey == null)
      {
         return;
      }
      if (flowStack != null && flowStack.size() > 0)
      {
         FlowStackElement flowStackElement = (FlowStackElement) flowStack.peek();

         flowStackElement.callReturnSessionKey = pOutsideKey;
         Log.debug("flowStackElement" , flowStackElement);
      } // end if
   }

   /**
    * @return
    */
   public String getReturnOutsideKey()
   {
      if (flowStack != null && flowStack.size() > 0)
      {
         FlowStackElement flowStackElement = (FlowStackElement) flowStack.peek();

         //Log.debug("" + flowStackElement);

         return flowStackElement.callReturnSessionKey;
      } // end if
      return null;
   } // end method addReturnOutsideKey

   //  /**
   //   * Vermerken Key-Mapping: Key-Aussen -> Key-Innen fr Return-Value
   //   *
   //   */
   //  public void addReturnInsideKey(String strPaInsideKey) {
   //    Log.debug("FlowStack#addReturnInsideKey " + strPaInsideKey);
   //    if (flowStack != null && flowStack.size() > 0) {
   //      FlowStackElement flowStackElement = (FlowStackElement) flowStack.peek();
   //
   //      flowStackElement.returnSessionKey = strPaInsideKey;
   //      Log.debug("" + flowStackElement);
   //    }// end if
   //  }// end method addReturnInsideKey

} // end class FlowStack
