package de.cnc.domainparser;


import java.util.*;

import de.cnc.domainparser.util.*;
import de.cnc.util.StringUtil;


/** 
 * Diese Klasse reprsentiert den Bereich zwischen
 * ffnender und schliessender geschwungener Klammer { ... } .
 */
public class BracesArea extends AbstractNonTerminalElement
{
    /** 
     * Liste der in diesem BracesArea enthaltenen Segmente
     */
    private ArrayList segmentList = new ArrayList();
    
    /** 
     * Referenz auf den kompletten Quelltext der Unit (Datei)
     */
    private final String completeSrcStr;

    /** 
     * Konstruktor, nicht ffentlich
     * 
     * @param pSrcPos Anfangsposition dieses BracesArea
     * @param pSrcStr kompletter Quelltext der Unit (Datei)
     */
    private BracesArea(SourcePosition pSrcPos, String pSrcStr)
    {
        // Anfangsposition vermerken
        srcPos = pSrcPos.getClone();

        completeSrcStr = pSrcStr;
    }

    /** 
     * Einlesen Quelltext und Parsen der im BracesArea enthaltenen Segmente
     * 
     * @exception DomainParseException 
     * @param pStr kompletter Quelltext der Unit (Datei)
     * @param pSrcPos Anfangsposition dieses BracesArea
     * @param pParserContext Einstellungen des Parsers
     * @return geparste Struktur
     */
    public static BracesArea parse(
            final String pStr, 
            final SourcePosition pSrcPos, // Achtung Seiteneffekt
            final DomainParserContext pParserContext)
        throws DomainParseException
    {
        if (pStr.charAt(pSrcPos.getPos()) != '{')
        {
            return null; // oder exception werfen
        }

        BracesArea brcsArea = new BracesArea(pSrcPos, pStr);
        
        // Vermerken der Anfangsposition
        int anfPos = pSrcPos.getPos();        

        pSrcPos.addOne(); // ffnende Klammer berspringen

        // nach der ffnenden Klammer folgende WhiteSpaces berspringen
        DomainParserUtil.parseWhitesAndComments(pStr, pSrcPos);

        while (pSrcPos.getPos() < pStr.length() && pStr.charAt(pSrcPos.getPos()) != '}')
        {
            // Segment parsen
            Segment segment = Segment.parse(pStr, pSrcPos, pParserContext);

            // System.out.println( "Unit#segment " + segment );

            // Segment vermerken
            brcsArea.segmentList.add(segment);

            // nach dem Segment folgende WhiteSpaces berspringen
            DomainParserUtil.parseWhitesAndComments(pStr, pSrcPos);
        }

        if (pSrcPos.getPos() >= pStr.length())
        {
            // keine schliessende Klammer aufgetreten
            throw new DomainParseException("right brace expected '}'", pSrcPos);
        }

        brcsArea.endSourcePosition = pSrcPos.getClone();

        pSrcPos.addOne(); // schliessende Klammer berspringen

        // Lnge vermerken
        brcsArea.length = pSrcPos.getPos() - anfPos;

        return brcsArea;
    }

    public int elementCount()
    {
        return segmentList.size();
    }

    /** 
     * 
     * @param pI 
     */
    public Segment getSegment(int pI)
    {
        return (Segment) segmentList.get(pI);
    }

    public String toString()
    {
        return toString(0);
    }

    /** 
     * 
     * @param indent 
     */
    public String toString(int indent)
    {
        StringBuffer strBuff = toStringHelp(indent);

        for (int i = 0; i < segmentList.size(); i++)
        {
            strBuff.append(((Segment) segmentList.get(i)).toString(indent + 1) + "\n");
        }
        if (segmentList.size() < 1)
        {
            strBuff.append(StringUtil.spaces(indent + 1) + "parenthesis area is empty" + "\n");
        }
        return strBuff.toString();
    }

    /** 
     * @see de.cnc.domainparser.Element#getContent()
     */
    public String getContent()
    {
        // throw new UnsupportedOperationException();
        return completeSrcStr.substring(srcPos.getPos(), srcPos.getPos() + length + 1 /* 2 */ );
    }

    /** "@see de.cnc.domainparser.Element#getNettoContent()
     */
    public String getNettoContent()
    {
        // throw new UnsupportedOperationException();
        return completeSrcStr.substring(srcPos.getPos() + 1, srcPos.getPos() + length + 1);
    }

    /** "@see de.cnc.domainparser.Element#isMultiLined()
     */
    public boolean isMultiLined()
    {
        return getContent().indexOf("\n") >= 0;
    }

    /** "@see de.cnc.domainparser.Element#isEmpty()
     */
    public boolean isEmpty()
    {
        return segmentList == null || segmentList.isEmpty();
    }

    /** 
     * Lschen Kommentare
     */
    public void stripComments(boolean pAllComments)
    {
        for (Iterator iter = segmentList.iterator(); iter.hasNext();)
        {
            Object o = iter.next();

            if (o instanceof InlineComment)
            {
                iter.remove();
            }
            else if (o instanceof BracketComment 
                && ( pAllComments || ! ( (BracketComment) o ).isJavadoc() ) )
            {
                iter.remove();
            }
            else if (!((AbstractElement) o).isTerminal())
            {
                ((AbstractElement) o).stripComments( pAllComments );
            }
        }
    }

    /**
     * Lschen leere Sub-Segmente
     */
    public void stripEmptySegments()
    {
        for (Iterator iter = segmentList.iterator(); iter.hasNext();)
        {
            Segment segment = (Segment) iter.next();
            if (segment.isEmpty())
            {
                iter.remove();
            }
            else
            {
                segment.stripEmptySegments();
            }
        }
    }

}
