package de.cnc.flowcompiler;

import java.util.ArrayList;

import de.cnc.*;
import de.cnc.domainparser.*;
import de.cnc.domainparser.util.SourcePosition;
import de.cnc.expression.Expression;
import de.cnc.expression.exceptions.ExpressionParseException;
import de.cnc.flowprogram.*;
import de.cnc.util.StringUtil;

/**
 * Compiler fr die Flow-Control-Datei flow.cnc . <br/>
 * Die Ablaufsteuerdatei flow.cnc wird mit der Library de.cnc.domainparser<br/>
 * eingeparst und auf Plausibilitt geprft.<br/>
 * Anhand dieser Datei wird die Steuertabelle {@link FlowControlTable} aufgebaut.
 *
 *
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner Kcker</a>
 * <br/>
 *
 * @author Heiner Kcker
 * @version $Id:$
 */
public class FlowControlCompiler
{
   /**
    * Die in der gesamten Web-Applikation nur einmal existierende
    * Tabelle mit den Low-Level (durch parsen aufgelsten) Kommandos.
    * Praktisch das Assembler-Programm des {@link FlowControl} .
    */
   private static FlowControlTable flowControlTable;

   /**
    * {@link FlowSrcFileLine}-Array mit den Zeilen der Steuerdatei <b>flow.cnc</b>
    * und den Prozedur-Namen. <br/>
    */
   private static FlowSrcFileLine[] srcFileLineArr;

   /**
    * Festlegung Name der Main-Procedur
    */
   public static final String MAIN_PROC_NAME = "main";

   /**
    * Name der gerade geparsten Procedure
    */
   private static String procOrFuncName;

    /**
     * ob die gerade geparste Procedure oder Function eine Function ist
     */
    private static boolean isFunction = false;

   /**
    * Liste mit Namen von Prozeduren und Labels mit Kontrolle auf doppelte Verwendung eines Namens
    */
   private static NameList nameList;

   /**
    * Fehlermeldungs-String fr allgemein verwendete Tags
    */
   // hier evtl. noch vervollstndigen
   public static final String genericExpectingStr =
      "Expecting if, while, do, call, return, goto, label, reset, invalidate";

   /**
    * Merker Zeilen- und Spalten-Nummer im XML-File
    */
   // private static int iClXmlLineNumber;

   /**
    * Merker Spalten-Nummer im XML-File
    */
   // private static int iClXmlColumnNumber;

   /**
    * Schachtelungstiefe do- und while-Schleifen . <br/>
    * Wird in {@link #resolveLabels} verwendet . <br/>
    */
   private static int iLoopLevel = 0;

   /**
    * Konstruktor private, instanziieren verboten
    */
   private FlowControlCompiler()
   {
      ;
   }

   /**
    * Parsen der als Parameter bergebenen Datei und zurckliefern
    * der erzeugten FlowControlTable
    */
   public static FlowControlTable parseFile(String fileNameWithPath /*,String pEncoding*/
   ) throws DomainParseException
   {
      // try {
      // parser = new XmlParser(new FileReader(fileNameWithPath));

      //Unit tUnit = Unit.parse( StrConst.FLOW , new DomainParserContext());
      Unit tUnit = Unit.parse(StringUtil.file2str(fileNameWithPath), new DomainParserContext());

      // System.out.println("vor stripComments" + tUnit);

      tUnit.stripComments(true);
      tUnit.stripEmptySegments();

      // Log.debug("nach stripComments" , tUnit);

      // } catch (Exception e) {
      // e.printStackTrace();
      // // throw new RuntimeException( "" + e ) ;
      // throw new ParserException("" + e);
      // }

      // Erzeugen String-Array mit Zeilen aus flow.cnc
      initSrcFileStrArr(fileNameWithPath /*,pEncoding*/
      );

      flowControlTable = new FlowControlTable();
      nameList = new NameList();

      parseFlow(tUnit); // Wurzel-Element

      nameList = null; // Speicherplatz freimachen

      return flowControlTable;
   } // end method parseFile

   /**
    * Reklamieren Fehler in der Flusskontroll-XML-Datei mit Position als Parameter
    *
    * @param iPaLine Zeile  im XML-File an welcher der Fehler aufgetreten ist
    * @param iPaCol  Spalte im XML-File an welcher der Fehler aufgetreten ist
    * @throws RuntimeException mit Error-Description
    */
   // public static void complain(String text, int iPaLine) {
   // // throw new RuntimeException( text +
   // throw new ParserException(text + " Line: " + iPaLine);
   // // " Col: " + iPaCol ) ;
   // }// end method complain

   /**
    * Parser fr das Wurzel-Element &lt;flow&gt; &lt;/flow&gt; . <br/>
    * Innerhalb des Wurzel-Tags flow ist nur das Tag &lt;procedure&gt; erlaubt.
    */
   protected static void parseFlow(Unit pUnit) throws DomainParseException
   {
      // throws IOException
      // try {
      // readStartTag("flow");
      // while (true) {
      // if (nextType == Xml.START_TAG) {
      // if (nextStartTag.getName().equals("procedure")) {
      // parseProcedure();
      // } else {
      // complain("Expecting <procedure name=\"...\">");
      // }
      // } else {
      // break;
      // }// end if
      // }// end while
      // readEndTag("flow");

      for (int i = 0; i < pUnit.segmentCount(); i++)
      {
         Segment segment = pUnit.getSegment(i);

         if (segment.isBracesSegment())
         {
            parseProcedureOrFunction(segment);
         }
         else if (!segment.isEmpty())
         {
            FlowControlUtil.complain(
               "Expecting procedure/function <name> ( <params> ) { <instructions> }",
               segment.getSourcePos());
         }
      }

      // auflsen calls und gotos
      resolveLabels();
      // } catch (Exception e) {
      // e.printStackTrace();
      // // throw new RuntimeException( "" + e ) ;
      // throw new ParserException("" + e);
      // }
   } // end method parseFlowControl

   /**
    * Parser fr das Tag &lt;procedure&gt;. <br/>
    * Das Tag &lt;procedure&gt; darf nur im Tag &lt;flow&gt; auftauchen.
    */
   protected static void parseProcedureOrFunction(final Segment pSegment) throws DomainParseException
   {
      // AttributeReader attrReader = readStartTag("procedure");
      final AbstractElement prcOrFncElem = pSegment.getElement(0);

      if ("procedure".equals(prcOrFncElem.getContent()))
      {
         isFunction = false;
      }
      else if ("function".equals(prcOrFncElem.getContent()))
      {
          isFunction = true;
      }
      else
      {
          FlowControlUtil.complain("procedure or function expected", prcOrFncElem.getSourcePos());
      }

      final Identifier nameElem = FlowControlCompilerUtil.checkForIdentifier(pSegment.getElement(1));

      // Setzen globalen Merker fr gerade geparste Procedure
      procOrFuncName = nameElem.getContent();

      if (StringUtil.contents(procOrFuncName, '.'))
      {
         FlowControlUtil.complain("no point in procedure or function name available", nameElem.getSourcePos());
      }

      // Vermerken xml-Zeile der Prozedur-Anfangs-Zeile
      final SourcePosition sourcePosition = prcOrFncElem.getSourcePos();

      // Vermerken des Beginns der Procedur, damit die calls spter aufgelst werden knnen
      final FlowControlInstruction anfangsElement = new FlowControlInstruction();

      anfangsElement.procName = procOrFuncName;
      anfangsElement.procOrLabelNameToCall = procOrFuncName;
      anfangsElement.sourcePosition = (sourcePosition);

      if (isFunction)
      {
          anfangsElement.controlCode = ControlCode.FUNC_START;
      }
      else
      {
          anfangsElement.controlCode = ControlCode.PROC_START;
      }

      nameList.add(procOrFuncName); // Kontrolle auf doppelte Verwendung des Namens

      final ParenthesisArea paramElem = FlowControlCompilerUtil.checkForParenthesisArea(pSegment.getElement(2));

      // NameList nameList = readProcParams(anfangsElement, attrReader); // Einlesen Parameter
      final NameList paramAndLocalVarsNameList = readProcParams(anfangsElement, paramElem.getCommaSeparatedList());
      // Einlesen Parameter

      // attrReader.checkAttributesConsumed(); // Prfen, ob alle Attribute verbraucht wurden

      anfangsElement.procName = procOrFuncName;
      anfangsElement.debugInfo = ("start " + ( isFunction ? "function" : "procedure" ) + " " + procOrFuncName);

      flowControlTable.add(anfangsElement);

      // Einlesen <localvar sessionKey="..."/>
      // while (nextType == Xml.START_TAG && "localvar".equals(nextStartTag.getName())) {
      //lokale Variable noch fertig machen
      //        while
      //         parseLocalvar(nameList);
      //        }// end while

      // while (nextGenericTag(true)) {
      // ; // nextGenericTag( true ) erledigt alles
      // }// end while

      //@todo: extra methode readProcStartGenericTags statt optionalem Parameter
      //@todo: umbennen in parseCommands oder -Deklaritions statt -Tags
      readGenericTags(pSegment, 3, paramAndLocalVarsNameList);

      // readEndTag("procedure");

      // return am Ende der Procedure
      //flowControlTable.add(
      //   new FlowControlInstruction(pSegment.getEndSourcePos(), ControlCode.PROC_RETURN, 0, "", null, "end procedure", procOrFuncName));
       // Vermerken des Endes der Procedur, damit die calls spter aufgelst werden knnen
       final FlowControlInstruction endElement = new FlowControlInstruction();

       endElement.procName = procOrFuncName;
       endElement.procOrLabelNameToCall = procOrFuncName;
       endElement.sourcePosition = (sourcePosition);
       anfangsElement.debugInfo = ("end " + ( isFunction ? "function" : "procedure" ) + " " + procOrFuncName);

       if (isFunction)
       {
           endElement.controlCode = ControlCode.FUNC_RETURN;
       }
       else
       {
           endElement.controlCode = ControlCode.PROC_RETURN;
       }

       flowControlTable.add(endElement);

      // zuordnen procedure-name zu sourcecode-zeilen
      for (int i = prcOrFncElem.getSourcePos().getRow() - 1;
         i < srcFileLineArr.length && i < pSegment.getEndSourcePos().getRow();
         i++)
      {
         srcFileLineArr[i].procName = procOrFuncName;
      } // end for

      procOrFuncName = null; // rcksetzen globalen Merker procedure-name
   } // end method parseProcedure

   /**
    * rekursives Einlesen aller erlaubten Elemente
    * in einem BracesArea.
    *
    * @param pSegment
    * @param pBracesPos
    * @throws DomainParseException
    */
   private static void readGenericTags(
      Segment pSegment,
      int pBracesPos,
      final NameList pParamAndLocalVarsNameList) // Liste mit Parametern und lokalen Variablen
   throws DomainParseException
   {
      final BracesArea bodyElem = FlowControlCompilerUtil.checkForBracesArea(pSegment.getElement(pBracesPos));

      //List mit den Referenzen auf eventuell vorherige 'if' oder 'else if'
      //je Segment lokal (rekursiv)
      ArrayList ifEndElementList = new ArrayList();

      //Durch den Schalter lastSegmentWasIf erfhrt die Methode, ob ein 'else' oder 'else if' erlaubt ist
      //Sie gibt 'true' bei Auftreten eines 'if' oder 'else if' zurck
      boolean lastSegmentWasIfOrElseIf = false;

      for (int i = 0; i < bodyElem.elementCount(); i++)
      {
         lastSegmentWasIfOrElseIf =
            readOneGenericTag(
               bodyElem.getSegment(i),
               lastSegmentWasIfOrElseIf,
               ifEndElementList,
               pParamAndLocalVarsNameList);
         //            if ( ! lastSegmentWasIfOrElseIf )
         //            {
         //                //wenn kein 'if' oder 'else if' auftauchte, Referenzliste leeren
         //                ifEndElementList.clear();
         //            }
      }
   }

   /**
    * Parsen der allgemein erlaubten Tags if, while, call ...noch fertig machen
    *
    * @param pIsIfOrElseIfAlowed ob ein 'else if' oder ein 'else' erlaubt ist.
    *                            Muss 'true' sein, wenn das vorherige Element
    *                            ein 'if' oder 'else if' war, sonst 'false'.
    * @return ob ein 'if' oder 'else if' auftrat
    */
   protected static boolean readOneGenericTag(
      final Segment pSegment,
      final boolean pIsIfOrElseIfAlowed,
      final ArrayList pIfEndElementList,
   //Referenzen auf vorherige 'if' und 'else if'
   final NameList pParamAndLocalVarsNameList) // Liste mit Parametern und lokalen Variablen
   throws DomainParseException
   {
      if (pSegment.isEmpty())
      {
         return false;
      }

      // if (nextType == Xml.START_TAG) {
      // String name = nextStartTag.getName();
      Identifier identifier = FlowControlCompilerUtil.checkForIdentifier(pSegment.getElement(0));
      String name = identifier.getContent();

      if (name.equals("else"))
      {
         if ( ! pIsIfOrElseIfAlowed)
         {
            FlowControlUtil.complain("else without if", pSegment.getElement(1).getSourcePos());
         }
         //nach 'else' ist nur der Identifier 'if' oder ein BracesArea erlaubt
         if (pSegment.getElement(1) instanceof Identifier)
         {
            parseElseIf(pSegment, pIfEndElementList);
            return true; // 'if' oder 'else if' aufgetreten
         }
         else
         {
            parseElse(pSegment, pIfEndElementList);
            return false;
         }
      }
      else if (name.equals("logon"))
      {
         parseLogon(pSegment, pParamAndLocalVarsNameList);
      }
      else if (name.equals("local"))
      {
         parseLocalvar(pSegment, pParamAndLocalVarsNameList);
      }
      else if (name.equals("if"))
      {
         parseIf(pSegment, pIfEndElementList);
         return true;
      }
      else if (name.equals("while"))
      {
         parseWhile(pSegment);
      }
      else if (name.equals("do"))
      {
         parseDo(pSegment);
      }
      else if (name.equals("action"))
      {
         parseAction(pSegment);
      }
      else if (name.equals("jsp"))
      {
         parseJsp(pSegment);
      }
      else if (name.equals("call"))
      {
         parseCall(pSegment);
      }
      else if (name.equals("return"))
      {
         parseReturn(pSegment);
      }
      else if (name.equals("label"))
      {
         parseLabel(pSegment);
      }
      else if (name.equals("goto"))
      {
         parseGoto(pSegment);
         // } else if (name.equals("set")) {
         // parseSet( pSegment );
      }
      else if (name.equals("reset"))
      {
         parseReset(pSegment);
         // } else if (name.equals("inc")) {
         // parseInc( pSegment );
         // } else if (name.equals("dec")) {
         // parseDec( pSegment );
      }
      else if (name.equals("continue"))
      {
         parseContinue(pSegment);
      }
      else if (name.equals("break"))
      {
         parseBreak(pSegment);
      }
      else if (name.equals("invalidate"))
      {
         parseInvalidate(pSegment);
      }
      // sesssion.variable := call function( ... );
      else if (
         pSegment.elementCount() == 5 // --
            && pSegment.getElement(0) instanceof Identifier // --
            && pSegment.getElement(1).getContent().equals(":=") // --
            && pSegment.getElement(2).getContent().equals("call") // --
            && pSegment.getElement(3) instanceof Identifier // --
            && pSegment.getElement(4) instanceof ParenthesisArea) // --
      {
         // <variable> := call <function> ( <parameterlist> )
         parseFunctionCall(pSegment);
      }
      // sesssion.variable := action( ... );
      else if (
         pSegment.elementCount() == 4 // --
            && pSegment.getElement(0) instanceof Identifier // --
            && pSegment.getElement(1).getContent().equals(":=") // --
            && pSegment.getElement(2).getContent().equals("action") // --
            && pSegment.getElement(3) instanceof ParenthesisArea) // --
      {
         // <variable> := call <function> ( <parameterlist> )
         parseActionAssignment(pSegment);
      }
      else
      {
         //complain(genericExpectingStr, pSegment.getSourcePos());

         //wenn nichts anderes passt, wird versucht eine Expression zu erkennen
         //in Expressions sind auch Zuweisungen erlaubt
         parseExpression(pSegment);
      } // end if

      return false;
   } // end method readGenericTag

   /**
    * Parsen &lt;if&gt;-Tag. <br/>
    * Im if-Tag drfen die Tags elseif und else auftauchen. <br/>
    */
   protected static void parseIf(Segment pSegment, ArrayList pIfEndElementList) throws DomainParseException
   {
      FlowControlCompilerUtil.checkForIdentifierParenthesisBraces(pSegment);

      addNop(pSegment.getSourcePos()); // NOP nur fr Einzelschritt-Debug

      FlowControlInstruction anfangsElement = new FlowControlInstruction();

      anfangsElement.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      anfangsElement.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      anfangsElement.controlCode = ControlCode.JUMP_IF_NOT;
      anfangsElement.debugInfo = ("if anfang");

      anfangsElement.setJumpConditionen(pSegment.getElement(1));

      anfangsElement.iLoopLevel = iLoopLevel;
      flowControlTable.add(anfangsElement);

      // *** if ***

      // allgemeine Tags sind nur solange erlaubt, wie noch kein elseif oder else aufgetaucht ist
      // in while-schleife alle generic-tags einziehen
      // while ( nextType == Xml.START_TAG && nextGenericTag( false ) ) {
      // while (readGenericTag(false)) {
      // ; // nextGenericTag( false ) erledigt alles
      // }// end while
      readGenericTags(pSegment, 2, null);

      // Sprungadresse fr das Anfangs-Element zum Anspringen if not
      // Springen auf else if
      anfangsElement.jumpDestination = (flowControlTable.size() + 1);

      // Endelement, alle Endelemente springen hinter des </if>
      FlowControlInstruction endElement = new FlowControlInstruction();

      endElement.procName = procOrFuncName;
      endElement.sourcePosition = (pSegment.getEndSourcePos());
      endElement.controlCode = ControlCode.JUMP;
      endElement.debugInfo = ("if ende jump");
      endElement.iLoopLevel = iLoopLevel;
      flowControlTable.add(endElement);

      //Referenzliste leeren
      pIfEndElementList.clear();

      pIfEndElementList.add(endElement);

      // *** else if ***
      //        while (nextType == Xml.START_TAG && "elseif".equals(nextStartTag.getName()))
      //        { // elseif ist nur innerhalb if erlaubt
      //
      //            // addNop() ; // NOP nur fr Einzelschritt-Debug
      //
      //            // letztem endElement die aktuelle XmlLineNumber geben
      //            endElement.setSourcePosition(iClXmlLineNumber);
      //
      //            attrReader = readStartTag("elseif");
      //
      //            addNop(); // NOP nur fr Einzelschritt-Debug
      //
      //            // Anfangselement auf jetziges Element setzen
      //            anfangsElement = new FlowControlInstruction();
      //            anfangsElement.procOrFuncName = procOrFuncName;
      //            anfangsElement.setSourcePosition(iClXmlLineNumber);
      //            anfangsElement.setControlCode(ControlCode.JUMP_IF_NOT);
      //            anfangsElement.setDebugInfo("elseif anfang");
      //            anfangsElement.setJumpConditionen(attrReader);
      //            attrReader.checkAttributesConsumed(); // Prfen, ob alle Attribute verbraucht wurden
      //            anfangsElement.iLoopLevel = iLoopLevel;
      //            flowControlTable.add(anfangsElement);
      //
      //            while (nextGenericTag(true))
      //            {
      //                ; // nextGenericTag( true ) erledigt alles
      //            }// end while
      //
      //            readEndTag("elseif");
      //
      //            // Sprungadresse fr das Anfangs-Element zum Anspringen if not
      //            // Springen auf else if
      //            anfangsElement.setJumpDestination(flowControlTable.size() + 1);
      //
      //            // Endelement, alle Endelemente springen hinter des </if>
      //            endElement = new FlowControlInstruction();
      //            endElement.procOrFuncName = procOrFuncName;
      //            endElement.setSourcePosition(iClXmlLineNumber);
      //            endElement.setControlCode(ControlCode.JUMP);
      //            endElement.setDebugInfo("elseif ende jump");
      //            endElement.iLoopLevel = iLoopLevel;
      //            flowControlTable.add(endElement);
      //            endElementList.add(endElement);
      //
      //        }// end while
      //
      //        // *** else ***
      //        if (nextType == Xml.START_TAG && "else".equals(nextStartTag.getName()))
      //        { // else ist nur innerhalb if erlaubt
      //
      //            // letztem endElement die aktuelle XmlLineNumber geben
      //            endElement.setSourcePosition(iClXmlLineNumber);
      //
      //            // addNop() ; // NOP nur fr Einzelschritt-Debug
      //
      //            // Sprungadresse fr das Anfangs-Element zum Anspringen if not
      //            // Springen auf else if
      //            // bei else das letzte Anfangs-Sprungziel um eins erhhen
      //            anfangsElement.setJumpDestination(flowControlTable.size());
      //
      //            readStartTag("else");
      //
      //            addNop(); // NOP nur fr Einzelschritt-Debug
      //
      //            while (nextGenericTag(true))
      //            {
      //                ; // nextGenericTag( true ) erledigt alles
      //            }// end while
      //
      //            readEndTag("else");
      //
      //            addNop(); // NOP nur fr Einzelschritt-Debug
      //
      //        }// end if
      //        readEndTag("if");

      addNop(pSegment.getEndSourcePos()); // NOP nur fr Einzelschritt-Debug

      // Sprung-Adresse fr alle EndElemente setzen
      for (int i = 0; i < pIfEndElementList.size(); i++)
      {
         FlowControlInstruction element = (FlowControlInstruction) pIfEndElementList.get(i);

         element.jumpDestination = (flowControlTable.size());
         // element.setXmlLineNumber( iClXmlLineNumber ) ;
      } // end for

      //am Ende kann gibt es ein unntiges JUMP auf sich selbst, das knnte gelscht werden

   } // end method parseIf

   /**
    * Parsen &lt;if&gt;-Tag. <br/>
    * Im if-Tag drfen die Tags elseif und else auftauchen. <br/>
    */
   protected static void parseElseIf(Segment pSegment, ArrayList pIfEndElementList) throws DomainParseException
   {
      FlowControlCompilerUtil.checkForIdentifierIdentifierParenthesisBraces(pSegment);

      //checken, ob dem 'else' ein 'if' folgt
      Identifier identifier = FlowControlCompilerUtil.checkForIdentifier(pSegment.getElement(1));
      String name = identifier.getContent();
      if (!"if".equals(name))
      {
         FlowControlUtil.complain("if or bracesarea expected", pSegment.getElement(1).getSourcePos());
      }

      addNop(pSegment.getSourcePos()); // NOP nur fr Einzelschritt-Debug

      FlowControlInstruction anfangsElement = new FlowControlInstruction();

      anfangsElement.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      anfangsElement.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      anfangsElement.controlCode = ControlCode.JUMP_IF_NOT;
      anfangsElement.debugInfo = ("else if start");

      anfangsElement.setJumpConditionen(pSegment.getElement(2));

      anfangsElement.iLoopLevel = iLoopLevel;
      flowControlTable.add(anfangsElement);

      // *** if ***

      // allgemeine Tags sind nur solange erlaubt, wie noch kein elseif oder else aufgetaucht ist
      // in while-schleife alle generic-tags einziehen
      // while ( nextType == Xml.START_TAG && nextGenericTag( false ) ) {
      // while (readGenericTag(false)) {
      // ; // nextGenericTag( false ) erledigt alles
      // }// end while
      readGenericTags(pSegment, 3, null);

      // Sprungadresse fr das Anfangs-Element zum Anspringen if not
      // Springen auf else if
      anfangsElement.jumpDestination = (flowControlTable.size() + 1);

      // Endelement, alle Endelemente springen hinter des </if>
      FlowControlInstruction endElement = new FlowControlInstruction();

      endElement.procName = procOrFuncName;
      endElement.sourcePosition = (pSegment.getEndSourcePos());
      endElement.controlCode = ControlCode.JUMP;
      endElement.debugInfo = ("else if end jump");
      endElement.iLoopLevel = iLoopLevel;
      flowControlTable.add(endElement);
      pIfEndElementList.add(endElement);

      // *** else if ***
      //        while (nextType == Xml.START_TAG && "elseif".equals(nextStartTag.getName()))
      //        { // elseif ist nur innerhalb if erlaubt
      //
      //            // addNop() ; // NOP nur fr Einzelschritt-Debug
      //
      //            // letztem endElement die aktuelle XmlLineNumber geben
      //            endElement.setSourcePosition(iClXmlLineNumber);
      //
      //            attrReader = readStartTag("elseif");
      //
      //            addNop(); // NOP nur fr Einzelschritt-Debug
      //
      //            // Anfangselement auf jetziges Element setzen
      //            anfangsElement = new FlowControlInstruction();
      //            anfangsElement.procOrFuncName = procOrFuncName;
      //            anfangsElement.setSourcePosition(iClXmlLineNumber);
      //            anfangsElement.setControlCode(ControlCode.JUMP_IF_NOT);
      //            anfangsElement.setDebugInfo("elseif anfang");
      //            anfangsElement.setJumpConditionen(attrReader);
      //            attrReader.checkAttributesConsumed(); // Prfen, ob alle Attribute verbraucht wurden
      //            anfangsElement.iLoopLevel = iLoopLevel;
      //            flowControlTable.add(anfangsElement);
      //
      //            while (nextGenericTag(true))
      //            {
      //                ; // nextGenericTag( true ) erledigt alles
      //            }// end while
      //
      //            readEndTag("elseif");
      //
      //            // Sprungadresse fr das Anfangs-Element zum Anspringen if not
      //            // Springen auf else if
      //            anfangsElement.setJumpDestination(flowControlTable.size() + 1);
      //
      //            // Endelement, alle Endelemente springen hinter des </if>
      //            endElement = new FlowControlInstruction();
      //            endElement.procOrFuncName = procOrFuncName;
      //            endElement.setSourcePosition(iClXmlLineNumber);
      //            endElement.setControlCode(ControlCode.JUMP);
      //            endElement.setDebugInfo("elseif ende jump");
      //            endElement.iLoopLevel = iLoopLevel;
      //            flowControlTable.add(endElement);
      //            endElementList.add(endElement);
      //
      //        }// end while
      //
      //        // *** else ***
      //        if (nextType == Xml.START_TAG && "else".equals(nextStartTag.getName()))
      //        { // else ist nur innerhalb if erlaubt
      //
      //            // letztem endElement die aktuelle XmlLineNumber geben
      //            endElement.setSourcePosition(iClXmlLineNumber);
      //
      //            // addNop() ; // NOP nur fr Einzelschritt-Debug
      //
      //            // Sprungadresse fr das Anfangs-Element zum Anspringen if not
      //            // Springen auf else if
      //            // bei else das letzte Anfangs-Sprungziel um eins erhhen
      //            anfangsElement.setJumpDestination(flowControlTable.size());
      //
      //            readStartTag("else");
      //
      //            addNop(); // NOP nur fr Einzelschritt-Debug
      //
      //            while (nextGenericTag(true))
      //            {
      //                ; // nextGenericTag( true ) erledigt alles
      //            }// end while
      //
      //            readEndTag("else");
      //
      //            addNop(); // NOP nur fr Einzelschritt-Debug
      //
      //        }// end if
      //        readEndTag("if");

      addNop(pSegment.getEndSourcePos()); // NOP nur fr Einzelschritt-Debug

      // Sprung-Adresse fr alle EndElemente setzen
      for (int i = 0; i < pIfEndElementList.size(); i++)
      {
         FlowControlInstruction element = (FlowControlInstruction) pIfEndElementList.get(i);

         element.jumpDestination = (flowControlTable.size());
         // element.setXmlLineNumber( iClXmlLineNumber ) ;
      } // end for

      //am Ende kann gibt es ein unntiges JUMP auf sich selbst, das knnte gelscht werden

   } // end method parseIf

   /**
    * Parsen &lt;if&gt;-Tag. <br/>
    * Im if-Tag drfen die Tags elseif und else auftauchen. <br/>
    */
   protected static void parseElse(Segment pSegment, ArrayList pIfEndElementList) throws DomainParseException
   {
      FlowControlCompilerUtil.checkForIdentifierBraces(pSegment);

      addNop(pSegment.getSourcePos()); // NOP nur fr Einzelschritt-Debug

      //        FlowControlInstruction anfangsElement = new FlowControlInstruction();
      //
      //        anfangsElement.procOrFuncName = procOrFuncName;
      //        anfangsElement.setSourcePosition(pSegment.getSourcePos());
      //        anfangsElement.setControlCode(ControlCode.JUMP_IF_NOT);
      //        anfangsElement.setDebugInfo("else if start");
      //
      //        anfangsElement.setJumpConditionen(pSegment.getElement(1));
      //
      //        anfangsElement.iLoopLevel = iLoopLevel;
      //        flowControlTable.add(anfangsElement);

      // *** if ***

      // allgemeine Tags sind nur solange erlaubt, wie noch kein elseif oder else aufgetaucht ist
      // in while-schleife alle generic-tags einziehen
      // while ( nextType == Xml.START_TAG && nextGenericTag( false ) ) {
      // while (readGenericTag(false)) {
      // ; // nextGenericTag( false ) erledigt alles
      // }// end while
      readGenericTags(pSegment, 1, null);

      // Sprungadresse fr das Anfangs-Element zum Anspringen if not
      // Springen auf else if
      //        anfangsElement.setJumpDestination(flowControlTable.size() + 1);

      // Endelement, alle Endelemente springen hinter des </if>
      //        FlowControlInstruction endElement = new FlowControlInstruction();
      //
      //        endElement.procOrFuncName = procOrFuncName;
      //        endElement.setSourcePosition(pSegment.getEndSourcePos());
      //        endElement.setControlCode(ControlCode.JUMP);
      //        endElement.setDebugInfo("else end");
      //        endElement.iLoopLevel = iLoopLevel;
      //        flowControlTable.add(endElement);
      //        pIfEndElementList.add(endElement);

      // *** else if ***
      //        while (nextType == Xml.START_TAG && "elseif".equals(nextStartTag.getName()))
      //        { // elseif ist nur innerhalb if erlaubt
      //
      //            // addNop() ; // NOP nur fr Einzelschritt-Debug
      //
      //            // letztem endElement die aktuelle XmlLineNumber geben
      //            endElement.setSourcePosition(iClXmlLineNumber);
      //
      //            attrReader = readStartTag("elseif");
      //
      //            addNop(); // NOP nur fr Einzelschritt-Debug
      //
      //            // Anfangselement auf jetziges Element setzen
      //            anfangsElement = new FlowControlInstruction();
      //            anfangsElement.procOrFuncName = procOrFuncName;
      //            anfangsElement.setSourcePosition(iClXmlLineNumber);
      //            anfangsElement.setControlCode(ControlCode.JUMP_IF_NOT);
      //            anfangsElement.setDebugInfo("elseif anfang");
      //            anfangsElement.setJumpConditionen(attrReader);
      //            attrReader.checkAttributesConsumed(); // Prfen, ob alle Attribute verbraucht wurden
      //            anfangsElement.iLoopLevel = iLoopLevel;
      //            flowControlTable.add(anfangsElement);
      //
      //            while (nextGenericTag(true))
      //            {
      //                ; // nextGenericTag( true ) erledigt alles
      //            }// end while
      //
      //            readEndTag("elseif");
      //
      //            // Sprungadresse fr das Anfangs-Element zum Anspringen if not
      //            // Springen auf else if
      //            anfangsElement.setJumpDestination(flowControlTable.size() + 1);
      //
      //            // Endelement, alle Endelemente springen hinter des </if>
      //            endElement = new FlowControlInstruction();
      //            endElement.procOrFuncName = procOrFuncName;
      //            endElement.setSourcePosition(iClXmlLineNumber);
      //            endElement.setControlCode(ControlCode.JUMP);
      //            endElement.setDebugInfo("elseif ende jump");
      //            endElement.iLoopLevel = iLoopLevel;
      //            flowControlTable.add(endElement);
      //            endElementList.add(endElement);
      //
      //        }// end while
      //
      //        // *** else ***
      //        if (nextType == Xml.START_TAG && "else".equals(nextStartTag.getName()))
      //        { // else ist nur innerhalb if erlaubt
      //
      //            // letztem endElement die aktuelle XmlLineNumber geben
      //            endElement.setSourcePosition(iClXmlLineNumber);
      //
      //            // addNop() ; // NOP nur fr Einzelschritt-Debug
      //
      //            // Sprungadresse fr das Anfangs-Element zum Anspringen if not
      //            // Springen auf else if
      //            // bei else das letzte Anfangs-Sprungziel um eins erhhen
      //            anfangsElement.setJumpDestination(flowControlTable.size());
      //
      //            readStartTag("else");
      //
      //            addNop(); // NOP nur fr Einzelschritt-Debug
      //
      //            while (nextGenericTag(true))
      //            {
      //                ; // nextGenericTag( true ) erledigt alles
      //            }// end while
      //
      //            readEndTag("else");
      //
      //            addNop(); // NOP nur fr Einzelschritt-Debug
      //
      //        }// end if
      //        readEndTag("if");

      addNop(pSegment.getEndSourcePos()); // NOP nur fr Einzelschritt-Debug

      // Sprung-Adresse fr alle EndElemente setzen
      for (int i = 0; i < pIfEndElementList.size(); i++)
      {
         FlowControlInstruction element = (FlowControlInstruction) pIfEndElementList.get(i);

         element.jumpDestination = (flowControlTable.size());
         // element.setXmlLineNumber( iClXmlLineNumber ) ;
      } // end for

      //am Ende kann gibt es ein unntiges JUMP auf sich selbst, das knnte gelscht werden

   } // end method parseIf

   /**
    * Parsen des &lt;while&gt;-Tags. <br/>
    */
   protected static void parseWhile(Segment pSegment) throws DomainParseException
   {
      FlowControlCompilerUtil.checkForIdentifierParenthesisBraces(pSegment);

      iLoopLevel++; // Schachtelungstiefe Schleife Erhhen
      FlowControlInstruction anfangsElement = new FlowControlInstruction();

      anfangsElement.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      anfangsElement.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      anfangsElement.controlCode = ControlCode.START_WHILE; // JUMP_IF_NOT ) ;
      anfangsElement.debugInfo = ("while anfang (JUMP_IF_NOT) " + pSegment.getElement(1).getContent());

      anfangsElement.setJumpConditionen(pSegment.getElement(1));

      anfangsElement.iLoopLevel = iLoopLevel;
      flowControlTable.add(anfangsElement);

      int rejumpAdress = flowControlTable.size() - 1;

      //        while (nextGenericTag(true))
      //        {
      //            ; // nextGenericTag( true ) erledigt alles
      //        }// end while

      readGenericTags(pSegment, 2, null);

      FlowControlInstruction endElement = new FlowControlInstruction();

      endElement.procName = procOrFuncName;
      endElement.sourcePosition = (pSegment.getEndSourcePos());
      endElement.controlCode = ControlCode.END_WHILE; // JUMP ) ;
      endElement.jumpDestination = (rejumpAdress);
      endElement.debugInfo = ("while ende (JUMP)");
      endElement.iLoopLevel = iLoopLevel;
      flowControlTable.add(endElement);
      iLoopLevel--; // Schachtelungstiefe Schleife Verringern

      // Sprungadresse fr das Anfangs-Element zum Anspringen if not
      // Springen hinter letzten Befehl
      anfangsElement.jumpDestination = (flowControlTable.size());
   } // end method parseWhile

   /**
    * Parsen des &lt;do&gt;-Tags. <br/>
    * Wie while-Schleife, wird aber mindestens einmal durchlaufen.<br/>
    */
   protected static void parseDo(Segment pSegment) throws DomainParseException
   {
      FlowControlCompilerUtil.checkForIdentifierParenthesisBraces(pSegment);

      iLoopLevel++; // Schachtelungstiefe Schleife Erhhen
      FlowControlInstruction doAnfangsElement = new FlowControlInstruction();

      doAnfangsElement.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      doAnfangsElement.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      doAnfangsElement.controlCode = ControlCode.START_DO;
      doAnfangsElement.debugInfo = ("do Start (NOP)");
      doAnfangsElement.iLoopLevel = iLoopLevel;
      flowControlTable.add(doAnfangsElement);

      int rejumpAdress = flowControlTable.size();

      FlowControlInstruction endElement = new FlowControlInstruction();

      // setzen conditionen usw. vor body
      endElement.procName = procOrFuncName;
      endElement.controlCode = ControlCode.END_DO; // JUMP_IF ) ;
      endElement.jumpDestination = (rejumpAdress);

      endElement.setJumpConditionen(pSegment.getElement(1));

      endElement.debugInfo = ("do ende (JUMP_IF)");

      // aufbau endelement hier unterbrochen und erst mal alles dazwischen einziehen
      //        while (nextGenericTag(true))
      //        {
      //            ; // nextGenericTag( true ) erledigt alles
      //        }// end while
      readGenericTags(pSegment, 2, null);

      //readEndTag("do");

      // setzen debug-zeile hinter body
      endElement.sourcePosition = (pSegment.getEndSourcePos());
      endElement.iLoopLevel = iLoopLevel;
      flowControlTable.add(endElement);
      iLoopLevel--; // Schachtelungstiefe Schleife Verringern
   } // end method parseDo

   /**
    * Parsen empty-Tag &lt;action name="..."/&gt; . <br/>
    */
   protected static void parseAction(Segment pSegment) throws DomainParseException
   {
      FlowControlCompilerUtil.checkForIdentifierParenthesis(pSegment);

      FlowControlInstruction element = new FlowControlInstruction();

      element.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      element.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      element.controlCode = ControlCode.ACTION;

      //element.setJumpConditionen(attrReader);

      //String name = attrReader.getValue("name");

      element.setNameExpression(pSegment.getElement(1)); // muss ein ParenthesisArea sein
      element.debugInfo = ("action " + pSegment.getElement(1).getContent());

      element.iLoopLevel = iLoopLevel;
      flowControlTable.add(element);

   } // end method parseAction

   /**
    * Parsen empty-Tag &lt;jsp name="..."/&gt; . <br/>
    */
   protected static void parseJsp(Segment pSegment) throws DomainParseException
   {
      FlowControlCompilerUtil.checkForIdentifierParenthesis(pSegment);

      FlowControlInstruction element = new FlowControlInstruction();

      element.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      element.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      element.controlCode = ControlCode.JSP;

      element.setNameExpression(pSegment.getElement(1)); // muss ein ParenthesisArea sein
      element.debugInfo = ("jsp " + pSegment.getElement(1).getContent());

      element.iLoopLevel = iLoopLevel;
      flowControlTable.add(element);

      //addNop( pSegment.getEndSourcePos() ); // NOP nur fr HTML-Debug

   } // end method parseJsp

   /**
    * Parsen empty-Tag &lt;call name="..."/&gt; . <br/>
    */
   protected static void parseCall(Segment pSegment)
   {
      FlowControlCompilerUtil.checkForIdentifierIdentifierParenthesis(pSegment);

      FlowControlInstruction element = new FlowControlInstruction();

      element.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      element.sourcePosition = (pSegment.getElement(0).getSourcePos());
      element.controlCode = ControlCode.CALL;

      //String name = attrReader.getValue("name");
      Identifier identifier = FlowControlCompilerUtil.checkForIdentifier(pSegment.getElement(1));
      String name = identifier.getContent();

      element.procOrLabelNameToCall = (name);

      element.debugInfo = ("call " + name);

      readCallParams(element, pSegment.getElement(2)); // Einlesen Parameter

      //addNop( pSegment.getSourcePos() ); // NOP nur fr Einzelschritt-Debug
      element.iLoopLevel = iLoopLevel;
      flowControlTable.add(element);

   } // end method parseCall

   /**
    * Parsen empty-Tag &lt;call name="..."/&gt; . <br/>
    */
   protected static void parseFunctionCall(Segment pSegment)
   {
      //FlowControlCompilerUtil.checkForIdentifierIdentifierParenthesis(pSegment);

      final FlowControlInstruction element = new FlowControlInstruction();

      element.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      element.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      element.controlCode = ControlCode.CALL;

      //String name = attrReader.getValue("name");
      Identifier identifier = FlowControlCompilerUtil.checkForIdentifier(pSegment.getElement(3));
      String name = identifier.getContent();

      element.procOrLabelNameToCall = (name);

      element.debugInfo = ("call " + name);

      //aufruf als function
      //element.returnSessionKey = attrReader.getValueOptional("returnSessionKey");
      identifier = FlowControlCompilerUtil.checkForIdentifier(pSegment.getElement(0));
      name = identifier.getContent();
      element.returnSessionKey = FlowControlUtil.checkAndCutPrefixSession(name, pSegment.getElement(0).getSourcePos());

      readCallParams(element, pSegment.getElement(4)); // Einlesen Parameter

      //addNop(pSegment.getSourcePos()); // NOP nur fr Einzelschritt-Debug
      element.iLoopLevel = iLoopLevel;
      flowControlTable.add(element);

   } // end method parseFunctionCall

   /**
    * Parsen empty-Tag &lt;call name="..."/&gt; . <br/>
    */
   protected static void parseActionAssignment(Segment pSegment) throws DomainParseException
   {
      FlowControlInstruction element = new FlowControlInstruction();

      element.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      element.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      element.controlCode = ControlCode.ACTION;

      //aufruf als function
      Identifier identifier = FlowControlCompilerUtil.checkForIdentifier(pSegment.getElement(0));
      String name = identifier.getContent();
      element.returnSessionKey = FlowControlUtil.checkAndCutPrefixSession(name, pSegment.getElement(0).getSourcePos());

      element.setNameExpression(pSegment.getElement(3)); // muss ein ParenthesisArea sein
      element.debugInfo = ("action " + pSegment.getElement(3).getContent());

      element.iLoopLevel = iLoopLevel;
      flowControlTable.add(element);

   } // end method parseActionAssignment

   /**
    * Parsen empty-Tag &lt;return/&gt; . <br/>
    * &lt;return/&gt; ist nicht in der main-Procedur erlaubt.<br/>
    */
   protected static void parseReturn(Segment pSegment) throws DomainParseException
   {
      if (MAIN_PROC_NAME.equals(procOrFuncName))
      {
         // wenn return in der main-procedur liegt
         FlowControlUtil.complain("return not allows in procedure " + MAIN_PROC_NAME, pSegment.getSourcePos());
      } // end if

      if (pSegment.elementCount() == 1)
      {
         //return;
         FlowControlCompilerUtil.checkForOneIdentifier(pSegment);
         if (isFunction)
         {
             //innnerhalb einer Function ist ein Rckgabewert notwendig
             FlowControlUtil.complain("return value expected", pSegment.getElement( 0 ).getSourcePos());
         }
      }
      else
      {
         //return ( <expression> )
         FlowControlCompilerUtil.checkForIdentifierParenthesis(pSegment);
         if ( ! isFunction)
         {
             //innnerhalb einer Procedure ist kein Rckgabewert erlaubt
             FlowControlUtil.complain("no return value available", pSegment.getElement( 0 ).getSourcePos());
         }
      }

      //AttributeReader attrReader = readStartTag("return");

      addNop(pSegment.getSourcePos()); // NOP nur fr Einzelschritt-Debug

      FlowControlInstruction element = new FlowControlInstruction();

      element.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      element.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());

      if (isFunction)
      {
          element.controlCode = ControlCode.FUNC_RETURN;
      }
      else
      {
          element.controlCode = ControlCode.PROC_RETURN;
      }

      //element.setJumpConditionen(attrReader);

      //element.returnSessionKey = attrReader.getValueOptional("returnSessionKey");
      //return value
      if (pSegment.elementCount() == 2)
      {
         element.setReturnExpression(pSegment.getElement(1)); // muss ein ParenthesisArea sein
      }

      element.debugInfo = ("return");

      element.iLoopLevel = iLoopLevel;
      flowControlTable.add(element);
   }

   /**
    * Parsen empty-Tag &lt;label name="..."/&gt; . <br/>
    * &lt;label/&gt; ist nur in der main-Procedur erlaubt.<br/>
    */
   protected static void parseLabel(Segment pSegment)
   {
      FlowControlCompilerUtil.checkForTwoIdentifier(pSegment);

      if (!MAIN_PROC_NAME.equals(procOrFuncName))
      {
         // wenn Label nicht in der main-procedur liegt
         FlowControlUtil.complain("label only allows in procedure " + MAIN_PROC_NAME, pSegment.getSourcePos());
      } // end if

      //AttributeReader attrReader = readStartTag("label");

      //addNop(pSegment.getSourcePos()); // NOP nur fr Einzelschritt-Debug

      FlowControlInstruction element = new FlowControlInstruction();

      element.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      element.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      element.controlCode = ControlCode.LABEL_FOR_GOTO;

      //String name = attrReader.getValue("name");
      Identifier identifier = FlowControlCompilerUtil.checkForIdentifier(pSegment.getElement(1));
      String name = identifier.getContent();

      nameList.add(name); // Kontrolle auf doppelte Verwendung des Namens

      element.procOrLabelNameToCall = (name);
      element.debugInfo = ("label " + name);

      element.iLoopLevel = iLoopLevel;
      flowControlTable.add(element);

   } // end method parseLabel

   /**
    * Parsen Befehl goto &lt;label&gt; . <br/>
    */
   protected static void parseGoto(Segment pSegment)
   {
      FlowControlCompilerUtil.checkForTwoIdentifier(pSegment);

      //addNop(pSegment.getSourcePos()); // NOP nur fr Einzelschritt-Debug

      FlowControlInstruction element = new FlowControlInstruction();

      element.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      element.sourcePosition = pSegment.getElement( 0 ).getSourcePos();
      element.controlCode = ControlCode.GOTO;

      Identifier identifier = FlowControlCompilerUtil.checkForIdentifier(pSegment.getElement(1));
      String name = identifier.getContent();

      element.procOrLabelNameToCall = (name);
      element.debugInfo = "goto " + name;

      element.iLoopLevel = iLoopLevel;
      flowControlTable.add(element);

   } // end method parseGoto

   //    /**
   //     * Parsen empty-Tag &lt;set sessionKey="..." value="..."/&gt; . <br/>
   //     */
   //    protected static void parseSet()
   //
   //    {
   //        AttributeReader attrReader = readStartTag("set");
   //
   //        FlowControlInstruction element = new FlowControlInstruction();
   //
   //        element.procOrFuncName = procOrFuncName;
   //        element.setSourcePosition(iClXmlLineNumber);
   //        element.setControlCode(ControlCode.SET);
   //
   //        String sessionKey = attrReader.getValue("sessionKey");
   //
   //        element.setSessionKey(sessionKey);
   //
   //        String debugStr = null;
   //
   //        String setClass = attrReader.getValueOptional("class");
   //        String value = attrReader.getValueOptional("value");
   //        String fromSessionKey = attrReader.getValueOptional("fromSessionKey");
   //        String fromRequestKey = attrReader.getValueOptional("fromRequestKey");
   //        String fromAction = attrReader.getValueOptional("fromAction");
   //
   //        if (!StringUtil.empty(setClass))
   //        {
   //            element.setSetClass(setClass);
   //            debugStr = "class=" + setClass;
   //        }
   //        else if (!StringUtil.empty(value))
   //        {
   //            element.setValue(value);
   //            debugStr = "value=" + value;
   //        }
   //        else if (!StringUtil.empty(fromSessionKey))
   //        {
   //            element.setFromSessionKey(fromSessionKey);
   //            debugStr = "fromSessionKey=" + fromSessionKey;
   //        }
   //        else if (!StringUtil.empty(fromRequestKey))
   //        {
   //            element.setFromRequestKey(fromRequestKey);
   //            debugStr = "fromRequestKey=" + fromRequestKey;
   //        }
   //        else if (!StringUtil.empty(fromAction))
   //        {
   //            element.setFromAction(fromAction);
   //            debugStr = "fromAction=" + fromAction;
   //        }
   //        else
   //        {
   //            complain("set: Option class, value, fromSessionKey, fromRequestKey or fromAction expected");
   //        }// end if
   //
   //        element.setDebugInfo("set " + sessionKey + " auf " + debugStr);
   //
   //        attrReader.checkAttributesConsumed(); // Prfen, ob alle Attribute verbraucht wurden
   //
   //        element.iLoopLevel = iLoopLevel;
   //        flowControlTable.add(element);
   //
   //        readEndTag("set");
   //    }// end method parseSet

   /**
    * Parsen empty-Tag &lt;reset sessionKey="..."/&gt; oder &lt;reset/&gt; . <br/>
    */
   protected static void parseReset(Segment pSegment) throws DomainParseException
   {
      //reset 'all' nicht fertig
      //berlegen ob reset identifier alternativ zu reset ( expression ) sinnvoll ist
      FlowControlCompilerUtil.checkForIdentifierParenthesis(pSegment);

      FlowControlInstruction element = new FlowControlInstruction();

      element.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      element.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      element.controlCode = ControlCode.RESET;

      element.setNameExpression(pSegment.getElement(1)); // muss ein ParenthesisArea sein
      element.debugInfo = "reset " + pSegment.getElement(1).getContent();

      element.iLoopLevel = iLoopLevel;
      flowControlTable.add(element);

      //addNop(pSegment.getEndSourcePos()); // NOP nur fr HTML-Debug

   } // end method parseReset

   //    /**
   //     * Parsen empty-Tag &lt;inc sessionKey="..." value="..."/&gt; . <br/>
   //     */
   //    protected static void parseInc()
   //
   //    {
   //        AttributeReader attrReader = readStartTag("inc");
   //
   //        FlowControlInstruction element = new FlowControlInstruction();
   //
   //        element.procOrFuncName = procOrFuncName;
   //        element.setSourcePosition(iClXmlLineNumber);
   //        element.setControlCode(ControlCode.INC);
   //        String sessionKey = attrReader.getValue("sessionKey");
   //        String value = attrReader.getValueOptional("value");
   //
   //        element.setSessionKey(sessionKey);
   //        element.setValue(value);
   //        element.setDebugInfo("inc " + sessionKey);
   //        attrReader.checkAttributesConsumed(); // Prfen, ob alle Attribute verbraucht wurden
   //        element.iLoopLevel = iLoopLevel;
   //        flowControlTable.add(element);
   //
   //        readEndTag("inc");
   //    }// end method parseInc

   //    /**
   //     * Parsen empty-Tag &lt;dec sessionKey="..." value="..."/&gt; . <br/>
   //     */
   //    protected static void parseDec()
   //
   //    {
   //        AttributeReader attrReader = readStartTag("dec");
   //
   //        FlowControlInstruction element = new FlowControlInstruction();
   //
   //        element.procOrFuncName = procOrFuncName;
   //        element.setSourcePosition(iClXmlLineNumber);
   //        element.setControlCode(ControlCode.DEC);
   //        String sessionKey = attrReader.getValue("sessionKey");
   //        String value = attrReader.getValueOptional("value");
   //
   //        element.setSessionKey(sessionKey);
   //        element.setValue(value);
   //        element.setDebugInfo("dec " + sessionKey);
   //        attrReader.checkAttributesConsumed(); // Prfen, ob alle Attribute verbraucht wurden
   //        element.iLoopLevel = iLoopLevel;
   //        flowControlTable.add(element);
   //
   //        readEndTag("dec");
   //    }// end method parseDec

   /**
    * Parsen empty-Tag &lt;continue/&gt; . <br/>
    */
   protected static void parseContinue(Segment pSegment)
   {
      FlowControlCompilerUtil.checkForOneIdentifier(pSegment);

      //addNop(pSegment.getSourcePos()); // NOP nur fr Einzelschritt-Debug

      FlowControlInstruction element = new FlowControlInstruction();

      element.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      element.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      element.controlCode = ControlCode.LOOP_CONTINUE;
      element.debugInfo = "continue";

      element.iLoopLevel = iLoopLevel;
      flowControlTable.add(element);
   } // end method parseContinue

   /**
    * Parsen empty-Tag &lt;break/&gt; . <br/>
    */
   protected static void parseBreak(Segment pSegment)
   {
      FlowControlCompilerUtil.checkForOneIdentifier(pSegment);

      //addNop(pSegment.getSourcePos()); // NOP nur fr Einzelschritt-Debug

      FlowControlInstruction element = new FlowControlInstruction();

      element.procName = procOrFuncName;
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      element.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      element.controlCode = ControlCode.LOOP_BREAK;
      element.debugInfo = "break";

      element.iLoopLevel = iLoopLevel;
      flowControlTable.add(element);
   } // end method parseBreak

   /**
    * Parsen des Befehls &lt;invalidate/&gt; . <br/>
    * Session-invalidate.
    */
   protected static void parseInvalidate(Segment pSegment)
   {
      FlowControlCompilerUtil.checkForOneIdentifier(pSegment);

      addNop(pSegment.getSourcePos());
      // NOP nur fr Einzelschritt-Debug, nach invalidate bricht debug sowieso zusammen

      FlowControlInstruction element = new FlowControlInstruction();

      element.procName = procOrFuncName;
      element.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      element.controlCode = ControlCode.SESSION_INVALIDATE;
      element.debugInfo = "invalidate";

      element.iLoopLevel = iLoopLevel;
      flowControlTable.add(element);
   } // end method parseBreak

   /**
    * Parsen des &lt;logon&gt;-Tags. <br/>
    * &lt;logon&gt; &lt;/logon&gt; ist nur in der main-Procedur erlaubt.<br/>
    */
   protected static void parseLogon(final Segment pSegment, final NameList pParamAndLocalVarsNameList)
      throws DomainParseException
   {
      FlowControlCompilerUtil.checkForIdentifierBraces(pSegment);

      if (!MAIN_PROC_NAME.equals(procOrFuncName))
      {
         // wenn Logon nicht in der main-procedur liegt
         FlowControlUtil.complain("logon only allows in procedure " + MAIN_PROC_NAME, pSegment.getSourcePos());
      } // end if

      if (pParamAndLocalVarsNameList == null)
      {
         // null als Hilfe, dass hier kein 'local' erlaubt ist
         FlowControlUtil.complain(
            "here is no logon area allowed, only at start of procedure/fucntion",
            pSegment.getSourcePos());
      }

      FlowControlInstruction anfangsElement = new FlowControlInstruction();

      anfangsElement.procName = procOrFuncName;
      anfangsElement.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      anfangsElement.controlCode = ControlCode.NOP;
      anfangsElement.debugInfo = "logon anfang (NOP)";

      anfangsElement.iLoopLevel = iLoopLevel;
      flowControlTable.add(anfangsElement);

      int iLogonBegin = flowControlTable.size() - 1;

      //        while (nextGenericTag(true))
      //        {
      //            ; // nextGenericTag( true ) erledigt alles
      //        }// end while
      readGenericTags(pSegment, 1, null);

      FlowControlInstruction endElement = new FlowControlInstruction();

      endElement.procName = procOrFuncName;
      endElement.sourcePosition = pSegment.getEndSourcePos();
      endElement.controlCode = ControlCode.NOP;
      endElement.debugInfo = "logon ende (NOP)";
      endElement.iLoopLevel = iLoopLevel;
      flowControlTable.add(endElement);

      // noUserValidation=true fr Elemente innerhalb <logon>...</logon> setzen
      for (int i = iLogonBegin; i < flowControlTable.size(); i++)
      {
         FlowControlInstruction instruction = (FlowControlInstruction) flowControlTable.get(i);

         instruction.noUserValidation = true;
      } // end for
   } // end method parseLogon

   /**
    * Parsen empty-Tag &lt;localvar sessionKey="..."/&gt; . <br/>
    *
    * @param parNameList NameList zur Kontrolle auf doppelt verwendete Parameter/lokale-Variable
    */
   protected static void parseLocalvar(final Segment pSegment, final NameList pParamAndLocalVarsNameList)
   {
      if (FlowControl.getBrowserLeading())
      {
         // Befehl localvar nur bei Session-Leading erlaubt
         FlowControlUtil.complain(
            "for local is flowleading session necessary (set in web.xml)",
            pSegment.getSourcePos());
      } // end if

      if (pParamAndLocalVarsNameList == null)
      {
         // null als Hilfe, dass hier kein 'local' erlaubt ist
         FlowControlUtil.complain(
            "here is no local declaration allowed, only at start of procedure/fucntion",
            pSegment.getSourcePos());
      }

      FlowControlCompilerUtil.checkForTwoIdentifier(pSegment);

      //addNop(pSegment.getSourcePos()); // NOP nur fr Einzelschritt-Debug

      FlowControlInstruction element = new FlowControlInstruction();

      element.procName = procOrFuncName;
      element.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());
      element.controlCode = ControlCode.LOCAL_VAR;
      element.debugInfo = "local variable";

      String localVarNameStr = FlowControlCompilerUtil.getIdentifierName(pSegment.getElement(1));

      pParamAndLocalVarsNameList.add(localVarNameStr); // Kontrolle auf doppelte Namen
      element.localVarName = localVarNameStr;
      element.debugInfo = "local var " + localVarNameStr;

      element.iLoopLevel = iLoopLevel;
      flowControlTable.add(element);

   }

   /**
    * Parsen eines Statements mit einer Expression
    */
   protected static void parseExpression(Segment pSegment) throws DomainParseException
   {
      //addNop(pSegment.getSourcePos()); // NOP nur fr Einzelschritt-Debug

      FlowControlInstruction element = new FlowControlInstruction();

      element.procName = procOrFuncName;

      if (pSegment.elementCount() < 1)
      {
         FlowControlUtil.complain(
            "expression expected",
            pSegment.getSourcePos());
      }
      //es wird die Position des ersten Elementes im Segment verwendet,
      //weil auf der Anfangsposition des Segmentes ein Kommentar stehen knnte
      element.sourcePosition = (pSegment.getElement( 0 ).getSourcePos());

      element.controlCode = ControlCode.EVAL_EXPRESSSION;
      element.debugInfo = "eval expression";

      element.setEvalExpression(pSegment);

      element.iLoopLevel = iLoopLevel;
      flowControlTable.add(element);
   } // end method parseExpression

   /**
    * Auflsen calls und gotos. <br/>
    * gotos drfen auch Prozeduren anspringen.<br/>
    * Beim Ausfhren eines GOTOs wird der Programmstack abgebaut.<br/>
    * Wird eine Prozdur verlassen, die ber ein GOTO angesprungen wurde,
    * so wird als nchstes der erste Befehl in der procedure main ausgefhrt.<br/>
    */
   protected static void resolveLabels()
   {
      // erst mal ganz langsame Methode implementiert, evtl. spter tunen

      // Auflsen Calls
      for (int i = 0; i < flowControlTable.size(); i++)
      {
         FlowControlInstruction instruction = (FlowControlInstruction) flowControlTable.get(i);

         if (instruction.controlCode == ControlCode.CALL)
         {
            String callName = instruction.procOrLabelNameToCall;
            int j = 0;

            for (j = 0; j < flowControlTable.size(); j++)
            {
               FlowControlInstruction suchElement = (FlowControlInstruction) flowControlTable.get(j);

               if (suchElement.controlCode == ControlCode.PROC_START //--
               || suchElement.controlCode == ControlCode.FUNC_START
               )
               {
                  if (callName.equals(suchElement.procName))
                  {
                     instruction.jumpDestination = j;

                     //Kontrolle, ob die Anzahl der Parameter beim CALL der Anzahl der Parameter der proc/func entsprechen
                     int callPrmNumber =
                        (instruction.callParamArrList == null) ? 0 : instruction.callParamArrList.size();
                     int procPrmNumber = (suchElement.procParamList == null) ? 0 : suchElement.procParamList.size();
                     if (callPrmNumber != procPrmNumber)
                     {
                        FlowControlUtil.complain(
                           "number of procedure call params "
                              + callPrmNumber
                              + " equals not with number of proc params "
                              + procPrmNumber,
                           instruction.sourcePosition);
                     }

                     if (suchElement.procParamList != null)
                     {
                        //Kontrolle auf call parameter is an variable identifier (session. ...) for refparam
                        for (int k = 0; k < suchElement.procParamList.size(); k++)
                        {
                           FlowProcDeclParam flowProcDeclParam = suchElement.procParamList.get(k);
                           if (flowProcDeclParam.isRefParam)
                           {
                              FlowProcCallParam flowProcedureCallParameter = instruction.callParamArrList.get(k);
                              String refParamKey =
                                 FlowControlUtil.checkAndCutPrefixSession(
                                    flowProcedureCallParameter.paramExpression.getSource(),
                                    flowProcedureCallParameter.sourcePosition);

                              //Kontrolle auf return sessionkey is no refparam
                              //System.out.println( "Kontrolle auf return sessionkey is no refparam" );
                              //System.out.println( "refParamKey: " + refParamKey );
                              //System.out.println( "instruction.returnSessionKey: " + instruction.returnSessionKey );
                              if (refParamKey.equals(instruction.returnSessionKey))
                              {
                                 FlowControlUtil.complain(
                                    "the same key for return value and refparam is forbidden",
                                    flowProcedureCallParameter.sourcePosition);
                              }
                           }
                        }
                     }

                     if ( instruction.returnSessionKey != null && suchElement.controlCode == ControlCode.PROC_START )
                     {
                         // bei einem Procedure-Aufruf ist keine Zuweisung eines Return-Wertes mglich
                         FlowControlUtil.complain(
                            "no assignment to variable on a procedure call available",
                            instruction.sourcePosition);
                     }

                     break;
                  } // end if
               } // end if
            } // end for
            if (j >= flowControlTable.size())
            { // nichts gefunden
               FlowControlUtil.complain("procedure " + callName + " not found", instruction.sourcePosition);
            } // end if
         } // end if
      } // end for

      // Auflsen GOTOs
      for (int i = 0; i < flowControlTable.size(); i++)
      {
         FlowControlInstruction element = (FlowControlInstruction) flowControlTable.get(i);

         if (element.controlCode == ControlCode.GOTO)
         {
            String gotoName = element.procOrLabelNameToCall;
            int j;

            for (j = 0; j < flowControlTable.size(); j++)
            {
               FlowControlInstruction suchElement = (FlowControlInstruction) flowControlTable.get(j);

               if (suchElement.controlCode == ControlCode.LABEL_FOR_GOTO
                  || suchElement.controlCode == ControlCode.PROC_START)
               {
                  if (gotoName.equals(suchElement.procOrLabelNameToCall))
                  {
                     element.jumpDestination = j;
                     break;
                  } // end if
               } // end if
            } // end for
            if (j >= flowControlTable.size())
            { // nichts gefunden
               FlowControlUtil.complain("procedure or label " + gotoName + " not found", element.sourcePosition);
            } // end if
         } // end if
      } // end for

      // Auflsen LOOP_CONTINUEs
      for (int i = 0; i < flowControlTable.size(); i++)
      {
         FlowControlInstruction element = (FlowControlInstruction) flowControlTable.get(i);

         if (element.controlCode == ControlCode.LOOP_CONTINUE)
         {
            // Rckwrts zum vorherigen START_WHILE oder START_DO laufen
            int j;

            for (j = i; j > -1; j--)
            {
               FlowControlInstruction suchElement = (FlowControlInstruction) flowControlTable.get(j);

               //. Problem Schachtelungstiefe:
               //. while ...    <----richtig-------+
               //. {                               |
               //.  while ...   <----falsch--+     |
               //.  {                        |     |
               //.    ...                    |     |
               //.  } //while                |     |
               //.                           |     |
               //.   continue;  -------------+-----+
               //. } //while

               if ((suchElement.controlCode == ControlCode.START_WHILE
                  || suchElement.controlCode == ControlCode.START_DO)
                  && suchElement.iLoopLevel == element.iLoopLevel)
               { // Schachtelungstiefe muss gleich sein, damit nicht eine falsche Schleife erwicht wird
                  element.jumpDestination = j;
                  break;
               } // end if

               if (suchElement.controlCode == ControlCode.PROC_START)
               {
                  // an den Anfang der aktuellen procedure gelaufen
                  FlowControlUtil.complain("while or do not found", element.sourcePosition);
               } // end if

            } // end for
            if (j < 0)
            { // nichts gefunden
               FlowControlUtil.complain("while or do for continue not found", element.sourcePosition);
            } // end if
         } // end if
      } // end for

      // Auflsen LOOP_BREAKs
      for (int i = 0; i < flowControlTable.size(); i++)
      {
         FlowControlInstruction element = (FlowControlInstruction) flowControlTable.get(i);

         if (element.controlCode == ControlCode.LOOP_BREAK)
         {
            // Vorwrts zum nchsten END_WHILE oder END_DO laufen
            int j;

            for (j = i; j < flowControlTable.size(); j++)
            {
               FlowControlInstruction suchElement = (FlowControlInstruction) flowControlTable.get(j);

               if ((suchElement.controlCode == ControlCode.END_WHILE || suchElement.controlCode == ControlCode.END_DO)
                  && suchElement.iLoopLevel == element.iLoopLevel)
               { // Schachtelungstiefe muss gleich sein, damit nicht eine falsche Schleife erwicht wird
                  element.jumpDestination = (j + 1);
                  // einen Befehl hinter die Adresse springen, um Schleife abzubrechen
                  break;
               } // end if
               if (suchElement.controlCode == ControlCode.PROC_START)
               {
                  // an den Anfang der nchsten procedure gelaufen
                  FlowControlUtil.complain("while or do for break not found", element.sourcePosition);
               } // end if
            } // end for
            if (j >= flowControlTable.size())
            { // nichts gefunden
               FlowControlUtil.complain("while or do not found", element.sourcePosition);
            } // end if
         } // end if
      } // end for

   } // end method resolveLabels

   /**
    * NOP (No-Operation-Befehl) einfgen
    */
   protected static void addNop(SourcePosition pSourcePosition)
   {
      FlowControlInstruction nopElement = new FlowControlInstruction();

      nopElement.procName = procOrFuncName;
      nopElement.sourcePosition = pSourcePosition;
      nopElement.controlCode = ControlCode.NOP;
      // nopElement.setName( "" ) ;
      nopElement.debugInfo = ("NOP for Single-Step-Debug");

      nopElement.iLoopLevel = iLoopLevel;
      flowControlTable.add(nopElement);
   } // end method addNop

   /**
    * Zurckliefern String-Array fr die Methode FlowControl#getHtmlDebug
    */
   public static FlowSrcFileLine[] getSrcFileLineArr()
   {
      return srcFileLineArr;
   } // end method getXmlFileLineArr

   /**
    * Erzeugen String-Array fr die Methode FlowControl#getHtmlDebug
    */
   private static void initSrcFileStrArr(String fileNameWithPath /*,String pEncoding*/
   )
   {
      // erst mal flow.cnc in eindimensionales Array einziehen
      String[] oneDimSrcFileStrArr = StringUtil.fileToStrArr(fileNameWithPath /*,pEncoding*/
      );

      // Array mit Namen der Procedure
      srcFileLineArr = new FlowSrcFileLine[oneDimSrcFileStrArr.length];
      // umkopieren
      for (int i = 0; i < oneDimSrcFileStrArr.length; i++)
      {
         srcFileLineArr[i] = new FlowSrcFileLine();
         srcFileLineArr[i].lineEncodet = StringUtil.encodeHtml(oneDimSrcFileStrArr[i]); // encodet
         srcFileLineArr[i].lineNotEncodet = oneDimSrcFileStrArr[i]; // nicht encodet
      } // end for

   } // end method initXmlFileStrArr

   /**
    * Einlesen Parameter-Liste procedure
    *
    * @return NameList zur Kontrolle auf doppelt verwendete Parameter/lokale-Variable
    */
   private static NameList readProcParams(
      FlowControlInstruction paInstruction,
      CommaSeparatedList pCommaSeparatedList) // AttributeReader attrReader)
   {
      NameList retNameList = new NameList();

      // if (attrReader.existAttribute("param")) {
      if (pCommaSeparatedList != null && !pCommaSeparatedList.isEmpty())
      {
         if (FlowControl.getBrowserLeading())
         {
            // Prozedur-Parameter nur bei Session-Leading erlaubt
            FlowControlUtil.complain(
               "for procedure or function parameters is flowleading session necessary (set in web.xml)",
               pCommaSeparatedList.getSourcePos());
         } // end if

         paInstruction.procParamList = new FlowProcDeclParamList();
         // while (attrReader.existAttribute("param"))
         for (int i = 0; i < pCommaSeparatedList.elementListCount(); i++)
         {
            ElementList elemList = pCommaSeparatedList.getElementList(i);

            FlowProcDeclParam flowProcDeclParam = new FlowProcDeclParam();

            Identifier nameElem = null;
            if (elemList.elementCount() == 1)
            {
               nameElem = FlowControlCompilerUtil.checkForIdentifier(elemList.getElement(0));
            }
            else if (
               elemList.elementCount() == 2
                  && FlowControlCompilerUtil.checkForIdentifier(elemList.getElement(0)).getNettoContent().equals(
                     "refparam"))
            {
               //Reference-Parameter
               nameElem = FlowControlCompilerUtil.checkForIdentifier(elemList.getElement(1));
               flowProcDeclParam.isRefParam = true;
            }
            else
            {
               FlowControlUtil.complain(
                  "'identifier' or 'refparam identifier' per comma area expected",
                  elemList.getSourcePos());
            }

            // String strInsideKey = attrReader.getValue("param");
            String strInsideKey =
               FlowControlUtil.checkAndCutPrefixSession(nameElem.getContent(), nameElem.getSourcePos());

            retNameList.add(strInsideKey); // Kontrolle auf doppelte Namen
            flowProcDeclParam.insideKey = strInsideKey;
            paInstruction.procParamList.add(flowProcDeclParam);

            //Log.debug( "\nFlowControlParser#readProcParams " + flowProcDeclParam );
         } // end while
      } // end if
      // Log.debug( "proc param " + paInstruction.getProcName() + " " + paInstruction.procParamList ) ;
      return retNameList;
   } // end method readProcParams

   /**
    * Einlesen Parameter-Liste call
    */
   private static void readCallParams(FlowControlInstruction paInstruction, AbstractElement pElement)
   {
      ParenthesisArea parenthesisArea = FlowControlCompilerUtil.checkForParenthesisArea(pElement);
      CommaSeparatedList commaSeparatedList = parenthesisArea.getCommaSeparatedList();

      //Log.debug( "commaSeparatedList" , commaSeparatedList );

      if (!(commaSeparatedList == null || commaSeparatedList.isEmpty()))
      {
         if (FlowControl.getBrowserLeading())
         {
            // Prozedur-Parameter nur bei Session-Leading erlaubt
            FlowControlUtil.complain(
               "for call parameter is flowleading session necessary (set in web.xml)",
               commaSeparatedList.getSourcePos());
         } // end if

         paInstruction.callParamArrList = new FlowProcParamList();
         //            while (attrReader.existAttribute("param")
         //                    || attrReader.existAttribute("sessionKeyParam")
         //                    || attrReader.existAttribute("refParam")
         //                    || attrReader.existAttribute("literalParam"))
         //            {
         //                FlowProcCallParam flowProcedureParameter = new FlowProcCallParam();
         //
         //                if (attrReader.isFirstAttribute("param"))
         //                {
         //                    flowProcedureParameter.isRefParam = false;
         //                    flowProcedureParameter.outsideKey = attrReader.getValue("param");
         //                }
         //                else if (attrReader.isFirstAttribute("sessionKeyParam"))
         //                {
         //                    // ist das gleiche wie param
         //                    flowProcedureParameter.isRefParam = false;
         //                    flowProcedureParameter.outsideKey = attrReader.getValue("sessionKeyParam");
         //                }
         //
         //                else if (attrReader.isFirstAttribute("refParam"))
         //                {
         //                    flowProcedureParameter.isRefParam = true;
         //                    flowProcedureParameter.outsideKey = attrReader.getValue("refParam");
         //                }
         //                else if (attrReader.isFirstAttribute("sessionKeyRefParam"))
         //                {
         //                    flowProcedureParameter.isRefParam = true;
         //                    flowProcedureParameter.outsideKey = attrReader.getValue("sessionKeyRefParam");
         //                }
         //
         //                else if (attrReader.isFirstAttribute("literalParam"))
         //                {
         //                    flowProcedureParameter.isLiteralParam = true;
         //                    flowProcedureParameter.literalValue = attrReader.getValue("literalParam");
         //                }
         //
         //                else
         //                {
         //                    complain("param-Attribute or refParam-Attribute expected");
         //                }// end if
         for (int i = 0; i < commaSeparatedList.elementListCount(); i++)
         {
            FlowProcCallParam flowProcedureParameter = new FlowProcCallParam();
            flowProcedureParameter.sourcePosition = commaSeparatedList.getElementList(i).getSourcePos();

            flowProcedureParameter.isRefParam = false;

            //hier noch refParam einbauen

            try
            {
               String exprsStr = commaSeparatedList.getElementList(i).getContent();
               flowProcedureParameter.paramExpression = Expression.parseStrict(exprsStr, 0, 0);
               flowProcedureParameter.sourcePosition = commaSeparatedList.getElementList(i).getSourcePos();

               //                    flowProcedureParameter.outsideKey = FlowControlUtil.checkAndCutPrefixSession(
               //                        commaSeparatedList.getElementList( i ).getNettoContent(),
               //                        commaSeparatedList.getElementList( i ).getSourcePos() );

            }
            catch (ExpressionParseException e)
            {
               FlowControlUtil.complain(e.getMessage(), commaSeparatedList.getElementList(i).getSourcePos());
               //System.err.println(e.getMessage());
               //e.printStackTrace();
               //throw new RuntimeException( e.getMessage() );
            }

            paInstruction.callParamArrList.add(flowProcedureParameter
            /*, commaSeparatedList.getElementList( i ).getSourcePos()*/
            );
         } // end while
      } // end if
      // Log.debug( "call param " + paInstruction.getProcName() + ": "  + paInstruction.getName() + " " + paInstruction.callParamArrList ) ;
   } // end method readCallParams

} // end class FlowControlCompiler
