package de.cnc.persistens;

import java.util.*;

import de.cnc.util.BeanUtil;
import de.cnc.util.config.ConfigManager;

/**
 * Diese Klasse ist ein Cache fr Daten aus der Datenbank
 *
 * @author Heiner Kcker
 * @since 20.09.2004
 */
public class DatabaseCache
{
   // @todo: long hash map einbauen, problem synchronized map

   /**
    * Konstruktor private.
    * Instanziieren verboten
    */
   private DatabaseCache()
   {
      ;
   }

   /**
    * Ein Cache mit den Caches fr die jeweiligen Klassen
    * fr den LookUp nach dem Primery-Key
    */
   private static Map primaryKeyClassCacheMap = Collections.synchronizedMap( new HashMap() );

   /**
    * Ein Cache mit den Caches fr die jeweiligen Klassen
    * fr den LookUp nach den Foreign-Keys
    */
   private static Map foreignKeyClassCacheMap = Collections.synchronizedMap( new HashMap() );

   /**
    * ein extra Cache fr gecachte SELECT-Statementes
    */
   private static SelectStmtCache selectStmtCache = new SelectStmtCache();

   /**
    * Mindestzeitspanne zwischen zwei Reorganisations-Lufen in Millisekunden
    */
   public static int REORG_OFF_TIME = /*5 * */
   60000;

   /**
    * Zeitpunkt der nchsten Reorganisation in Millisekunden ab 1.1.1970
    */
   private static long nextReorgTime = System.currentTimeMillis() + REORG_OFF_TIME;

   /**
    * maximale Lebensdauer im Cache, Default 15 Minuten
    */
   private static int cacheExpirationTime = ( 15 * 60 * 1000);

   /**
    * Cache-Reorganisastion mit einem eigenen Thread
    */
   static Reorganizer reorganizer;

   /**
    * Initialisieren des Cache
    */
   public static synchronized void init()
   {
      int newCacheExpirationTime = ConfigManager.getDefaultProperties().getInt("cacheexpirationtime");
      if (newCacheExpirationTime > 0)
      {
         cacheExpirationTime = newCacheExpirationTime;
      }

      //Reorganisations-Thread starten
      if (reorganizer == null)
      {
         reorganizer = new Reorganizer();
         reorganizer.start();
      }
   }

   /**
    *
    * @return ob der Cache leer ist
    */
   static synchronized boolean reorg()
   {
      //      //reorganisieren maximal alle REORG_OFF_TIME Milli-Sekunden
      //      if (nextReorgTime > System.currentTimeMillis() )
      //      {
      //         return;
      //      }

      //Log.debug( "DatabaseCache#reorg");
      
      final long expiredTime = System.currentTimeMillis() - cacheExpirationTime;

      //Caches fr Primery-Key reorganisieren
      boolean primeryKeyCachesEmpty = reorgPrimeryKeyCaches(expiredTime);
//Log.debug( "primeryKeyCachesEmpty" , "" + primeryKeyCachesEmpty);

      //Caches fr Foreign-Keys reorganisieren
      boolean foreignKeyCachesEmpty = reorgForeignKeyCaches(expiredTime);
//Log.debug( "foreignKeyCachesEmpty" , "" + foreignKeyCachesEmpty);


      //Cache fr SELECT-WHERE-Statementes reorganisieren
      boolean selectStmtCachesEmpty = selectStmtCache.reorg(expiredTime);
//Log.debug( "selectStmtCachesEmpty" , "" + selectStmtCachesEmpty);

      nextReorgTime = System.currentTimeMillis() + REORG_OFF_TIME;
      
      return (primeryKeyCachesEmpty && foreignKeyCachesEmpty && selectStmtCachesEmpty);
   }

   /**
    * Cache fr Primery-Key reorganisieren
    * 
    * @param expiredTime
    * @return
    */
   private static boolean reorgPrimeryKeyCaches(final long expiredTime)
   {
      boolean allCachesEmpty = true;
         Set entrySet = primaryKeyClassCacheMap.entrySet();
         Iterator iterator = entrySet.iterator();
         while (iterator.hasNext())
         {
            Map.Entry entry = (Map.Entry) iterator.next();
            AbstractTableCache tableCache = (AbstractTableCache) entry.getValue();
            tableCache.reorg(expiredTime);
            
            if (tableCache.isEmpty())
            {
               iterator.remove();
            }
            else
            {
               allCachesEmpty = false;
            }
         }
      return allCachesEmpty;
   }

   /**
    * Cache fr Foreign-Keys reorganisieren
    * 
    * @param expiredTime
    * @return
    */
   private static boolean reorgForeignKeyCaches(final long expiredTime)
   {
      boolean allCachesEmpty = true;
      Set cacheEntrySet = foreignKeyClassCacheMap.entrySet();
      Iterator cacheIterator = cacheEntrySet.iterator();
      while (cacheIterator.hasNext())
      {
         Map.Entry cacheEntry = (Map.Entry) cacheIterator.next();
         Map tableCache = (Map) cacheEntry.getValue();
      
         Set tableEntrySet = tableCache.entrySet();
         Iterator tableEntryIterator = tableEntrySet.iterator();
         while (tableEntryIterator.hasNext())
         {
            Map.Entry tableEntry = (Map.Entry) tableEntryIterator.next();
            Map tableFieldCache = (Map) tableEntry.getValue();
      
            Set tableFieldEntrySet = tableFieldCache.entrySet();
            Iterator tableFieldIterator = tableFieldEntrySet.iterator();
            while (tableFieldIterator.hasNext())
            {
               Map.Entry tableFieldEntry = (Map.Entry) tableFieldIterator.next();
               PrimeryKeyRecordCache tableFieldValueCache = (PrimeryKeyRecordCache) tableFieldEntry.getValue();
      
               tableFieldValueCache.reorg(expiredTime);
               if (tableFieldValueCache.isEmpty())
               {
                  tableFieldIterator.remove();
               }
               else
               {
                  allCachesEmpty = false;
               }
            }
            if (tableFieldCache.isEmpty())
            {
               tableEntryIterator.remove();
            }
            else
            {
               allCachesEmpty = false;
            }
         }
         if (tableCache.isEmpty())
         {
            cacheIterator.remove();
         }
         else
         {
            allCachesEmpty = false;
         }
      }
      return allCachesEmpty;
   }

   /**
    * put mit Object-ID.
    *
    * @param pStorable
    */
   private static synchronized void putForPrimeryKey(final IStorable pStorable)
   {
      //Log.debug( "Databasecache#put: " + pStorable);
      //wenn die DB-ID fehlt, nicht cachen
      //if (pStorable.getObjectId() == IStorable.NOT_STORED)
      //{
      //   return;
      //}

      PrimeryKeyRecordCache tableCache = (PrimeryKeyRecordCache) primaryKeyClassCacheMap.get(pStorable.getClass());
      if (tableCache == null)
      {
         tableCache = new PrimeryKeyRecordCache();
         primaryKeyClassCacheMap.put(pStorable.getClass(), tableCache);
      }

      tableCache.put(pStorable);
   }

   static synchronized void put(final IStorable pStorable)
   {
      //Initialisieren Cache, wenn notwendig
      if (reorganizer == null)
      {
         init();
      }

      //Speichern nach Primery-Key
      String primeryKeyFieldName = PersistensHelper.getPrimaryKeyFieldName(pStorable);
      if (primeryKeyFieldName != null)
      {
         putForPrimeryKey(pStorable);
      }
      //Speichern nach Foreign-Keys
      String[] foreignKeyPropertyNameArr = PersistensHelper.getForeignKeyPropertyNameArr(pStorable);
      if (foreignKeyPropertyNameArr != null && foreignKeyPropertyNameArr.length > 0)
      {
         for (int i = 0; i < foreignKeyPropertyNameArr.length; i++)
         {
            putForForeignKey(pStorable, foreignKeyPropertyNameArr[i]);
         }
      }
   }

   /**
    * put mit Object-ID.
    *
    * @param pStorable
    */
   private static synchronized void putForForeignKey(//--
   final IStorable pStorable, //--
   final String pKeyPropertyName)
   {
      //Log.debug( "Databasecache#put: " + pStorable);
      //wenn die DB-ID fehlt, nicht cachen
      //if (pStorable.getObjectId() == IStorable.NOT_STORED)
      //{
      //   return;
      //}

      //Ermittlung Map fr die Tabelle
      Map tableCache = (Map) foreignKeyClassCacheMap.get(pStorable.getClass());
      if (tableCache == null)
      {
         tableCache = Collections.synchronizedMap( new HashMap() );
         foreignKeyClassCacheMap.put(pStorable.getClass(), tableCache);
      }

      //Ermittlung Cache fr das spezifizierte Feld
      Map fieldCache = (Map) tableCache.get(pKeyPropertyName);
      if (fieldCache == null)
      {
         fieldCache = Collections.synchronizedMap( new HashMap() );
         tableCache.put(pKeyPropertyName, fieldCache);
      }

      //Ermittlung der Map der zum Foreign-Key gehrenden Objekte
      //      ForeignKeyRecordCache tableFieldCache = (ForeignKeyRecordCache) fieldCache.get( (Long) BeanUtil.get( pStorable , pKeyPropertyName ) );
      //      if (tableFieldCache == null)
      //      {
      //         tableFieldCache = new ForeignKeyRecordCache();
      //         fieldCache.put( pKeyPropertyName , tableFieldCache );
      //      }
      //
      //      tableFieldCache.put( pStorable );

      //Ermittlung Cache fr den spezifizierte Foreign-Key-Value
      final Long foreignKeyValue = (Long) BeanUtil.get(pStorable, pKeyPropertyName);
      PrimeryKeyRecordCache foreignKeyValueCache = (PrimeryKeyRecordCache) fieldCache.get(foreignKeyValue);
      if (foreignKeyValueCache == null)
      {
         foreignKeyValueCache = new PrimeryKeyRecordCache();
         fieldCache.put(foreignKeyValue, foreignKeyValueCache);
      }

      foreignKeyValueCache.put(pStorable);
   }

   /**
    * put mit WHERE-String.
    *
    * @param pStorable
    */
   //   /*package*/ static synchronized void put( // --
   //         final IStorable pStorable, // --
   //         final String pWhereStr)
   //   {
   //Log.debug( "Databasecache#put: " + pStorable + " where: " + pWhereStr);
   //
   //      if (PersistensHelper.getPrimaryKeyFieldName(pStorable) != null)
   //      {
   //         //hat eine ID, also nach ID cachen
   //         DatabaseCache.put(pStorable);
   //         return;
   //      }
   //
   //      WhereTableCache tableCache = (WhereTableCache) primKeyClassCacheMap.get( pStorable.getClass() );
   //      if (tableCache == null)
   //      {
   //         tableCache = new WhereTableCache();
   //         primKeyClassCacheMap.put( pStorable.getClass() , tableCache );
   //      }
   //
   //      tableCache.put( pStorable , pWhereStr );
   //   }

   /**
    * put mit WHERE-String.
    *
    * @param pStorable
    */
   static synchronized void putForWhereStr(// --
   final List pValList, // --
   final String pSelectStr)
   {
      //Log.debug( "Databasecache#put: " + pValList + " where: " + pSelectStr);

      selectStmtCache.put(pValList, pSelectStr);
   }

   static synchronized IStorable getForPrimeryKey(// --
   final Class pClass, //--
   final Long pIdObj)
   {
      final PrimeryKeyRecordCache tableCache = (PrimeryKeyRecordCache) primaryKeyClassCacheMap.get(pClass);
      if (tableCache != null)
      {
         //Log.debug( "DatabaseCache#get: " + tableCache.get( pIdObj ) );
         return tableCache.get(pIdObj);
      }
      return null;
   }

   //   /*package*/ static synchronized IStorable get( // --
   //      final Class pClass, //--
   //      final String pWhereStr)
   //   {
   //      final WhereTableCache tableCache = (WhereTableCache) primKeyClassCacheMap.get( pClass );
   //      if (tableCache != null)
   //      {
   //Log.debug( "DatabaseCache#get: " + tableCache.get( pWhereStr ) + " where: " + pWhereStr );
   //         return tableCache.get( pWhereStr );
   //      }
   //      return null;
   //   }

   /*package*/
   static synchronized List getForWhereStr(// --
   //final Class pClass, //--
   final String pWhereStr)
   {
      return selectStmtCache.get(pWhereStr);
   }

   /*package*/
   static synchronized IStorable[] getForForeignKey(// --
   final Class pClass, //--
   final String pKeyPropertyName, //--
   final Long pIdObj)
   {
      //Ermittlung Map fr die Tabelle
      Map tableCache = (Map) foreignKeyClassCacheMap.get(pClass);
      if (tableCache == null)
      {
         return null;
      }

      //Ermittlung Cache fr das spezifizierte Feld
      Map fieldCache = (Map) tableCache.get(pKeyPropertyName);
      if (fieldCache == null)
      {
         return null;
      }

      //Ermittlung Cache fr den spezifizierte Foreign-Key-Value
      //    ForeignKeyRecordCache tableFieldCache = (ForeignKeyRecordCache) tableCache.get( pKeyPropertyName );
      //    if (tableFieldCache == null)
      //    {
      //       return null;
      //    }
      PrimeryKeyRecordCache foreignKeyValueCache = (PrimeryKeyRecordCache) fieldCache.get(pIdObj);
      if (foreignKeyValueCache == null)
      {
         return null;
      }

      return foreignKeyValueCache.valueArr();
   }

   /**
    * Remove Cache-Element . <br>
    * Removen des SELECT-WHERE-Cache funktioniert nicht.
    *
    * @param pStorable
    */
   static synchronized void remove(final IStorable pStorable)
   {
      //wenn die DB-ID fehlt, nicht cachen
//      if (pStorable.getObjectId() == IStorable.NOT_STORED)
//      {
//         return;
//      }

      PrimeryKeyRecordCache tableCache = (PrimeryKeyRecordCache) primaryKeyClassCacheMap.get(pStorable.getClass());
      if (tableCache != null)
      {
         tableCache.remove(new Long(pStorable.getObjectId()));
      }
      
      //@todo: remove im Foreign-Key-Cache
   }

   /**
    * @param pStorable
    */
   static synchronized void removeAll(final Class pClass)
   {
      primaryKeyClassCacheMap.remove(pClass);
   }

   /**
    * Diese Klasse ...
    *
    * @author Heiner Kcker
    * @since 20.09.2004
    */
   private static abstract class AbstractTableCache
   {
      /**
       *
       */
      protected Map rowCacheMap = Collections.synchronizedMap( new HashMap() );

      /**
       *
       * @param pExpiredTime
       * @return ob der Cache leer ist
       */
      protected boolean reorg(final long pExpiredTime)
      {
         //Log.debug( "AbstractTableCache" , "reorg" );
         //Log.debug( this.getClass().getName() , "reorg" );
         
         boolean isCacheEmpty = true;
         
         Set entrySet = rowCacheMap.entrySet();
         Iterator iterator = entrySet.iterator();
         while (iterator.hasNext())
         {
            Map.Entry mapEntry = (Map.Entry) iterator.next();
            TableCacheEntry tableCacheEntry = (TableCacheEntry) mapEntry.getValue();
            if (tableCacheEntry.lastUsed < pExpiredTime)
            {
               //Eintrag abgelaufen, entfernen
               //Log.debug( "Cache#remove" , tableCacheEntry.value );
               iterator.remove();
            }
            else
            {
               isCacheEmpty = false;
            }
         }
         return isCacheEmpty;
      }

      /**
       * @return
       */
      public boolean isEmpty()
      {
         return rowCacheMap.isEmpty();
      }

   }

   /**
    * Diese Klasse ...
    *
    * @author Heiner Kcker
    * @since 20.09.2004
    */
   private static class TableCacheEntry
   {
      /*private*/
      long lastUsed;

      /*private*/
      Object value;
   }

   /**
    * Diese Klasse ...
    *
    * @author Heiner Kcker
    * @since 20.09.2004
    */
   private static class PrimeryKeyRecordCache extends AbstractTableCache
   {
      /**
       *
       * @param pStorable
       * @param pId
       */
      void put(//--
      final IStorable pStorable) //--
      //final Long pId ) //--
      //final Object pKeyObject)
      {
         long pId = pStorable.getObjectId();
         Long idObj = new Long(pId);
         TableCacheEntry entry = (TableCacheEntry) rowCacheMap.get(idObj);
         if (entry == null)
         {
            entry = new TableCacheEntry();
            rowCacheMap.put(idObj, entry);
         }
         entry.lastUsed = System.currentTimeMillis();
         entry.value = pStorable;
      }

      /**
       *
       * @param pIdObj
       * @return
       */
      IStorable get(final Long pIdObj)
      {
         TableCacheEntry entry = (TableCacheEntry) rowCacheMap.get(pIdObj);
         if (entry == null)
         {
            return null;
         }
         entry.lastUsed = System.currentTimeMillis();
         return (IStorable) entry.value;
      }

      /**
       *
       * @param pIdObj
       */
      void remove(final Long pIdObj)
      {
         rowCacheMap.remove(pIdObj);
      }

      /**
       * 
       * @return
       */
      Collection values()
      {
         return rowCacheMap.values();
      }

      /**
       * 
       * @return
       */
      IStorable[] valueArr()
      {
         TableCacheEntry[] entryArr =
            (TableCacheEntry[]) rowCacheMap.values().toArray(new TableCacheEntry[rowCacheMap.size()]);
         if (entryArr == null || entryArr.length < 1)
         {
            return null;
         }
         IStorable[] storableArr = new IStorable[entryArr.length];

         //java.lang.reflect.Array.newInstance(
         //      entryArr[0].getClass(),
         //      entryArr.length);
         for (int i = 0; i < entryArr.length; i++)
         {
            storableArr[i] = (IStorable) entryArr[i].value;
         }
         return storableArr;
      }
   }

   /**
    * Diese Klasse enthlt Cache-Items, die entsprechend dem SELECT-Statement-String gecacht werden.
    *
    * @author Heiner Kcker
    * @since 20.09.2004
    */
   private static class SelectStmtCache extends AbstractTableCache
   {
      /*private*/
      void put(final List pValList, final String pWhereStr)
      {
         TableCacheEntry entry = (TableCacheEntry) rowCacheMap.get(pWhereStr);
         if (entry == null)
         {
            entry = new TableCacheEntry();
            rowCacheMap.put(pWhereStr, entry);
         }
         entry.lastUsed = System.currentTimeMillis();
         entry.value = pValList;
      }

      /*private*/
      List get(final String pWhereStr)
      {
         TableCacheEntry entry = (TableCacheEntry) rowCacheMap.get(pWhereStr);
         if (entry == null)
         {
            return null;
         }
         entry.lastUsed = System.currentTimeMillis();
         return (List) entry.value;
      }
   }

   /**
    * Diese Klasse ...
    *
    * @author Heiner Kcker
    * @since 20.09.2004
    */
//   private static class ForeignKeyRecordCache extends AbstractTableCache
//   {
//      /**
//       *
//       * @param pStorable
//       * @param pId
//       */
//      void put(//--
//      final IStorable pStorable, //--
//      final Long pId) //--
//      //final Object pKeyObject)
//      {
//         //long pId = pStorable.getObjectId();
//         //Long idObj = new Long( pId );
//         PrimeryKeyRecordCache entry = (PrimeryKeyRecordCache) rowCacheMap.get(/*idObj*/
//         pId);
//         if (entry == null)
//         {
//            entry = new PrimeryKeyRecordCache();
//            rowCacheMap.put(/*idObj*/
//            pId, entry);
//         }
//         entry.put(pStorable);
//      }
//
//      /**
//       *
//       * @param pIdObj
//       * @return
//       */
//      Object[] get(final Long pIdObj)
//      {
//         PrimeryKeyRecordCache fieldValueCache = (PrimeryKeyRecordCache) rowCacheMap.get(pIdObj);
//         if (fieldValueCache == null)
//         {
//            return null;
//         }
//         return fieldValueCache.valueArr();
//      }
//
//      /**
//       *
//       * @param pIdObj
//       */
//      void remove(final Long pIdObj)
//      {
//         rowCacheMap.remove(pIdObj);
//      }
//
//   }

   /**
    * Diese Klasse ist ein Thread zum Reorganisieren
    * des Caches in einem bestimmeten Intervall.
    * Er sorgt dafr, dass die abgelaufenen
    * Cache-Eintrge auch ohne User-Aktivitt gelscht
    * werden.
    *
    * @author Heiner Kcker
    * @since 23.09.2004
    */
   private static class Reorganizer extends Thread
   {
      public void run()
      {
         //Log.debug( "Reorganizer" , "run" );
         while (true)
         {
            try
            {
               Thread.sleep(REORG_OFF_TIME);
            }
            catch (InterruptedException e)
            {
            }
            boolean allCachesEmpty = DatabaseCache.reorg();
            
            if (allCachesEmpty)
            {
               //alle caches sind leer, scheinbar arbeitet niemand mehr auf der Applikation
               //Log.debug( "Reorganizer" , "stop" );
               //reorganizer.stop();
               reorganizer = null;
               return;
            }
         }
      }
   }

}
