package de.cnc.persistens;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

/**
 * Speichert Connection-Objekte fr die diversen verwendeten Datenbanken (zum Datenbanknamen)
 * SINGLETON
 * @author Anja Odenthal
 * 
 * nderungen:
 * 03.09.02 Erzeugen der SAMExceptions nachgezogen.
 * 17.09.02 O.G. Die Klasse synchronisiert, da evt. ohne Probleme auftreten knnen
 * 22.12.03 Bestimmung der DB-Schemata jetzt ber SAMServiceLocatorClient [DH]
 */
public class DbConnectionFactory
{

   /**
    * Konstanten fr die vorhandenen (JNDI-) Datenbanknamen
    */
   public static final String DBNAME_NATDB = "NART_REF";	
   public static final String DBNAME_SORTDB = "SORT_REF"; 
   public static final String DBNAME_BESIDB = "BESI_REF";
   public static final String DBNAME_REBIZDB = "REBIZ_REF";
   public static final String DBNAME_SPACEMANDB = "SPACEMAN_REF";
   
   /**
    * Kommt aus dbConnection.properties:
    * Wenn true, werden alle open- und close-Operationen auf Connections sowie die close-
    * Operationen auf Statements sowie die Anzahl offener Connections je Datenbank geloggt.
    */
   public static boolean iDebug = false;

   /**
    * Qualifier vor dem Datenbanknamen fr den JNDI-Lookup
    * Ist aus Kompatibilittsgrnden mit dem WAS auf OS/390 erforderlich
    * Der Qualifier ist unter NT anders als unter OS/390, er wird daher beim Erzeugen des Singletons
    * aus einer Properties-Datei geladen.
    */
   public static String iJndiQualifier;

   /**
    * die Singleton-Instanz der Factory
    */
   private static DbConnectionFactory dbConnectionFactory;

   /**
    * Hashtable, welche die Zuordnung zwischen DB-Namen und Connection-Objekten macht
    */
   private static HashMap iDataSources;

   /**
    * Enthlt DB-Zugriffsinformationen wie Schemanamen, JNDI-Qualifier, UIDs.
    * Aus derDatei dbConnection.properties
    */
   //private static Properties iDbProps;

   /**
    * Fhrt die Anzahl momentan offener Connections mit.
    */
   private static ConnectionCounter iConnCounter;

   /**
    * Einmaliges Erzeugen der Singleton-Instanz
    */
   static
   {
      dbConnectionFactory = new DbConnectionFactory();
   }

   /**
    * Privater Konstruktor, weil wir ein Singleton verwenden
    * Initialisiert die Zuordnungen Datenbankname - UserId und Datenbankname - Passwort (iDbProps).
    */
   private DbConnectionFactory()
   {
      synchronized (this)
      {
         iDataSources = new HashMap();

         //iDbProps = Configuration.getProperties(this.getClass());

         iJndiQualifier = "java:comp/env/" ; //iDbProps.getProperty("JNDI_QUALIFIER");
         //iDebug = Boolean.getBoolean(iDbProps.getProperty("DEBUG"));

         iConnCounter = new ConnectionCounter();
      }
   }

   /**
    * @see getDataSource(String)
    */
   private DataSource _getDataSource(String pDbName)
   {
      DataSource ds = null;

      // Wenn das DataSource-Objekt bereits existiert: liefere es aus dem Cache zurck.
      ds = (DataSource) iDataSources.get(pDbName);
      if (ds != null)
      {
         return ds;
      }

      // Sonst: Neue DataSource-Instanz erzeugen (JNDI-Lookup)
      try
      {
         InitialContext ic = new InitialContext();

         ds = (DataSource) ic.lookup(iJndiQualifier + pDbName);
      }
      catch (NamingException e)
      {
         e.printStackTrace();
         throw new RuntimeException( "JNDI lookup failed: " + iJndiQualifier + pDbName );
      }

      // Gefundenes DataSource-Objekt in die Hashtable tun
      if (ds != null)
      {
         synchronized (iDataSources)
         {
            iDataSources.put(pDbName, ds);
         }
         return ds;
      }

      // ToDo: kann der Fall ds == null berhaupt auftreten?
      throw new RuntimeException( "JNDI lookup failed: " + iJndiQualifier + pDbName );
   }

   /**
    * ffnet eine Connection fr die DB
    * @param pDbName	Name der gewnschten DB, wie in der Properties-Datei eingegeben
    * @return	die erzeugte Connection
    */
   public static final synchronized Connection makeDBConnection(String pDbName)
   {
//   	  if(iDebug)
//   	  {
//   	  	de.rewe.sam.test.DebugUtils.println( "Con-Request:"+de.rewe.sam.test.DebugUtils.methodInfo(1) );
//   	  }
   	  
      return dbConnectionFactory._makeDBConnection(pDbName);
   }

   /**
    * @see makeDBConnection( String )
    */
   private Connection _makeDBConnection(String pDbName)
   {
      Connection conn;
      DataSource dataSource = _getDataSource(pDbName);

      // create connection
      try
      {
         StringBuffer logMsg = null;

         if (iDebug)
         {
            logMsg = new StringBuffer("Opening connection... ");
         }
         // Uid und Pw stehen in den entspr. Datasources
         conn = dataSource.getConnection();
         iConnCounter.add(conn, pDbName);
//         if (iDebug)
//         {
//            logMsg.append("Connection #" + conn + " opened! DB: " + pDbName);
//            log.debug(Log.TYPE_REBIZ_SYSTEM_LOGS, Log.HANDLER_REBIZ_ADMIN, null,
//                  "DbConnectionFactory: " + logMsg, DbConnectionFactory.class);
//         }
      }
      catch (SQLException e)
      {
         e.printStackTrace();
         throw new RuntimeException( e.getMessage() + " " + pDbName );
      }

      if (conn == null)
      {
         throw new NullPointerException( pDbName );
      }

      return conn;
   }

   /**
    * Schliet eine Connection. Das wird zentral hier gemacht, um besser
    * berwachen zu knnen, welche Connections geffnet und geschlossen wurden.
    * @param pConn	die Connection, die geschlossen werden soll
    * @return	false, wenn es zu einem Fehler kam (z.B. SQLException)
    */
   public static final synchronized boolean close(Connection pConn)
   {
      Statement stats[];

      stats = new Statement[0];

      return close(pConn, stats);
   }

   /**
    * Schliet ein Statement und eine Connection. Das wird zentral hier gemacht, um besser
    * berwachen zu knnen, welche Connections geffnet und geschlossen wurden.
    * @param pConn	die Connection, die geschlossen werden soll
    * @param pStat	ein Statement, das geschlossen werden soll
    * @return	false, wenn es zu einem Fehler kam (z.B. SQLException)
    */
   public static final synchronized boolean close(Connection pConn, Statement pStat)
   {
      Statement stats[];

      stats = new Statement[] { pStat };

      return close(pConn, stats);
   }

   /**
    * Schliet ein Statement und eine Connection. Das wird zentral hier gemacht, um besser
    * berwachen zu knnen, welche Connections geffnet und geschlossen wurden.
    * @param pConn	die Connection, die geschlossen werden soll
    * @param pStat	ein Statement, das geschlossen werden soll
    * @return	false, wenn es zu einem Fehler kam (z.B. SQLException)
    */
   public static final synchronized boolean close(Connection pConn, Statement[] pStats)
   {
      boolean okay = false;

      try
      {
         StringBuffer logMsg = null;

         // Alle Statements dichtmachen
         if (pStats != null && pStats.length > 0)
         {
            for (int cnt = 0; cnt < pStats.length; cnt++)
            {
               if (iDebug)
               {
                  logMsg = new StringBuffer("Closing statement #" + pStats[cnt] + "...");
               }
               
               if( pStats[cnt] != null )
               {
			      pStats[cnt].close();
               }

//               if (iDebug)
//               {
//                  logMsg.append("Statement closed!");
//                  log.debug(Log.TYPE_REBIZ_SYSTEM_LOGS, Log.HANDLER_REBIZ_ADMIN, null,
//                        "DbConnectionFactory: " + logMsg, DbConnectionFactory.class);
//               }
            }
         }

         // Connection dichtmachen
         if (pConn != null)
         {
            if (iDebug)
            {
               logMsg = new StringBuffer("Closing connection #" + pConn + "... DB: "
                     + iConnCounter.getDbName(pConn));
            }

            iConnCounter.remove(pConn);
            pConn.close();

//            if (iDebug)
//            {
//               logMsg.append(" Connection closed!");
//               log.debug(Log.TYPE_REBIZ_SYSTEM_LOGS, Log.HANDLER_REBIZ_ADMIN, null,
//                     "DbConnectionFactory: " + logMsg, DbConnectionFactory.class);
//            }
         }
      }
      catch (Exception ignore)
      {
         return false;
      }

      return okay;
   }

   /**
    * Gibt das Schema zur Datenbank pDnName zurck. Die mglichen Datenbanknamen sind in dieser
    * Klasse als Konstanten definiert. Das Mapping Datenbankname-Schema wird ber den
    * SAMServiceLocatorClient ermittelt.
    * @param pDbName	Name der Datenbank, dessen Schema gesucht wird
    * @return	das Schema der Datenbank (inclusive des Punktes am Ende!) oder ein Leerstring, falls
    * 			    der Datenbankname nicht gefunden wurde oder kein Schemaeintrag fr ihn vorliegt.
    */
//   public static final String getSchema(String pDbName)
//   {
//      //String envName = pDbName + "_SCHEMA";
//      String schema = "NAP0001" ; // SAMServiceLocatorClient.getStringEnvEntry(iJndiQualifier + envName);
//      
//      if ( schema == null )
//      {
//         schema = "";	
//      }
//      else
//      {
//         schema += ".";	
//      }
//      
//      return schema;      
//   }

   /**
    * Gibt das Schema zur Datenbank pDbName zurck. Die mglichen Datenbanknamen sind in dieser
    * Klasse als Konstanten definiert. Das Mapping Datenbankname-Schema wird ber den
    * SAMServiceLocatorClient ermittelt.
    * @param pDbName	Name der Datenbank, dessen Schema gesucht wird
    * @param pSchemaQualifier	gibt an, welches Schema der Datenbank verwendet
    * 		        werden soll. Bsp: Schema "RZO" fr DB-Name DBNAME_NATDB gibt das Schema ("RZO0001") zurck.
    * @return	das Schema der Datenbank (inclusive des Punktes am Ende!) oder ein Leerstring, falls
    * 			    der Datenbankname nicht gefunden wurde oder kein Schemaeintrag fr ihn vorliegt.
    */
//   public static final String getSchema(String pDbName, String pSchemaQualifier)
//   {
//      String envName = pDbName + "_" + pSchemaQualifier;
//      String schema = SAMServiceLocatorClient.getStringEnvEntry(iJndiQualifier + envName);
//      
//      if ( schema == null )
//      {
//         schema = "";	
//      }
//      else
//      {
//         schema += ".";	
//      }
//      
//      return schema;
//   }

   /**
    * Enthlt eine Liste aller momentan geffneten Connections.
    * Speichert zu jeder Connection den Datenbanknamen und den Zeitpunkt des ffnens der Connection.
    */
   private class ConnectionCounter
   {

      /**
       * Speichert die geffneten Connections und deren DB-Namen sowie einen Timestamp
       * (ffnen der Connection bzw. Anlegen dieses Eintrags).
       * Key: 	Connection-Instanz
       * Value:	Object[]
       * 			0	:	Datenbankname, String
       * 			1	:	Open-Timestamp. Long
       */
      HashMap conns;

      /**
       * Speichert die Anzahl geffneter Connections pro Datenbank
       */
      HashMap numConnsPerDB;	// Anzahl der offenen Connections je Datenbank, Key ist DB-Name

      public ConnectionCounter()
      {
         conns = new HashMap();
         numConnsPerDB = new HashMap();
      }

      /**
       * Trgt eine neu geffnete Connection in die Connection-Liste(n) ein
       */
      public void add(Connection pConn, String pDbName)
      {
         Object[] connInfo = new Object[2];

         connInfo[0] = pDbName;
         connInfo[1] = new Long(System.currentTimeMillis());
         conns.put(pConn, connInfo);

         Integer numConns = (Integer) numConnsPerDB.get(pDbName);

         if (numConns == null)
         {
            numConns = new Integer(0);
         }
         else
         {
            numConns = new Integer(numConns.intValue() + 1);
         }

         numConnsPerDB.put(pDbName, numConns);
      }

      /**
       * Lscht eine Connection aus der Connection-Liste
       */
      public void remove(Connection pConn)
      {
         Object[] connInfo = (Object[]) conns.get(pConn);

         // Wenn connInfo null ist, wurde die Connection bereits geschlossen
         if (connInfo != null)
         {
            String dbName = (String) connInfo[0];

            Long openTimestamp = (Long) connInfo[1];
            long diff = System.currentTimeMillis() - openTimestamp.longValue();

            conns.remove(pConn);

            //log.debug(Log.TYPE_REBIZ_SYSTEM_LOGS, Log.HANDLER_REBIZ_ADMIN, null, "600403",
            //      DbConnectionFactory.class, new String[] { dbName, Long.toString(diff) });

            if (dbName != null)
            {
               Integer numConns = (Integer) numConnsPerDB.get(dbName);

               if (numConns != null)
               {
                  numConns = new Integer(numConns.intValue() - 1);
                  numConnsPerDB.put(dbName, numConns);
               }
            }
         }
      } // remove

      /**
       * Gibt die Anzahl aktuell offener Connections zu einer DB zurck.
       */
      public int numConns(String pDbName)
      {
         Integer numConns = (Integer) numConnsPerDB.get(pDbName);

         if (numConns != null)
         {
            return numConns.intValue();
         }

         return 0;
      }

      /**
       * Gibt den Namen der DB zurck, fr die die Connection pConn aufgemacht wurde.
       */
      public String getDbName(Connection pConn)
      {
         Object[] connInfo = (Object[]) conns.get(pConn);
         String dbName = (String) connInfo[0];

         return dbName;
      }
   }
}

