package de.cnc.persistens;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.HashMap;

import de.cnc.util.Log;

/**
 * Diese Klasse dient fr die Key-Generation durch den PersistensHelper.
 * Dieser KeyGenerator holt sich beim ersten Zugriff ber
 * SELECT MAX( %lt;field&gt; ) FROM %lt;table&gt;
 * den maximalen Key von der DB-Tabelle und zhlt von da an nur noch im 
 * Speicher weiter.
 * 
 * @author Heiner Kcker
 * @since 13.09.2004
 */
public class KeyGenerator
{
   //static final String selectStr = "SELECT MAX(?) FROM ?";

   /**
    * Speicher fr vorgemerkte Keys,
    * damit nicht bei jedem Key auf die 
    * Datenbank gegangen werden muss.
    * Ebene Datenbank.
    * @see {@link #keyStep}
    */
   private static HashMap dbMemoryKeyMap = new HashMap();
   
   /**
    * Methode fr Key-Generation durch den PersistensHelper
    * @param pString Tabellen-Name
    * @return der nchste freie Key (Long, BIGINT)
    */
   synchronized public static long getNextKey(
      String pDbName, 
      String pTableName, 
      String pFldName)
   {
      //Abfrage auf DB-Ebene
      HashMap tblKeyHashMap = (HashMap) dbMemoryKeyMap.get( pDbName );
      if (tblKeyHashMap == null)
      {
         tblKeyHashMap = new HashMap();
         dbMemoryKeyMap.put( pDbName, tblKeyHashMap );
      }
      
      //Abfrage auf Tabellen-Ebene
      KeyEntry keyEntry = (KeyEntry) tblKeyHashMap.get( pTableName );
      if (keyEntry == null)
      {
         keyEntry = new KeyEntry();
         tblKeyHashMap.put( pTableName , keyEntry );

         Long lastKeyObj = getMaxFromDbTable(pDbName, pTableName, pFldName);
         if (lastKeyObj != null)
         {
            //es gibt schon Datenstze
            keyEntry.lastKey = lastKeyObj.longValue();
         }
      }
      long nextKey = ++keyEntry.lastKey;
      return nextKey;
   }

   /**
    * Diese Klasse trgt die im Speicher gehaltenen
    * Keys.
    * 
    * @author Heiner Kcker
    */
   private static class KeyEntry
   {
      /** 
       * letzter vergebener Key 
       * Zhlung beginnt mit 1.
       */
      long lastKey = 0;
   }

   /**
    * Abfrage des maximalen Keys von der Datenbank.
    * Wenn unbekannt, wird null zurckgegeben.
    * 
    * @param dbName
    * @param tableName
    * @param genKeyFldName
    */
   private static synchronized Long getMaxFromDbTable(
      final String pDbName, 
      final String pTableName, 
      final String pFldName)
   {
      Log.debug( "Keygenerator#getMaxFromDbTable" );

      //UserTransaction userTransaction = null;
      java.sql.Connection con = null;

      try
      {
         //final InitialContext ctx = new InitialContext();
         //userTransaction = (UserTransaction) ctx.lookup("java:comp/UserTransaction");
         //userTransaction.begin();
         
         // Vorgehensweise Tomcat: con = DatabaseConnectionFactory.getConnection( pDbName );
         con = DbConnectionFactory.makeDBConnection( pDbName );

         String selectStr = "SELECT MAX(" + pFldName + ") FROM " + pTableName;

Log.debug( selectStr );

         PreparedStatement myPreparedStatement = con.prepareStatement(selectStr);
         //myPreparedStatement.setString(1, pFldName);
         //myPreparedStatement.setString(2, pTableName);
         ResultSet rsltSet = myPreparedStatement.executeQuery();

         if (rsltSet != null && rsltSet.next())
         {
            long maxKey = rsltSet.getLong(1);
            Log.debug("lastKey gelesen " + maxKey);
            if (rsltSet.wasNull())
            {
               return null;
            }
            return new Long(maxKey);
            
         }
         else
         {
            //erstes Anlegen des Key fr diese Tabelle
            return null;
         }
      }
      catch (Exception e)
      {
         Log.fatal(e);

         return null;
      }
      finally
      {
         //try
         //{
         //   userTransaction.commit();
         //}
         //catch (Exception e1)
         //{
         //   Log.fatal( e1 );
         //}
         if (con != null)
         {
            try
            {
               con.close();
            }
            catch (Exception e)
            {
               Log.fatal(e);
               return null;
            }
         }
      }

   }

}
