package de.cnc.persistens;

import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Hashtable;

import de.cnc.CncContext;
import de.cnc.util.*;
import de.cnc.util.config.ConfigManager;

/**
 * Klasse fr das Bean-zu-Datenbank-Mapping . <br/>
 * Liest das Mapping aus der Datei <default-app-props>.properties und
 * speichert es im Cache {@link #databaseDefinitionAndMethodsCache} . <br/>
 * Das gecachte Mapping wird zum Erzeugen der SQL-Statements und <br/>
 * zum Umkopieren der Daten von den JDBC-Result-Sets in die Beans verwendet. <br/>
 * Achtung, Achtung: Boolean-Properties werden in SMALLINT-Feldern gespeichert, <br/>
 * wobei als Feldtyp Boolean angegeben wird. <br/>
 * z.B.:
 * <pre>
 *                 Feldname   Map-Typ   Property
 * databasefield = STATUS     Boolean   status
 *
 * Feld:      STATUS SMALLINT
 * Property:  status boolean
 * </pre>
 *
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner Kcker</a>
 * <br/>
 *
 * @author Heiner Kcker
 * @version $Id: PersistensHelper.java,v 1.3 2004/10/05 22:50:19 aex116e Exp $
 */
public class PersistensHelper
{
   /**
    * Hashtable zum Cachen der Datenbankinformationen und Methoden fr die jeweiligen ActionFormBeans.<br/>
    */
   private static Hashtable databaseDefinitionAndMethodsCache = new Hashtable();

   /**
    * Array mit den zulssigen Feldtypen
    */
   private static final String[] FIELD_TYPES =
      { "Boolean", "Integer", "Long", "Double", "BigDecimal", "String", "Date", "Timestamp" };

   /**
    * Constructor declaration<br/>
    * Instanziieren nicht erlaubt
    */
   private PersistensHelper()
   {
      ;
   } // end constructor

   /**
    * Init
    */
   public static void init()
   {
      initDatabaseDefinitionAndMethods();
   }

   /**
    * Hilfsmethode zum Liefern eines Class-Objekts entsprechend dem bergebenen Field-Type-Namen
    * (Boolean|Integer|Long|Double|String|Date|Timestamp) . <br/>
    * Wird verwendet in {@link #initDatabaseDefinitionAndMethods initDatabaseDefinitionAndMethods}.<br/>
    *
    * @param strTyp
    * @return entsprechende Klasse
    *
    * @throws Exception wenn Typ nicht verwertbar
    */
   private static final Class getParamClass(String strTyp)
   {
      // Boolean|Integer|Long|Double|BigDecimal|String|Date|Timestamp
      Class retClass = null;

      if ("Boolean".equals(strTyp))
      {
         retClass = Boolean.class;
      }
      else if ("Integer".equals(strTyp))
      {
         retClass = Integer.class;
      }
      else if ("Long".equals(strTyp))
      {
         retClass = Long.class;
      }
      else if ("Double".equals(strTyp))
      {
         retClass = Double.class;
      }
      else if ("BigDecimal".equals(strTyp))
      {
         retClass = BigDecimal.class;
      }
      else if ("String".equals(strTyp))
      {
         retClass = String.class;
      }
      else if ("Date".equals(strTyp))
      {
         retClass = java.sql.Date.class;
      }
      else if ("Timestamp".equals(strTyp))
      {
         retClass = java.sql.Timestamp.class;
      }
      else
      {
         throw new RuntimeException("Method getParamClass: Class " + strTyp + " not available for DatabaseObject");
      }

      return retClass;
   }

   /**
    * Umladen der Felder aus dem ResultSet in die Bean-Properties der Kindklasse unter Verwendung des Methoden-Caches.
    *
    * @param rs ResultSet aus SQL-Select-Statement
    *
    * @exception  alles was da kommt und auf falsche Einstellungen in <b><default-app-props>.properties</b> zurckgeht, wird weitergeworfen
    */
   public static final void readResultSet(//--
   final IStorable databaseBean, //--
   final ResultSet rs)
   {
      try
      {
      DbTableDefinitionAndMethods databaseDefinitionAndMethods = getDatabaseDefinitionAndMethods(databaseBean);

      for (int i = 0; i < databaseDefinitionAndMethods.fieldArr.length; i++)
      {
         DbFieldDefinitionAndMethods oneField = databaseDefinitionAndMethods.fieldArr[i];
         Object valueObject = null;

         //         Log.debug("setResultSet");
         //         Log.debug("valueObject: " + valueObject);
         //         if (valueObject != null)
         //         {
         //            Log.debug("valueObject.class: " + valueObject.getClass().getName());
         //         }
         //         Log.debug("feldName: " + oneField.fieldName);

//@todo: hier einen switch-Verteiler machen, int-Konstanten und Werte in DbFieldDefinitionAndMethods hinterlegen

         // Feld-Typ (Boolean|Integer|Long|Double|BigDecimal|String|Date|Timestamp)
         if ("Boolean".equals(oneField.fieldTyp))
         {
            int tmpInt = rs.getInt(oneField.fieldName);
            if ( rs.wasNull() )
            {
               valueObject = null;
            }
            else
            {
               // Sonderbehandlung bei Boolean, Feldtyp SMALLINT umwandeln in Boolean
               valueObject = StringUtil.numStr2boolObject("" + tmpInt);
            }
         }
         else if ("String".equals(oneField.fieldTyp))
         {
            if (oneField.trimByRead)
            {
               valueObject = StringUtil.trim( rs.getString(oneField.fieldName ) );
            }
            else
            {
               valueObject = rs.getString(oneField.fieldName);
            }
         }
         else if ("Date".equals(oneField.fieldTyp))
         {
            valueObject = rs.getDate(oneField.fieldName);
         }
         else if ("Timestamp".equals(oneField.fieldTyp))
         {
            valueObject = rs.getTimestamp(oneField.fieldName);
         }
         else if ("Integer".equals(oneField.fieldTyp))
         {
//@todo: Umwandelbare Typen aus Buch Wolfgang Dehnhardt, Anwendungsprogrammierung mit JDBC hier eintragen            
            //wegen Typ-Problematik umwandeln in String
            String tmpStr = rs.getString(oneField.fieldName);
            if ( ! rs.wasNull() )
            {
               valueObject = StringUtil.str2IntegerStrict( tmpStr );
            }
         }
         else if ("Long".equals(oneField.fieldTyp))
         {
            //wegen Typ-Problematik umwandeln in String
            String tmpStr = rs.getString(oneField.fieldName);
            if ( ! rs.wasNull() )
            {
               valueObject = StringUtil.str2LongStrict( tmpStr );
            }
         }
         else if ("Double".equals(oneField.fieldTyp))
         {
            //wegen Typ-Problematik umwandeln in String
            String tmpStr = rs.getString(oneField.fieldName);
            if ( ! rs.wasNull() )
            {
               valueObject = StringUtil.str2Double( tmpStr );
            }
         }
         else if ("BigDecimal".equals(oneField.fieldTyp))
         {
            //wegen Typ-Problematik umwandeln in String
            String tmpStr = rs.getString(oneField.fieldName);
            if ( ! rs.wasNull() )
            {
               valueObject = StringUtil.str2BigDecimalStrict( tmpStr );
            }
         }
         //else // irgendetwas anderes
         //{
         //   valueObject = rs.getObject(oneField.fieldName);
         //}

         //      if (valueObject == null)
         //      {
         //        // alle Nullpointer wegfangen
         //        // null in Vorbelegungswert umwandeln
         //        valueObject = ClassBuilder.buildSql(databaseDefinitionAndMethods.feldTypArr[i]);
         //      }

         // Feld-Typ (Boolean|Integer|Long|Double|BigDecimal|String|Date|Timestamp)
         // set-Method holen
         Method setMethod = oneField.setMethod;

         //Log.debug("setMethod: " + setMethod);

         //         if (setMethod == null)
         //         {
         //            Log.debug("setMethod is null");
         //         }

         //      if (valueObject != null)
         //      {
         //         // Prfen auf Zuweisungskompatibilitt
         //         // wenn nicht zuweisungskompatibel aber beide vom Type Number, dann passend umwandeln
         //         Class[] classArr = setMethod.getParameterTypes();
         //         if (! valueObject instanceof classArr[0])
         //      }

         MethodUtil.invoke( setMethod, databaseBean, valueObject);
      } //end for

      databaseBean.setChanged(false);
   }
      catch (SQLException e)
      {
         Log.fatal( e );
         throw new RuntimeException( e.getMessage() );
      }
   } //end method setResultSet

   /**
    * Erzeugen DELETE-Statement fr Parameter-Bean
    *
    * @param databaseBean Parameter-Bean, auf die sich das SQL-Statement bezieht
    * @return Statement-String
    *
    * @throws DatabaseStoreException
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in <default-app-props>.properties zurckgeht, wird weitergeworfen
    */
   public static final String getDeleteSQLStatement(IStorable databaseBean)
   {
      StringBuffer sb = new StringBuffer();

      sb.append("DELETE FROM " + getTableName(databaseBean));
      sb.append(" WHERE " + getPrimaryKeyFieldName(databaseBean) + "=");
      sb.append(databaseBean.getObjectId());

      // Log.debugOut (sb.toString ());
      return sb.toString();
   }

   /**
    * Erzeugen DELETE-Statement fr alle Records der Tabelle
    *
    * @param databaseBean Parameter-Bean, auf die sich das SQL-Statement bezieht
    * @return Statement-String
    *
    * @throws DatabaseStoreException
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in <default-app-props>.properties zurckgeht, wird weitergeworfen
    */
   public static final String getDeleteAllSQLStatement(IStorable databaseBean)
   {
      StringBuffer sb = new StringBuffer();

      sb.append("DELETE FROM " + getTableName(databaseBean));

      // Log.debugOut (sb.toString ());
      return sb.toString();
   }

   /**
    * Insert-Statement ohne ID-Feld holen fr als Parameter bergebene Bean
    *
    * @param databaseBean Parameter-Bean, auf die sich das SQL-Statement bezieht
    * @return Statement-String
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in <default-app-props>.properties zurckgeht, wird weitergeworfen
    */
   //  public static final String getInsertSQLStatement(IStorable databaseBean)
   //    throws Exception
   //  {
   //    return getInsertSQLStatement(databaseBean, false);
   //  }

   /**
    * Insert-Statement mit oder ohne ID-Feld holen fr Parameter-Bean.<br/>
    *
    * Bei Neuanlegen Datensatz wird normalerweise der Datenbank die Vergabe des Schlssels berlassen, <br/>
    * aber beim SQL-Dump {@link #writeSQLDumpToFile} werden die IDs eingebunden. <br/>
    * Ist z.B. ntig bei Konditionen, da der Zugriff auf die Konditionen ber die ID erfolgt und diese in Bonn und Berlin bereinstimmen muss. <br/>
    *
    * @param databaseBean Parameter-Bean, auf die sich das SQL-Statement bezieht
    * @param bPaIncludeId ob die objectID (meist Feld ID) in das Insert-Statement mit aufgenommen werden soll(ja/nein). <br/>
    * @return Statement-String
    *
    * @throws Exception alles was da kommt und auf falsche Einstellungen in <default-app-props>.properties zurckgeht, wird weitergeworfen
    */
   public static final String getInsertSQLStatement(IStorable databaseBean, CncContext cncContext
   /*, boolean bPaIncludeId*/
   )
   {
      // Mimik fr Key-Erzeugung in Server, statt in Datenbank, komplett entfernt
      StringBuffer sb = new StringBuffer();

      sb.append("INSERT INTO " + getTableName(databaseBean));

      StringBuffer[] sba = null;
      DbTableDefinitionAndMethods databaseDefinitionAndMethods = getDatabaseDefinitionAndMethods(databaseBean);

      for (int i = 0; i < databaseDefinitionAndMethods.fieldArr.length; i++)
      {
         DbFieldDefinitionAndMethods oneField = databaseDefinitionAndMethods.fieldArr[i];
         //      // beim Insert mit automatischer Key-Vergabe nicht das Key-Field mit zum INSERT-Statement hinzufgen
         //      // aber bei SQL-Dump mitschicken
         //      if ((bPaIncludeId) ||!getPrimaryKeyFieldName(databaseBean).equals(databaseDefinitionAndMethods.feldNameArr[i]))
         //      {
         //        Object[]  paramArr = new Object[0]; // leeres Objekt-Array fr die Parameter-Liste des invoke-Aufrufs der Get-Methode
         //        Object    valueObject = (databaseDefinitionAndMethods.getMethodArr[i]).invoke(databaseBean, paramArr);
         //
         ////        // Timestamp auf aktuelle Zeit setzen, aber nicht wenn auch die ID mitgeliefert wird
         ////        if ((!bPaIncludeId) && "TIMESTAMP".equals(databaseDefinitionAndMethods.feldNameArr[i]))
         //        if ("ANLAGE_ZEITPUNKT".equals(databaseDefinitionAndMethods.feldNameArr[i]))
         //        {
         //          // hier wird vorausgesetzt, da bei Feldname ANLAGE_ZEITPUNKT auch der Feldtyp TIMESTAMP auftritt und dieses Feld als ANLAGE_ZEITPUNKT benutzt wird
         //          // Timestamp auf aktuelle Zeit setzen
         //          valueObject = new java.sql.Timestamp((new java.util.Date()).getTime());
         //        }
         //        if ("AENDER_ZEITPUNKT".equals(databaseDefinitionAndMethods.feldNameArr[i]))
         //        {
         //          // hier wird vorausgesetzt, da bei Feldname AENDER_ZEITPUNKT auch der Feldtyp TIMESTAMP auftritt und dieses Feld als AENDER_ZEITPUNKT benutzt wird
         //          // Timestamp auf aktuelle Zeit setzen
         //          valueObject = new java.sql.Timestamp((new java.util.Date()).getTime());
         //        }
         //
         //        // Boolean|Integer|Long|Double|String|Date|Timestamp
         //        sba = SqlUtil.appendInsertKlausel(sba, databaseDefinitionAndMethods.feldNameArr[i], valueObject);
         //      }

         Object valueObject = null;

         if (oneField.readOnly)
         {
            ; //dieses Feld nur Lesen, niemals Schreiben
         }
         else if (oneField.currentTimeStampOnInsert)
         {
            // hier wird vorausgesetzt, da bei Feldname AENDER_ZEITPUNKT auch der Feldtyp TIMESTAMP auftritt und dieses Feld als AENDER_ZEITPUNKT benutzt wird
            // Timestamp auf aktuelle Zeit setzen
            //valueObject = new java.sql.Timestamp((new java.util.Date()).getTime());
            //Rckspeichern gesetzten Wert
            //MethodUtil.invoke( oneField.setMethod, databaseBean, valueObject );

            valueObject = SqlUtil.CURRENT_TIMESTAMP_JOKER;

            // Boolean|Integer|Long|Double|String|Date|Timestamp|BigDecimal
            sba = SqlUtil.appendInsertKlausel(sba, oneField.fieldName, valueObject);
         }
         else if (oneField.currentUserOnInsert)
         {
            valueObject = cncContext.getUserId();

            //Rckspeichern gesetzten Wert
            MethodUtil.invoke( oneField.setMethod, databaseBean, valueObject );

            // Boolean|Integer|Long|Double|String|Date|Timestamp|BigDecimal
            sba = SqlUtil.appendInsertKlausel(sba, oneField.fieldName, valueObject);
         }
         else if (oneField.generateKey)
         {
            valueObject =
               new Long(
                  KeyGenerator.getNextKey(
                     databaseDefinitionAndMethods.dbName,
                     databaseDefinitionAndMethods.tableName,
                     oneField.fieldName));

            //Rckspeichern gesetzten Wert
            MethodUtil.invoke( oneField.setMethod, databaseBean, valueObject);

            sba = SqlUtil.appendInsertKlausel(sba, oneField.fieldName, valueObject);
         }
         else if (oneField.insertValue != null)
         {
            valueObject = oneField.insertValue;

            //Rckspeichern gesetzten Wert
            MethodUtil.invoke( oneField.setMethod , databaseBean, valueObject );

            sba = SqlUtil.appendInsertKlausel(sba, oneField.fieldName, valueObject);
         }
         else
         {
            //default-Fall, Wert aus Bean die Datensatz speichern
            valueObject = MethodUtil.invoke( oneField.getMethod, databaseBean);

            // Boolean|Integer|Long|Double|String|Date|Timestamp|BigDecimal
            sba = SqlUtil.appendInsertKlausel(sba, oneField.fieldName, valueObject);
         }
      }

      sb.append(SqlUtil.completeInsertKlausel(sba));

      // Log.debugOut (sb.toString ());
      return sb.toString();
   }

   /**
    * Zurckgeben SQL-Statement SELECT FOR UPDATE . <br/>
    * Wird fr Record-Locking bentigt
    *
    * @param databaseBean Parameter-Bean, auf die sich das SQL-Statement bezieht
    * @return Statement-String
    *
    * @throws DatabaseStoreException
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in <default-app-props>.properties zurckgeht, wird weitergeworfen
    */
   public static final String getSelectForUpdateSQLStatement(IStorable databaseBean)
      throws DatabaseStoreException, Exception
   {
      return getDatabaseDefinitionAndMethods(databaseBean).selectForUpdateSQLStatement;
   } // end method getSelectForUpdateSQLStatement

   /**
    * Erzeugen SQL-Statement SELECT FOR UPDATE . <br/>
    * Wird fr Record-Locking bentigt
    *
    * @param databaseBean Parameter-Bean, auf die sich das SQL-Statement bezieht
    * @return Statement-String
    *
    * @throws DatabaseStoreException
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in <default-app-props>.properties zurckgeht, wird weitergeworfen
    */
   private static final String createSelectForUpdateSQLStatement(String pTableName, String pPrimaryKeyFieldName)
   {
      StringBuffer sb = new StringBuffer();

      sb.append("SELECT * FROM " + pTableName + " WHERE ");
      sb.append(" " + pPrimaryKeyFieldName + "=");
      sb.append("?"); // sb.append(databaseBean.getObjectId());
      sb.append(" FOR UPDATE");

      // Log.debugOut (sb.toString ());
      return sb.toString();
   } // end method getSelectForUpdateSQLStatement

   /**
    * Erzeugen UPDATE-SQL-Statement
    *
    * @param databaseBean Parameter-Bean, auf die sich das SQL-Statement bezieht
    * @return Statement-String
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in <default-app-props>.properties zurckgeht, wird weitergeworfen
    */
   public static final String getUpdateSQLStatement(IStorable databaseBean, CncContext cncContext)
   {
      StringBuffer sbu = new StringBuffer();
      DbTableDefinitionAndMethods databaseDefinitionAndMethods = getDatabaseDefinitionAndMethods(databaseBean);

      for (int i = 0; i < databaseDefinitionAndMethods.fieldArr.length; i++)
      {
         DbFieldDefinitionAndMethods oneField = databaseDefinitionAndMethods.fieldArr[i];
         //      // beim Insert mit automatischer Key-Vergabe nicht das Key-Field mit zum INSERT-Statement hinzufgen
         //      if (!getPrimaryKeyFieldName(databaseBean).equals(oneField.fieldName))
         //      {
         //        Object  valueObject = (oneField.getMethod).invoke(databaseBean, null);
         //
         //        if ("AENDER_ZEITPUNKT".equals(oneField.fieldName))
         //        {
         //          // hier wird vorausgesetzt, da bei Feldname AENDER_ZEITPUNKT auch der Feldtyp TIMESTAMP auftritt und dieses Feld als AENDER_ZEITPUNKT benutzt wird
         //          // Timestamp auf aktuelle Zeit setzen
         //          valueObject = new java.sql.Timestamp((new java.util.Date()).getTime());
         //        }
         //
         //        // Boolean|Integer|Long|Double|String|Date|Timestamp
         //        SqlUtil.appendUpdateSetKlausel(sb1, oneField.fieldName, valueObject);
         //      }

         Object valueObject = null;

         if (oneField.readOnly)
         {
            ; //dieses Feld nur Lesen, niemals Schreiben
         }
         else if (oneField.noUpdate)
         {
            ; //dieses Feld nur beim Insert beschreiben, niemals beim Update
         }
         else if (oneField.currentTimeStampOnUpdate)
         {
            // hier wird vorausgesetzt, da bei Feldname AENDER_ZEITPUNKT auch der Feldtyp TIMESTAMP auftritt und dieses Feld als AENDER_ZEITPUNKT benutzt wird
            // Timestamp auf aktuelle Zeit setzen
            //valueObject = new java.sql.Timestamp((new java.util.Date()).getTime());
            //Rckspeichern gesetzten Wert
            //MethodUtil.invoke( oneField.setMethod, databaseBean, valueObject );

            valueObject = SqlUtil.CURRENT_TIMESTAMP_JOKER;

            // Boolean|Integer|Long|Double|String|Date|Timestamp|BigDecimal
            sbu = SqlUtil.appendUpdateSetKlausel(sbu, oneField.fieldName, valueObject);
         }
         else if (oneField.currentTimeStampOnInsert)
         {
            ; // dirty hack, nur bei INSERT beschreiben
         }
         else if (oneField.currentUserOnUpdate)
         {
            // hier wird vorausgesetzt, da bei Feldname AENDER_ZEITPUNKT auch der Feldtyp TIMESTAMP auftritt und dieses Feld als AENDER_ZEITPUNKT benutzt wird
            // Timestamp auf aktuelle Zeit setzen
            //valueObject = new java.sql.Timestamp((new java.util.Date()).getTime());
            valueObject = cncContext.getUserId();

            //Rckspeichern gesetzten Wert
            MethodUtil.invoke( oneField.setMethod, databaseBean, valueObject );

            // Boolean|Integer|Long|Double|String|Date|Timestamp|BigDecimal
            sbu = SqlUtil.appendUpdateSetKlausel(sbu, oneField.fieldName, valueObject);
         }
         else if (oneField.updateValue != null)
         {
            valueObject = oneField.updateValue;

            //Rckspeichern gesetzten Wert
            MethodUtil.invoke( oneField.setMethod, databaseBean, valueObject );

            sbu = SqlUtil.appendUpdateSetKlausel(sbu, oneField.fieldName, valueObject);
         }
         else
         {
            //default-Fall, Wert aus Bean die Datensatz speichern
            valueObject = MethodUtil.invoke( oneField.getMethod, databaseBean);

            // Boolean|Integer|Long|Double|String|Date|Timestamp|BigDecimal
            sbu = SqlUtil.appendUpdateSetKlausel(sbu, oneField.fieldName, valueObject);
         }
      }

      StringBuffer sb = new StringBuffer();

      sb.append("UPDATE " + getTableName(databaseBean) + " SET ");
      sb.append(sbu);
      sb.append(" WHERE " + getPrimaryKeyFieldName(databaseBean) + "=");
      sb.append(databaseBean.getObjectId());

      // Log.debugOut (sb.toString ());
      return sb.toString();
   } // end method getUpdateSQLStatement

   /**
    * Erzeugen WHERE-Klausel zum Selectieren des Datensatzes, entsprechend Parameter-Bean
    *
    * @param databaseBean Parameter-Bean, auf die sich das SQL-Statement bezieht
    * @return Statement-String
    *
    * @throws DatabaseStoreException
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final String getSelectIdWhereKlausel(IStorable databaseBean)
   {
      StringBuffer sb = new StringBuffer();

      sb.append(getPrimaryKeyFieldName(databaseBean) + "=");
      sb.append(databaseBean.getObjectId());

      sb.append(" FETCH FIRST ROW ONLY");

      // Log.debugOut (sb.toString ());
      return sb.toString();
   } // end method getSelectForUpdateSQLStatement

   /**
    * Holen des Bean-zu-Datenbank-Mappings mit Methoden-Cache fr die Parameter-Bean
    *
    * @param databaseBean Parameter-Bean
    * @return entsprechendes Mapping mit Methoden-Cache
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   private static final DbTableDefinitionAndMethods getDatabaseDefinitionAndMethods(IStorable databaseBean)
   {
      Class beanClass = databaseBean.getClass();
      DbTableDefinitionAndMethods databaseDefinitionAndMethods = getDatabaseDefinitionAndMethods(beanClass);
      return databaseDefinitionAndMethods;
   }

   private static DbTableDefinitionAndMethods getDatabaseDefinitionAndMethods(Class beanClass) 
   {
      DbTableDefinitionAndMethods databaseDefinitionAndMethods =
         (DbTableDefinitionAndMethods) databaseDefinitionAndMethodsCache.get(beanClass);
      
      if (databaseDefinitionAndMethods == null)
      {
         throw new RuntimeException("No database definition found for class " + beanClass.getName());
      }
      
      return databaseDefinitionAndMethods;
   }

   /**
    * Holen Namen der Datenbank (bei Betrieb mit mehreren Datenbanken)
    * fr die jeweilige Parameter-Bean.
    *
    * @param databaseBean Parameter-Bean
    * @return Tabellen-Name
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in <default-app-props>.properties zurckgeht, wird weitergeworfen
    */
   public static final String getDbName(IStorable databaseBean)
   {
      return getDatabaseDefinitionAndMethods(databaseBean).dbName;
   }

   /**
    * Holen Namen der Datenbank-Tabelle fr die jeweilige Parameter-Bean
    *
    * @param databaseBean Parameter-Bean
    * @return Tabellen-Name
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in <default-app-props>.properties zurckgeht, wird weitergeworfen
    */
   public static final String getTableName(IStorable databaseBean)
   {
      return getDatabaseDefinitionAndMethods(databaseBean).tableName;
   }

   /**
    * Holen Namen der Datenbank-Tabelle fr die jeweilige Parameter-Bean
    *
    * @param databaseBean Parameter-Bean
    * @return Tabellen-Name
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in <default-app-props>.properties zurckgeht, wird weitergeworfen
    */
   public static final String getTableName(Class pClass)
   {
      return getDatabaseDefinitionAndMethods(pClass).tableName;
   }

   /**
    * Holen Namen des Feldes, welches der Primrschlssel ist, fr die jeweilige Parameter-Bean
    *
    * @param databaseBean Parameter-Bean
    * @return Feldname Primery-Key
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final String getPrimaryKeyFieldName(IStorable databaseBean)
   {
      return getDatabaseDefinitionAndMethods(databaseBean).primaryKeyFieldName;
   }

   /**
    * Holen Namen des Feldes, welches der Primrschlssel ist, fr die jeweilige Parameter-Bean
    *
    * @param databaseBean Parameter-Bean
    * @return Feldname Primery-Key
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final String getPrimaryKeyFieldName(Class pClass)
   {
      return getDatabaseDefinitionAndMethods(pClass).primaryKeyFieldName;
   }

   /**
    * Holen der Namen der Felder, welche Sekundrschlssel sind, fr die jeweilige Parameter-Bean
    *
    * @param databaseBean Parameter-Bean
    * @return Feldnamen-Array Foreign-Keys
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final String[] getForeignKeyFieldNameArr(IStorable databaseBean)
   {
      return getDatabaseDefinitionAndMethods(databaseBean).foreignKeyFieldNameArr;
   }

   /**
    * Holen der Namen der Felder, welche Sekundrschlssel sind, fr die jeweilige Parameter-Bean
    *
    * @param databaseBean Parameter-Bean
    * @return Feldnamen-Array Foreign-Keys
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final String[] getForeignKeyFieldNameArr(Class pClass)
   {
      return getDatabaseDefinitionAndMethods(pClass).foreignKeyFieldNameArr;
   }

   /**
    * Holen der Namen der Felder, welche Sekundrschlssel sind, fr die jeweilige Parameter-Bean
    *
    * @param databaseBean Parameter-Bean
    * @return Feldnamen-Array Foreign-Keys
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final String[] getForeignKeyPropertyNameArr(IStorable databaseBean)
   {
      return getDatabaseDefinitionAndMethods(databaseBean).foreignKeyPropertyNameArr;
   }

   /**
    * Holen der Namen der Felder, welche Sekundrschlssel sind, fr die jeweilige Parameter-Bean
    *
    * @param databaseBean Parameter-Bean
    * @return Feldnamen-Array Foreign-Keys
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final String[] getForeignKeyPropertyNameArr(Class pClass)
   {
      return getDatabaseDefinitionAndMethods(pClass).foreignKeyPropertyNameArr;
   }

   /**
    * ob ein cache benutzt wird.
    *
    * @return maximale Trefferanzahl
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final boolean cacheUse(IStorable databaseBean)
   {
      return getDatabaseDefinitionAndMethods(databaseBean).cacheUse;
   }

   /**
    * ob ein cache benutzt wird.
    *
    * @return maximale Trefferanzahl
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final boolean cacheUse(Class pClass)
   {
      return getDatabaseDefinitionAndMethods(pClass).cacheUse;
   }

   /**
    * Maximale Treffer-Anzahl bei SELECT.
    * Wird realisiert ber "FETCH FIRST &gt;num&lt; ONLY".
    * 0 oder kleiner bedeutet keine Angabe.
    *
    * @return maximale Trefferanzahl
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final int getMaxFetch(IStorable databaseBean)
   {
      return getDatabaseDefinitionAndMethods(databaseBean).maxFetch;
   }

   /**
    * Maximale Treffer-Anzahl bei SELECT.
    * Wird realisiert ber "FETCH FIRST &gt;num&lt; ONLY".
    * 0 oder kleiner bedeutet keine Angabe.
    *
    * @return maximale Trefferanzahl
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final int getMaxFetch(Class pClass)
   {
      return getDatabaseDefinitionAndMethods(pClass).maxFetch;
   }

   /**
    * Feldliste fr SELECT
    *
    * @return 
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final String getSelectFieldsStr(IStorable databaseBean)
   {
      return getDatabaseDefinitionAndMethods(databaseBean).selectFieldsStr;
   }

   /**
    * Feldliste fr SELECT
    *
    * @return 
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final String getQualifiedSelectFieldsStr(Class pClass)
   {
      return getDatabaseDefinitionAndMethods(pClass).qualifiedSelectFieldsStr;
   }

   /**
    * Feldliste fr SELECT
    *
    * @return 
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final String getQualifiedSelectFieldsStr(IStorable databaseBean)
   {
      return getDatabaseDefinitionAndMethods(databaseBean).qualifiedSelectFieldsStr;
   }

   /**
    * Feldliste fr SELECT
    *
    * @return 
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in xxx.properties zurckgeht, wird weitergeworfen
    */
   public static final String getSelectFieldsStr(Class pClass)
   {
      return getDatabaseDefinitionAndMethods(pClass).selectFieldsStr;
   }

   /**
    * Holen Timestamp der Parameter-Bean, nicht des entsprechenden Datensatzes
    *
    * @param databaseBean Parameter-Bean
    * @return Timestamp oder null, wenn kein TIMESTAMP-Feld vorhanden
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in <default-app-props>.properties zurckgeht, wird weitergeworfen
    */
   //  public static final Timestamp getTimestamp(IStorable databaseBean)
   //   
   //  {
   //    DbTableDefinitionAndMethods  DbTableDefinitionAndMethods = getDatabaseDefinitionAndMethods(databaseBean);
   //
   //    if (DbTableDefinitionAndMethods.timestampIndex < 0)
   //    {
   //      // kein TIMESTAMP-Feld vorhanden
   //      return null;
   //    }
   //
   //    DbTableDefinitionAndMethods  databaseDefinitionAndMethods = getDatabaseDefinitionAndMethods(databaseBean);
   //    int                           index = databaseDefinitionAndMethods.timestampIndex;
   //    Object                        valueObject = (databaseDefinitionAndMethods.getMethodArr[index]).invoke(databaseBean, null);
   //    Timestamp                     retTimestamp = null;
   //
   //    if (valueObject instanceof Timestamp)
   //    {
   //      retTimestamp = (Timestamp) valueObject;
   //    }
   //
   //    return retTimestamp;
   //  } // end method getTimestamp

   /**
    * Initialisieren des Bean-zu-Datenbank-Mappings mit Methoden-Cache fr die Parameter-Bean . <br/>
    * Dazu wird die Konfigurationsdatei <default-app-props>.properties als String-Array eingelesen und eingeparst. <br/>
    *
    *
    * @param databaseBean Parameter-Bean
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in <default-app-props>.properties zurckgeht, wird weitergeworfen
    */
   private static final synchronized void initDatabaseDefinitionAndMethods()
   {
      Log.debug("Parsing database definitions from file" , ConfigManager.getDefaultConfigFileName() + ".properties");

      //Einlesen <defaultconfigfile>.properties als String[]
      String[] propStrArr = ConfigManager.getPropertiesFileAsStringArr(ConfigManager.getDefaultConfigFileName());
      //String[] propStrArr = StrConst.PROP;

      int zeiPtr = 0; // Zeilenzeiger

      while (zeiPtr < propStrArr.length)
      {
         zeiPtr = stepNextAndOverCommentOrEmptyLines(propStrArr, zeiPtr);

         // Vormarschieren bis databasebeanclass = de.cnc.persistens.xxxActionForm
         String suchString = "databasebeanclass=";
         while (zeiPtr < propStrArr.length && (!(StringUtil.stripTabSpace(propStrArr[zeiPtr])).startsWith(suchString)))
         {
            zeiPtr++;
         }

         if (zeiPtr >= propStrArr.length)
         {
            break;
         }
         
         
         // Die Bean-Klasse ermitteln
         final String className = cutPropertyName("databasebeanclass", propStrArr[zeiPtr]);

         Log.debug( "line" , "" + zeiPtr );
         Log.debug( "databasebeanclass" , className );

         final Class beanClass;
         try
         {
            beanClass = Class.forName(className);
         }
         catch (ClassNotFoundException e)
         {
            throw new RuntimeException("class " + className + " not found , line " + (zeiPtr + 1));
         }

         //final IStorable databaseBean = (IStorable) beanClass.newInstance();

         final DbTableDefinitionAndMethods dbTableDefinitionAndMethods = new DbTableDefinitionAndMethods();

         zeiPtr = stepNextAndOverCommentOrEmptyLines(propStrArr, zeiPtr);

         // databasename=
         suchString = "databasename";

         checkForPropertyName( suchString, propStrArr, zeiPtr );

         String dbName = cutPropertyName(suchString, propStrArr[zeiPtr]);

Log.debug( "dbName" , dbName );

         dbTableDefinitionAndMethods.dbName = dbName;

         zeiPtr = stepNextAndOverCommentOrEmptyLines(propStrArr, zeiPtr);

         // databasetable=
         suchString = "databasetable";

         checkForPropertyName( suchString, propStrArr, zeiPtr );

         String tableName = cutPropertyName(suchString, propStrArr[zeiPtr]);

Log.debug( "tableName" , tableName );

         dbTableDefinitionAndMethods.tableName = tableName;

         zeiPtr = stepNextAndOverCommentOrEmptyLines(propStrArr, zeiPtr);

         // databaseprimerykeyfield=
         suchString = "databaseprimerykeyfield";

         checkForPropertyName( suchString, propStrArr, zeiPtr );

         String primeryKeyField = cutPropertyName(suchString, propStrArr[zeiPtr]);

Log.debug( "primeryKeyField" , primeryKeyField );

         dbTableDefinitionAndMethods.primaryKeyFieldName = primeryKeyField;

         zeiPtr = stepNextAndOverCommentOrEmptyLines(propStrArr, zeiPtr);

         // maxfetch=
         suchString = "maxfetch";

         checkForPropertyName( suchString, propStrArr, zeiPtr );

         String maxfetchStr = cutPropertyName(suchString, propStrArr[zeiPtr]);

Log.debug( "maxfetchStr" , maxfetchStr );

         dbTableDefinitionAndMethods.maxFetch = StringUtil.str2int(maxfetchStr);

         zeiPtr = stepNextAndOverCommentOrEmptyLines(propStrArr, zeiPtr);

         // selectasterisk= wenn ja, dann "SELECT * ..." sonst mit Feld-Liste
         suchString = "selectasterisk";

         checkForPropertyName( suchString, propStrArr, zeiPtr );

         String selectasteriskStr = cutPropertyName(suchString, propStrArr[zeiPtr]);

Log.debug( "selectasteriskStr" , selectasteriskStr );

         dbTableDefinitionAndMethods.selectAsterisk = StringUtil.str2bool(selectasteriskStr);

         zeiPtr = stepNextAndOverCommentOrEmptyLines(propStrArr, zeiPtr);

         // cache-Benutzung
         suchString = "cache";

         checkForPropertyName( suchString, propStrArr, zeiPtr );

         String cacheStr = cutPropertyName(suchString, propStrArr[zeiPtr]);

Log.debug( "cacheStr" , cacheStr );

         dbTableDefinitionAndMethods.cacheUse = StringUtil.str2bool(cacheStr);

         zeiPtr = stepNextAndOverCommentOrEmptyLines(propStrArr, zeiPtr);

         // Vormarschieren bis databasefield=
         suchString = "databasefield";

         checkForPropertyName( suchString, propStrArr, zeiPtr );

         boolean isGenerateKeys = false;

         // Einlesen alle databasefield=
         suchString = "databasefield=";
         while (zeiPtr < propStrArr.length //--
         && ( ( StringUtil.stripTabSpace(propStrArr[zeiPtr]) ).startsWith(suchString) ) ) // databasefield=
         {

            DbFieldDefinitionAndMethods oneField = new DbFieldDefinitionAndMethods();

            String oneFieldStr = propStrArr[zeiPtr].trim();

            //Abschneiden databasefield= bzw. databasefield =
            oneFieldStr = StringUtil.cutLeft(oneFieldStr, "databasefield");
            oneFieldStr = oneFieldStr.trim();
            oneFieldStr = StringUtil.cutLeft(oneFieldStr, "=");

            //String aufteilen in Feldname, Feldtyp und Propterty-Name
            //StringTokenizer defStringTokenizer = new StringTokenizer(oneFieldStr);
            String[] defStrArr = StringUtil.tokenize(oneFieldStr, " ", false, false);

            // Kontrolle, ob genug Eintrge in der Zeile sind
            if (defStrArr.length < 3)
            {
               throw new RuntimeException("databasefield-Definition not valid, line " + (zeiPtr + 1));
            }

            //Feldname
            //String          feldName = defStringTokenizer.nextToken();
            String feldName = defStrArr[0];
            oneField.fieldName = feldName.trim();

            //Feldtyp
            //String  feldTyp = defStringTokenizer.nextToken();
            String feldTyp = defStrArr[1];
            oneField.fieldTyp = feldTyp;

            //Prfung Feldtyp auf Gltigkeit
            if (!StringUtil.isInArray(feldTyp, FIELD_TYPES))
            {
               throw new RuntimeException("FieldType " + feldTyp + " not available, line " + (zeiPtr + 1));
            }

            //Property-Name
            //String propName = defStringTokenizer.nextToken();
            String propName = defStrArr[2];
            oneField.propertyName = propName;

            //restliche Optionen einlesen
            for (int i = 3; i < defStrArr.length; i++)
            {
               if (defStrArr[i].startsWith("//") || defStrArr[i].startsWith("#"))
               {
                  //Kommentar, Schluss mit einlesen dieser Zeile
                  break;
               }
               else if (defStrArr[i].equals("foreignKey"))
               {
                  //Foreign-Key vermerken
                  oneField.isForeignKey = true;
                  dbTableDefinitionAndMethods.foreignKeyFieldNameArr = 
                     StringUtil.arrAdd( dbTableDefinitionAndMethods.foreignKeyFieldNameArr , feldName);
                  dbTableDefinitionAndMethods.foreignKeyPropertyNameArr = 
                     StringUtil.arrAdd( dbTableDefinitionAndMethods.foreignKeyPropertyNameArr , propName);
               }
               else if (defStrArr[i].equals("currentTimeStampOnInsert"))
               {
                  //in diesem Feld den aktuellen TIMESTAMP beim INSERT setzen
                  oneField.currentTimeStampOnInsert = true;
               }
               else if (defStrArr[i].equals("currentTimeStampOnUpdate"))
               {
                  //in diesem Feld den aktuellen TIMESTAMP beim UPDATE setzen
                  oneField.currentTimeStampOnUpdate = true;
               }
               else if (defStrArr[i].equals("currentUserOnInsert"))
               {
                  //in diesem Feld den aktuellen User beim INSERT setzen
                  oneField.currentUserOnInsert = true;
               }
               else if (defStrArr[i].equals("currentUserOnUpdate"))
               {
                  //in diesem Feld den aktuellen User beim UPDATE setzen
                  oneField.currentUserOnUpdate = true;
               }
               else if (defStrArr[i].equals("currentUserOnUpdate"))
               {
                  //in diesem Feld den aktuellen User beim UPDATE setzen
                  oneField.currentUserOnUpdate = true;
               }
               else if (defStrArr[i].equals("readOnly"))
               {
                  //dieses Feld nur Lesen, niemals Schreiben
                  oneField.readOnly = true;
               }
               else if (defStrArr[i].equals("generateKey"))
               {
                  //fr dieses Feld eine laufende Nummer als Key erzeugen
                  if (isGenerateKeys)
                  {
                     throw new RuntimeException("only one field for option generateKey allowed, line " + (zeiPtr + 1));
                  }
                  isGenerateKeys = true;
                  oneField.generateKey = true;
               }
               else if (defStrArr[i].equals("noUpdate"))
               {
                  //dieses Feld nur beim Insert beschreiben, niemals beim Update
                  oneField.noUpdate = true;
               }
               else if (defStrArr[i].startsWith("insertValue:"))
               {
                  //diesen Wert beim Insert setzen
                  oneField.insertValue = StringUtil.cutLeft(defStrArr[i], "insertValue:");
               }
               else if (defStrArr[i].startsWith("updateValue:"))
               {
                  //diesen Wert beim Update setzen
                  oneField.updateValue = StringUtil.cutLeft(defStrArr[i], "updateValue:");
               }
               else if (defStrArr[i].startsWith("trimByRead"))
               {
                  //@todo: trimByRead/Write, leftTrimBy..., reightTrimBy
                  //nach Lesen von der DB trimmen
                  oneField.trimByRead = true;
               }
               else
               {
                  throw new RuntimeException("option " + defStrArr[i] + " not available, line " + (zeiPtr + 1));
               }
            }

            dbTableDefinitionAndMethods.addField(oneField);

            zeiPtr++;
            //zeiPtr = stepNextAndOverCommentOrEmptyLines(propStrArr, zeiPtr);
         } //end while databasefield

         //databaseDefinitionAndMethods.timestampIndex = timestampIndex;

         // Kontrolle, ob berhaupt Felder in der Tabelle sind
         if (dbTableDefinitionAndMethods.fieldArr.length < 1)
         {
            throw new RuntimeException("databasefield-Definition not found, line " + (zeiPtr + 1));
         }

         //Primerykey-Field darf fehlen bei nur lesendem Zugriff
         //      boolean keyFieldOk = false;
         //
         //      // Fehlerkontrolle PrimeryKeyField
         //      for (int i = 0; i < databaseDefinitionAndMethods.feldNameArr.length; i++)
         //      {
         //        if (primeryKeyField.equals(databaseDefinitionAndMethods.feldNameArr[i]))
         //        {
         //          keyFieldOk = true;
         //
         //          break;
         //        }
         //
         //        break;
         //      }
         //
         //      if (!keyFieldOk)
         //      {
         //        throw new Exception("primaryKeyField not contained in FieldPropertyClassMapping");
         //      }

         // *** Erzeugen Method-Cache ***
         for (int i = 0; i < dbTableDefinitionAndMethods.fieldArr.length; i++)
         {
            DbFieldDefinitionAndMethods oneField = dbTableDefinitionAndMethods.fieldArr[i];

            // Get-Method besorgen
            // Property-Name
            String getMethodName = oneField.propertyName;

            // Name der Get-Method
            getMethodName = "get" + StringUtil.upper1(getMethodName);

            // Log.debugOut ("Class " + databaseBean.getClass ());
            // Get-Method
            Method getMethod = BeanUtil.getMethod(beanClass, getMethodName, null);

            if (getMethod == null)
            {
               throw new RuntimeException(" Method " + getMethodName + "() in " + beanClass + " not found ");
            }

            // Get-Method
            oneField.getMethod = getMethod;

            // Set-Method besorgen
            // Property-Name
            String setMethodName = oneField.propertyName;

            // Name der Set-Method
            setMethodName = "set" + StringUtil.upper1(setMethodName);

            // ein Objekt der Parameterklasse, ist ntig fr Class.getMethod
            Class paramClass = getParamClass(oneField.fieldTyp);

            // Set-Method
            Method setMethod = BeanUtil.getMethod(beanClass, setMethodName, new Class[]{ paramClass });

            if (setMethod == null)
            {
               throw new RuntimeException(" Method " + setMethodName + "(" + paramClass + ") in " + beanClass + " not found ");
            }

            // Set-Method
            oneField.setMethod = setMethod;
         }
         // *** Ende Erzeugen Method-Cache ***

         // Erzeugen selectForUpdateSQLStatement
         dbTableDefinitionAndMethods.selectForUpdateSQLStatement =
            createSelectForUpdateSQLStatement(tableName, primeryKeyField);
            
         //Erzeugen FeldListe fr SELECT
         if (dbTableDefinitionAndMethods.selectAsterisk)
         {
            dbTableDefinitionAndMethods.selectFieldsStr = "*";
            dbTableDefinitionAndMethods.qualifiedSelectFieldsStr = "*";
         }
         else
         {
            dbTableDefinitionAndMethods.selectFieldsStr = createSelectFieldsStr(dbTableDefinitionAndMethods);
            dbTableDefinitionAndMethods.qualifiedSelectFieldsStr = createQualifiedSelectFieldsStr(dbTableDefinitionAndMethods);
         }

         // Abspeichern in Cache
         databaseDefinitionAndMethodsCache.put(beanClass, dbTableDefinitionAndMethods);
      }
   }

   /**
    * @param databaseDefinitionAndMethods
    * @return
    */
   private static String createSelectFieldsStr(DbTableDefinitionAndMethods databaseDefinitionAndMethods)
   {
      final StringBuffer strBuff = new StringBuffer();
      for (int i = 0; i < databaseDefinitionAndMethods.fieldArr.length; i++)
      {
         DbFieldDefinitionAndMethods oneField = databaseDefinitionAndMethods.fieldArr[i];
         strBuff.append( oneField.fieldName );
         if ( i < databaseDefinitionAndMethods.fieldArr.length - 1 ) // nicht letzter
         {
            strBuff.append( "," );
         }
      }
      return strBuff.toString();
   }

   /**
    * mit Tabellen-Prefix.
    * 
    * @param databaseDefinitionAndMethods
    * @return
    */
   private static String createQualifiedSelectFieldsStr(DbTableDefinitionAndMethods databaseDefinitionAndMethods)
   {
      final StringBuffer strBuff = new StringBuffer();
      for (int i = 0; i < databaseDefinitionAndMethods.fieldArr.length; i++)
      {
         DbFieldDefinitionAndMethods oneField = databaseDefinitionAndMethods.fieldArr[i];
         strBuff.append( databaseDefinitionAndMethods.tableName + "." + oneField.fieldName );
         if ( i < databaseDefinitionAndMethods.fieldArr.length - 1 ) // nicht letzter
         {
            strBuff.append( "," );
         }
      }
      return strBuff.toString();
   }

   /**
    * Abschneiden des am Anfang einer Zeile der Datei xxx.properties stehenden
    * Namens und des Gleichheitszeichens.
    * 
    * @param pSpecifier Property-name
    * @param pLine gesamte Zeile
    * @return
    */
   private static final String cutPropertyName(String pSpecifier, String pLine)
   {
      String retStr = (StringUtil.cutLeft(pLine, pSpecifier)).trim();
      retStr = (StringUtil.cutLeft(retStr, "=")).trim();
      return retStr;
   }
   
   /**
    * @param suchString
    * @param propStrArr
    * @param zeiPtr
    */
   private static void checkForPropertyName(String pSuchString, String[] pPropStrArr, int pZeiPtr) 
   {
      if (pZeiPtr >= pPropStrArr.length
      || ! ( StringUtil.stripTabSpace(pPropStrArr[pZeiPtr] ).startsWith( pSuchString + "=" ) ) )
      {
         throw new RuntimeException( "Property " + pSuchString + " in xxx.properties expected, line " + (pZeiPtr + 1) );
      }
   }

   /**
    * Laufen ber Leerzeilen oder Kommentarzeilen
    * 
    * @param pStrArr
    * @param pStartPos
    * @return
    */
   private static int stepNextAndOverCommentOrEmptyLines( //--
      final String[] pStrArr, //--
      final int pStartPos )
   {
      //mindestens eine Zeile weiter schalten
      int retPos = pStartPos + 1;
      
      //Leerzeilen und Kommentarzeilen (beginnen mit #) berspringen
      while (retPos < pStrArr.length
            && ( StringUtil.empty( pStrArr[ retPos ] ) 
               || pStrArr[ retPos ].trim().startsWith( "#" ) ) )
      {
         retPos++;
      }
      return retPos;
   }

   /**
    * Erzeugen SQL-Dump als String . <br/>
    * Der erzeugt SQL-Dump besteht aus einem DELETE-Statement fr alle alten Records und <br/>
    * INSERT-Statements fr alle Records der bergebenen ArrayList. <br/>
    * Ein SQL-Dump eignet sich gut zum bertragen von Datenbank-Daten oder <br/>
    * zum Zwischenspeichern bei Datenbank-Modifikationen . <br/>
    *
    * @param aClass Bean-Klasse, Information ber die Bean-Klasse {@link IStorable} wird bentigt falls ArrayList leer
    * @param databaseBeanList zu dumpende Treffer(Bean)-Liste
    * @return Dump-String
    *
    * @throws Exception  alles was da kommt und auf falsche Einstellungen in <default-app-props>.properties zurckgeht, wird weitergeworfen
    */
   public static final String generateSQLDump(Class aClass, ArrayList databaseBeanList, CncContext cncContext)
     
   {
      StringBuffer retStrBuff = new StringBuffer();

      //Statement fr alle vorhandenen Datenstze lschen
      retStrBuff.append("DELETE FROM " + getTableName((IStorable) ClassBuilder.build(aClass)) + "\n");

      if (databaseBeanList != null)
      {
         for (int i = 0; i < databaseBeanList.size(); i++)
         {
            // Achtung: Insert-Statement mit ID wird bentigt
            String strTmp = getInsertSQLStatement((IStorable) databaseBeanList.get(i), cncContext /*, true*/
            );

            // hier werden CR-LF zu \n encodet
            strTmp = StringUtil.encodeBreak(strTmp);

            retStrBuff.append(strTmp + "\n");
         } // end for
      } // end if

      return retStrBuff.toString();
   } // end method generateSQLDump

   /**
    * Erzeugen SQL-Dump als File . <br/>
    * Der erzeugt SQL-Dump besteht aus einem DELETE-Statement fr alle alten Records und <br/>
    * INSERT-Statements fr alle Records der bergebenen ArrayList. <br/>
    * Ein SQL-Dump eignet sich gut zum bertragen von Datenbank-Daten oder <br/>
    * zum Zwischenspeichern bei Datenbank-Modifikationen . <br/>
    *
    * @param aClass Parameter-Bean-Klasse
    * @param fileName Datei-Name
    * @return ob erfolgreich
    */
   public static final boolean writeSQLDumpToFile(Class aClass, String fileName, CncContext cncContext)
   {
      boolean bReturn = false;

      try
      {
         DatabaseStore dbStore = new DatabaseStore();
         ArrayList tmpArrayList = (ArrayList) dbStore.getList(aClass);
         String outStr = generateSQLDump(aClass, tmpArrayList, cncContext);

         bReturn = StringUtil.str2file(outStr, fileName);
      }
      catch (Exception e)
      {
         Log.fatal(e);
      }

      return bReturn;
   } //end method writeSQLDumpToFile

   /**
    * Umwandeln Object-ID eines {@link IStorable} in ein
    * Long-Objekt oder null, wenn Object-ID == {@link IStorable#NOT_STORED}
    * 
    * @param pL
    * @return
    */
   public static final Long objId2LongObj(long pObjId)
   {
      if (pObjId == IStorable.NOT_STORED)
      {
         return null;
      }
      return new Long(pObjId);
   }

} //end class PersistensHelper
