package de.cnc.taglib;


import java.io.IOException;

import javax.servlet.http.HttpSession;
import javax.servlet.jsp.*;
import javax.servlet.jsp.tagext.TagSupport;

import de.cnc.*;
import de.cnc.util.*;
import de.cnc.validator.FormFieldErrorMap;


/**
 * Tag zum Erzeugen des Input-Tag fr ActionFormBeans. <br/><br/>
 * Die Typen Text, TextArea und Checkbox werden untersttzt.<br/><br/>
 * Der aktuelle Wert wird dem Tag als value mitgegeben.<br/><br/>
 * <pre>
 * Verwendung:
 *
 * Action:
 *
 * Ablegen ActionFormBean auf festem SessionKey {@link SessionKeys#CNC_FORM_KEY}
 * <b>
 * JSP:
 *
 * &lt;cnc:form action="xxx"/&gt;
 *
 *    &lt;cnc:input type=&quot;text&quot; property=&quot;...&quot;/&gt;
 *
 * &lt;/form&gt;
 * </b>
 * Das Einparsen erfolgt im {@link CncServlet} ber {@link CncUtil#parseAndSet}.
 *
 * Auswertung der editierten Daten in Action
 * auf festem SessionKey {@link SessionKeys#CNC_FORM_KEY}
 *
 * Beim Type textarea wird folgendes Tag erzeugt:
 * &lt;textarea  name=&quot;meldung&quot; rows=&quot;5&quot; cols=&quot;50&quot;&gt;anfangswert&lt;/textarea&gt;
 *
 * Beim Type checkbox (boolean) erfolgt die Umwandlung vom String zum bool-Wert in
 * der Methode {@link de.cnc.util.BeanUtil#convertToClass}.
 * Die Browser liefern bei nicht gecheckter Checkbox keinen Parameter in den Request
 * wodurch einmal gecheckte Checkboxen niemals zurckgestzt werden.
 * Deshalb wird beim Type checkbox ein Hidden-Field
 *  &lt;input type=&quot;hidden&quot; name=&quot;CHECKBOX_DEFAULT_checkboxname&quot; value=&quot;false&quot;/&gt;
 * mitgesendet.
 * Im {@link CncServlet} wird dieses Hidden-Field ausgewertet,
 * so da ein Rcksetzen erfolgt.
 * </pre>
 * siehe Datei <b>cnc.tld</b> . <br/>
 *
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner Kcker</a>
 * <br/>
 *
 * @author Heiner Kcker
 * @version $Id: $
 */
public class InputTag extends TagSupport {

  /**
   * ob das Input-Feld indiziert wird ( name[nr] ) . <br/>
   * Default ist false
   */
  private String indexed;

  /**
   * Input-Type text|textarea|checkbox|radio|select
   */
  protected String  type = "";

  /**
   * Property der Session-Bean auf dem auf festem SessionKey {@link SessionKeys#CNC_FORM_KEY SessionKeys#CNC_FORM_KEY}
   */
  protected String  property = "";

  /**
   * size-Option
   */
  protected String  size = "";

  /**
   * maxlength-Option
   */
  protected String  maxlength = "";

  /**
   * readonly-Option
   */
  protected String  readonly = null;

  /**
   * disabled-Option
   */
  protected String  disabled = "";

  /**
   * rows-Option fr Textarea
   */
  protected String  rows = "";

  /**
   * cols-Option fr Textarea
   */
  protected String  cols = "";

  /**
   * value-Option fr Radiobutton
   */
  private String value = "";

  /**
   * property multiple for select box
   */
  private String multiple;

  /**
   * sessionvalue-Option fr Werte mit dem Prefix {@link RequestKeys.CNC_SESSION_VALUE_PREFIX}<br>
   * @see CncUtil.parseAndSetImplSessionValues
   */
  private String sessionvalue;

  /**
   * die Maske wird in {@link CncUtil#parseAndSetImplSessionValues(HttpServletRequest)}
   * ausgewertet, um eine Umwandlung in Grossbuchstaben, Ausfiltern onlydigits, Number
   * (digits und Dezimaltrenner) und so weiter Auszulsen.
   */
  private String mask;

   /**
    * ob es ein Muss-Feld ist.
    * Wenn es ein Muss-Feld und leer ist,
    * wird id="mustfield" ausgegeben, so dass
    * die CSS fr mustfield fr die Anzeige 
    * zum Einsatz kommen.
    */
   //private String mustfield;  

   private String tabindex;

  /**
   * get indexed
   *
   * @return indexed
   */
//  public String getIndexed() {
//    return indexed;
//  }// end method getIndexed

  /**
   * set indexed
   *
   * @param indexed new Value
   */
  public void setIndexed(String newIndexed) {
    this.indexed = newIndexed;
  }// end method setIndexed

  /**
   * get type
   *
   * @return type
   */
//  public String getType() {
//    return type;
//  }

  /**
   * set type
   *
   * @param type new Value
   */
  public void setType(String newType) {
    newType = newType.toUpperCase();
    if (!(newType.equals("TEXT") //-
        || newType.equals("TEXTAREA") //-
        || newType.equals("CHECKBOX") //-
        || newType.equals("RADIO") //-
        || newType.equals("SELECT") //-
        || newType.equals( "PASSWORD" )
        )) {
      throw new RuntimeException("unknown input type " + newType + " (text|textarea|checkbox|radio|select)");
    }
    this.type = newType;
  }

  /**
   * get property
   *
   * @return property
   */
//  public String getProperty() {
//    return property;
//  }

  /**
   * set property
   *
   * @param property new Value
   */
  public void setProperty(String newProperty) {
    this.property = newProperty;
  }

  /**
   * get size
   *
   * @return size
   */
//  public String getSize() {
//    return size;
//  }

  /**
   * set size
   *
   * @param size new Value
   */
  public void setSize(String newSize) {
    this.size = newSize;
  }

  /**
   * get maxlength
   *
   * @return maxlength
   */
//  public String getMaxlength() {
//    return maxlength;
//  }

  /**
   * set maxlength
   *
   * @param maxlength new Value
   */
  public void setMaxlength(String newMaxlength) {
    this.maxlength = newMaxlength;
  }

  /**
   * get readonly
   *
   * @return readonly
   */
//  public String getReadonly() {
//    return readonly;
//  }

  /**
   * set readonly
   *
   * @param readonly new Value
   */
  public void setReadonly(String newReadonly) {
    this.readonly = newReadonly;
  }

  /**
   * get disabled
   *
   * @return disabled
   */
//  public String getDisabled() {
//    return disabled;
//  }

  /**
   * set disabled
   *
   * @param disabled new Value
   */
  public void setDisabled(String newDisabled) {
    this.disabled = newDisabled;
  }

  /**
   * get rows
   *
   * @return rows
   */
//  public String getRows() {
//    return rows;
//  }

  /**
   * set rows
   *
   * @param rows new Value
   */
  public void setRows(String newRows) {
    this.rows = newRows;
  }

  /**
   * get cols
   *
   * @return cols
   */
//  public String getCols() {
//    return cols;
//  }

  /**
   * set cols
   *
   * @param cols new Value
   */
  public void setCols(String newCols) {
    this.cols = newCols;
  }

  /**
   * get value
   *
   * @return value
   */
//  public String getValue() {
//    return value;
//  }// end method getValue

  /**
   * set value
   *
   * @param newValue value new Value
   */
  public void setValue(String newValue) {
    this.value = newValue;
  }// end method setValue

  /**
   * get multiple
   *
   * @return multiple
   */
//  public String getMultiple() {
//    return multiple;
//  }// end method getMultiple

  /**
   * set multiple
   *
   * @param newMultiple multiple new Value
   */
  public void setMultiple(String newMultiple) {
    this.multiple = newMultiple;
  }// end method setMultiple

  /**
   * get sessionvalue
   *
   * @return sessionvalue
   */
//  public String getSessionvalue() {
//    return sessionvalue;
//  }// end method getSessionvalue

  /**
   * set sessionvalue
   *
   * @param newSessionvalue sessionvalue new Value
   */
  public void setSessionvalue(String newSessionvalue) {
    this.sessionvalue = newSessionvalue;
  }// end method setSessionvalue

  /**
   * @param pString
   */
  public void setMask(String pString)
  {
     mask = pString;
  }

  /**
   * @param pString
   */
//  public void setMustfield(String pString)
//  {
//     mustfield = pString;
//  }

  /**
   * @param pString
   */
  public void setTabindex(String pString)
  {
     tabindex = pString;
  }

  /**
   * Erzeugen HTML-Code
   *
   * @return Code fr JSP-Seite weiter abarbeiten
   *
   * @exception JspException if a JSP exception has occurred
   */
  public int doStartTag()
      throws JspException {

    HttpSession session = pageContext.getSession();

    //boolean isDate = false;
    String  outString;

    if ("TEXTAREA".equals(type)) {
      // textarea
      // Sonderbehandlung fr Textarea
      // <textarea ... (kein <input type="...)
      outString = "<textarea ";
    }
    else if ("SELECT".equals(type)) {
      // select
      // Sonderbehandlung fr Selectbox
      // <select ... (kein <input type="...)
      outString = "<select ";
    }
    else {
      // kein textarea
      outString = "<input type=\"" + type + "\"";
    }

    // *** Ermittlung Name
    // noch auf stringbuffer umstellen
    outString += " name=\"";

    // Name, so wie er endgltig in das Tag geschrieben wird
    String propIndexed = property;

    final boolean isIndexed = StringUtil.str2bool(indexed);
    int indexInt = -1;
    if (isIndexed) { // indiziert
      // Ermitteln laufende Nummer
      // index wurde vom cnc:iterate/iteratepaged-Tag anhand der Option indexId als Integer-Objekt gesetzt
      indexInt = ( (Integer) session.getAttribute(SessionKeys.CNC_LIST_ELEMENT_ID_KEY) ).intValue() ;

      // input-name ist property-name des Elements + [index]
      // Index an Namen anhngen
      propIndexed += "[" + indexInt + "]";
    }// end if indiziert

    boolean bSessionValue = StringUtil.str2bool(sessionvalue);

    // impliziter sessionvalue
    if (bSessionValue) {
      propIndexed = RequestKeys.CNC_SESSION_VALUE_PREFIX + propIndexed;
    }

    outString += propIndexed;

    outString += "\"";

    // *** Ende Name

    // *** Ermittlung Value

    String  valueStr = "";

    Object valueObject = null;
    Object elementObject = null;

    // impliziter sessionvalue
    if (bSessionValue) {
      valueObject = session.getAttribute(property);
    }
    else {
      // Bean-Value
      if (isIndexed) { // indiziert
        // value wird aus dem Property des element-Objekts ermittelt
        // das element-Objekt reprsentiert die jeweilige Listenzeile und wird vom cnc:iterate/iteratepaged-Tag auf den Session-Key CNC_LIST_ELEMENT_KEY gelegt
        // das inputindexed-tag findet das element-objekt auf dem session-key, der mit der Option elementname bergeben wird

        elementObject = session.getAttribute(SessionKeys.CNC_LIST_ELEMENT_KEY);
      }
      else { // nicht indiziert
        // value wird aus dem Property des CNC_FORM_KEY-Session-Objekts ermittelt
        elementObject = pageContext.findAttribute(SessionKeys.CNC_FORM_KEY);
      }// end if indiziert
      if (elementObject != null) {
        // auf dem Element-Key liegt schon die einzelne Bean aus der ArrayList,
        // deshalb Zugriff ohne [nr]
        valueObject = BeanUtil.getIndexed(elementObject, property);
      }
    }

    if (valueObject != null) {
      if (valueObject instanceof java.util.Date) {
        valueStr = StringUtil.date2str((java.util.Date) valueObject);
        //isDate = true;  // dieser merker wird fr die weitergabe des hidden-fields datetyp_check_ gebraucht
      }
      else {
        valueStr = (valueObject).toString();
      }
    }

    valueStr = StringUtil.encodeHtml(valueStr);

    if ("CHECKBOX".equals(type)) {
      // checkbox
      if (StringUtil.str2bool(valueStr)) {
        outString += " checked ";
      }
    }
    else if ("RADIO".equals(type)) {
      // radio button
      if (StringUtil.stringEqual(valueStr, value)) {
        outString += " checked ";
      }
    }
    else if (!"TEXTAREA".equals(type)) {
      // keine checkbox und kein Textarea
      outString += " value=\"" + valueStr + "\"";
    }

    // size-Option
    if (!StringUtil.empty(size)) {
      outString += " size=\"" + size + "\"";
    }

    // maxlength-Option
    if (!StringUtil.empty(maxlength)) {
      outString += " maxlength=\"" + maxlength + "\"";
    }

    // readonly-Option
    //if (!StringUtil.empty(readonly))
    if ( StringUtil.str2bool( readonly ) )
    {
      //tString += " readonly=\"" + readonly + "\"";
      outString += " readonly=\"true\"";
    }

    // disabled-Option
    if (StringUtil.str2bool(disabled)) {
      outString += " disabled=\"" + disabled + "\"";
    }

//    // mustfield-Option
//    if (StringUtil.str2bool(mustfield) && StringUtil.empty( valueStr )) 
//    {
//      outString += " id=\"mustfield\"";
//    }

    // Behandlung invalid-Field
      FormFieldErrorMap formFieldErrorMap = (FormFieldErrorMap) session.getAttribute(SessionKeys.CNC_FORM_FIELD_ERROR_MAP);
//System.out.println( "formFieldErrorMap: " + formFieldErrorMap );
      if (formFieldErrorMap != null && formFieldErrorMap.get(propIndexed) != null)
      {
          outString += " id=\"CNC_INVALID_FIELD_STYLE\"";
//System.out.println( "outString: " + outString );
      }

    // rows-Option
    if (!StringUtil.empty(rows)) {
      outString += " rows=\"" + rows + "\"";
    }

    // cols-Option
    if (!StringUtil.empty(cols)) {
      outString += " cols=\"" + cols + "\"";
    }

    // value-Option
    // nur bei checkbox und radio
    if (!StringUtil.empty(value)) {
      outString += " value=\"" + value + "\"";
    }

    // tabindex-Option
    if (!StringUtil.empty(tabindex)) {
      outString += " tabindex=\"" + tabindex + "\"";
    }

    if (!"TEXTAREA".equals(type)) {
      // kein textarea
      outString += "/";
    }

    // select box
    if ("SELECT".equals(type) && StringUtil.str2bool(multiple)) {
      outString += " multiple";
    }

    outString += ">";

    if ("TEXTAREA".equals(type)) {
      // textarea
      outString += valueStr + "</textarea>";
    }
    else if ("SELECT".equals(type)) {
      // selectbox
      if (elementObject == null) {
        throw new JspException("no bean for input type=\"select\" found");
      }
      Object optionsObject = BeanUtil.getIndexed(elementObject, property + "SelectboxOptions"); // noch Konstante machen

      final boolean isValuableSelect;
      if (optionsObject instanceof String[][])
      {
         isValuableSelect = true;
      }
      else if (optionsObject instanceof String[])
      {
         isValuableSelect = false;
      }
      else
      {
        throw new JspException("no options String[] or String[][] for input type=\"select\" get" + StringUtil.upper1(property) + "SelectboxOptions() found");
      }

      if (isValuableSelect)
      {
         String[][] optionsStrArrArr = (String[][]) optionsObject;

         // optionsStrArrArr[i][0] ist der value und optionsStrArrArr[i][1] ist die Anzeige
         for (int i = 0; i < optionsStrArrArr.length; i++) 
         {
           outString += "<option value=\"" + optionsStrArrArr[i][0] + "\"" ;
           
           if (StringUtil.stringEqual(valueStr, optionsStrArrArr[ i ][ 0 ])) 
           {
             outString += " selected";
           }
           outString += ">" + optionsStrArrArr[ i ][ 1 ] + "</option>";
         }
      }
      else
      {
         String[] optionsStrArr = (String[]) optionsObject;

         for (int i = 0; i < optionsStrArr.length; i++) {
           outString += "<option";
           if (StringUtil.stringEqual(valueStr, optionsStrArr[ i ])) {
             outString += " selected";
           }
           outString += ">" + optionsStrArr[ i ] + "</option>";
         }
      }
      outString += "</select>";
    }
    else if ("CHECKBOX".equals(type)) {
      // wenn es sich bei dem Input-feld um eine checkbox handelt,
      // wird ein spezielles hidden-field checkbox_default_name=false mitgegeben,
      // weil der Browser nicht gecheckte Checkboxen (value == false) nicht mit in den request sendet,
      // wodurch einmal gecheckte boolean-Checkboxen nicht mehr zurckgesetzt werden.
      // Im CncServlet werden diese hidden-fields ausgewertet,
      // so da ein Rcksetzen der checkboxen erfolgt.
      outString += "<input type=\"hidden\" name=\"" + RequestKeys.CNC_CHECKBOX_DEFAULT_PREFIX + property
          + "\" value=\"false\"/>";
    }

   if ( ! StringUtil.empty( mask ) )
   {
      outString += "<input type=\"hidden\" name=\"" + RequestKeys.CNC_SESSION_VALUEMASK_PREFIX + property
          + "\" value=\"" + mask + "\"/>";
   }

//Log.debug( outString ) ;

    try {
      JspWriter writer = pageContext.getOut();

      writer.print(outString);
    } catch (IOException e) {
      throw new JspException(e.toString());
    }

    return EVAL_PAGE;
  }// end method doStartTag

  /**
   * nichts zu tun
   *
   * @return Code fr Weiterverarbeitung JSP
   *
   * @throws JspException
   */
  public int doEndTag()
      throws JspException {
    return EVAL_PAGE;
  }// end method doEndTag

}// end class InputTag
