package de.cnc.util;

import java.util.ArrayList;

import javax.servlet.http.HttpSession;

import de.cnc.SessionKeys;

/**
 * Hilfsklasse zum seitenweisen Anzeigen. <body><br/>
 * arbeitet zusammen mit: <br/>
 * {@link de.cnc.taglib.IterateBrowsedTag IterateBrowsedTag} <br/>
 * <br/>
 * <pre>
 * Vorgehensweise:
 *
 * Action aus {@link de.cncdemo.actions} holt Liste von Datenbank
 *      |
 *      +--&gt; legt Liste in Session {@link SessionKeys#CNC_LIST_KEY}
 *                 |
 *                 +--&gt; ruft {@link #resetBrowseControl resetBrowseControl}
 *                       |
 *                       +--&gt; forward zur JSP
 *
 *
 *
 *       +---------------------------------------------------+
 *       | JSP-Seite                                         |
 *       |                                                   |&lt;------------+
 *       |                                                   |             |
 *       | {@link de.cnc.taglib.ButtonPrevPageTag}                   {@link de.cnc.taglib.ButtonNextPageTag} |             |
 *       | Hidden-Field srcjsp           Hidden-Field srcjsp |             |
 *       |                                                   |             |
 *       | {@link de.cnc.taglib.IterateBrowsedTag IterateBrowsedTag}                                   |             |
 *       |  &lt;cnc:iteratebrowsed&gt;                        |             |
 *       |     Liste                                         |             |
 *       |      ...                                          |             |
 *       |      ...                                          |             |
 *       |      ...                                          |      zurck zu srcjsp
 *       |      ...                                          |             |
 *       |      ...                                          |             |
 *       |      ...                                          |             |
 *       |  &lt;/cnc:iteratebrowsed&gt;                       |             |
 *       | {@link de.cnc.taglib.IterateBrowsedTag IterateBrowsedTag}                                   |             |
 *       |                                                   |             |
 *       |                                                   |             |
 *       |                                                   |             |
 *       | {@link de.cnc.taglib.ButtonPrevPageTag}                   {@link de.cnc.taglib.ButtonNextPageTag} |             |
 *       |      |                                    |       |             |
 *       +------|------------------------------------|-------+             |
 *              |                                    |                     |
 *              |                                    +-- {@link CncServlet#doRequest} implizite Action -----+
 *              |                                             |            |
 *              +--------------------------------------- {@link CncServlet#doRequest} implizite Action -----+
 *                                                            |
 *                                                       {@link #staticGoNext staticGoNext}
 *                                                       {@link #staticGoPrev staticGoPrev}
 *
 * </pre>
 * Achtung: nach jedem Einfgen oder Lschen eines Listenelements mu {@link #validate} aufgerufen werden.
 * <br/>
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner Kcker</a><br/>
 * <br/>
 *
 * @author Heiner Kcker
 * @version $Id: $
 */
public final class BrowseControl
{

    /**
     * anzuzeigende Seite
     */
    private float currentPage = 1;

    /**
     * Anzahl Positionen je Seite
     */
    private int itemsPerPage = 10;

    /**
     * Anzahl Positionen Gesamt
     */
    private int itemCount = 0;

    /**
     * Konstruktor parameterlos. <br/>
     * Hier kann man itemsPerPage voreinstellen
     * sinnvoll wre evtl. noch static preset fr itemsPerPage . <br/>
     */
    public BrowseControl()
    {
        super();
    } // end constructor

    /**
     * Konstruktor mit Angabe itemsPerPage. <body><br/>
     *
     * @param newitemsPerPage Anzahl Positionen je Seite
     */
    public BrowseControl(int newitemsPerPage)
    {
        if (newitemsPerPage > 0)
        {
            itemsPerPage = newitemsPerPage;
        } // end if
    } // end constructor(int)

    /**
     * Sichern, da die anzuzeigende Seite im darstellbaren Bereich liegt
     *
     * @param itemCount Anzahl anzuzeigender Elemente gesamt
     */
    public void validate(int newItemCount)
    {
        // erstes und einmaliges Speichern der Anzahl Postionen gesamt
        // jedes ndern der Anzahl Positionen (Hinzufgen.Lschen) mu ber die validate-Methode mitgeteilt werden
        itemCount = newItemCount;

        if (currentPage < 1)
        {
            currentPage = 1;
        }
        else if (
            ((currentPage - 1) * itemsPerPage) >= itemCount)
            // aktuelle Seite umrechnen auf Basis Null und Prfung, ob erstes Element der anzuzeigenden Seite (Basis 0) >= Anzahl Gesamt (Basis 1)
        {
            currentPage =
                (int) Math.ceil((float) itemCount / (float) itemsPerPage);

            currentPage = Math.max(currentPage, 1);
            // immer mindestens auf Seite 1 stellen
        } // end if
        currentPage =
            (float) Math.round(currentPage * (float) itemsPerPage)
                / (float) itemsPerPage;
        // nachrunden auf Positionen je Seite
    } // end method validate

    /**
     * Sichern, da die anzuzeigende Seite im darstellbaren Bereich liegt . <br/>
     * Die Anzahl Positionen wird vom festen Session-Key {@link SessionKeys#CNC_LIST_KEY}
     */

    /*
     public void validate( HttpSession session ) {
     // erstes und einmaliges Speichern der Anzahl Postionen gesamt
     // jedes ndern der Anzahl Positionen (Hinzufgen.Lschen) mu ber die validate-Methode mitgeteilt werden
     itemCount = getItemCountFromSession( session ) ;
    
     if ( currentPage < 1 ) {
     currentPage = 1;
     }
     else if (((currentPage - 1) * itemsPerPage) >= itemCount) // aktuelle Seite umrechnen auf Basis Null und Prfung, ob erstes Element der anzuzeigenden Seite (Basis 0) >= Anzahl Gesamt (Basis 1)
     {
     currentPage = (int) Math.ceil((float) itemCount / (float) itemsPerPage);
    
     currentPage = Math.max(currentPage, 1); // immer mindestens auf Seite 1 stellen
     } // end if
     currentPage = (float) Math.round( currentPage * (float) itemsPerPage ) / (float) itemsPerPage ; //nachrunden auf Positionen je Seite
     } // end method validate
     */

    /**
     * Anzeigebereich (Fokus) auf ein bestimmtes Element setzen
     *
     * @param iPaPosition auf dieses Listenelelement soll der Listen-Fokus gesetzt werden.
     */
    public void setFocus(int iPaPosition)
    {
        if (iPaPosition >= 0 && iPaPosition < itemCount)
        {
            currentPage = ((float) iPaPosition / (float) itemsPerPage) + 1;
            currentPage =
                (float) Math.round(currentPage * (float) itemsPerPage)
                    / (float) itemsPerPage;
            // nachrunden auf Positionen je Seite
        } // end if
    } // end method setFocus

    /**
     * get Offset  (Basis 0)
     *
     * @return Offset
     */
    public int getOffset()
    {
        return Math.round((currentPage - 1) * itemsPerPage);
    } // end method

    /**
     * get currentPage  (Basis 1)
     *
     * @return currentPage
     */
    public float getCurrentPage()
    {
        return currentPage;
    } // end method

    /**
     * set currentPage (Basis 1)
     *
     * @param newCurrentPage new Value
     */
    public void setCurrentPage(int newCurrentPage)
    {
        this.currentPage = newCurrentPage;
    } // end method

    /**
     * get itemsPerPage (Basis 1)
     *
     * @return itemsPerPage
     */
    public int getItemsPerPage()
    {
        return itemsPerPage;
    } // end method

    /**
     * Anzahl Seiten
     *
     * @return pageCount
     */
    public int getPageCount()
    {
        return Math.max(
            ((int) Math.ceil((double) itemCount / (double) itemsPerPage)),
            1);
    } // end method

    /**
     * set itemsPerPage (Basis 1)
     *
     * @param itemsPerPage new Value
     */
    public void setitemsPerPage(int newItemsPerPage)
    {
        this.itemsPerPage = newItemsPerPage;
    } // end method

    /**
     * eine Seite zurck schalten
     */
    public void goPrev()
    {
        // validate hier eigentlich nicht ntig, weil immer vor Anzeige aufgerufen
        currentPage--;
    } // end method

    /**
     * eine Position zurck schalten
     */
    public void goPrevPos()
    {
        // validate hier eigentlich nicht ntig, weil immer vor Anzeige aufgerufen
        currentPage -= 1 / (float) itemsPerPage;
        // Log.debug( "goPrevPos " + currentPage ) ;
    } // end method

    /**
     * eine Position vorwrts schalten
     */
    public void goNextPos()
    {
        // validate hier eigentlich nicht ntig, weil immer vor Anzeige aufgerufen
        currentPage += 1 / (float) itemsPerPage;
        // Log.debug( "goNextPos " + currentPage ) ;
    } // end method

    /**
     * eine Seite vorwrts schalten
     */
    public void goNext()
    {
        // validate hier eigentlich nicht ntig, weil immer vor Anzeige aufgerufen
        currentPage++;
    } // end method

    /**
     * zur ersten Seite schalten
     */
    public void goFirst()
    {
        // validate hier eigentlich nicht ntig, weil immer vor Anzeige aufgerufen
        currentPage = 1;
    } // end method

    /**
     * zur letzten Seite schalten
     */
    public void goLast()
    {
        // validate hier eigentlich nicht ntig, weil immer vor Anzeige aufgerufen
        currentPage = getPageCount();
    } // end method

    /**
     * Abfrage, ob die aktuelle Seite die erste Seite ist
     */
    public boolean isFirstPage()
    {
        return (currentPage < 2);
    } // end method

    /**
     * Abfrage, ob die aktuelle Seite die erste Seite ist
     */
    public boolean isFirstPos()
    {
        return (currentPage <= 1);
    } // end method

    /**
     * Abfrage, ob die aktuelle Seite die letzte Seite ist
     */
    public boolean isLastPos()
    {
        // aktuelle Seite umrechnen auf Basis Null und Prfung,
        // ob erstes Element der nchsten anzuzeigenden Seite (Basis 0) >= Anzahl Gesamt (Basis 1)
        // return ( ( (float) currentPage * (float) itemsPerPage ) - 1 > itemCount + 1 ) ;
        // Log.debug( "isLastPos(): getOffset(): " + getOffset() + " itemCount " + itemCount ) ;
        return (getOffset() + 1 >= itemCount);
    } // end method

    /**
     * Abfrage, ob die aktuelle Seite die letzte Seite ist
     */
    public boolean isLastPage()
    {
        // aktuelle Seite umrechnen auf Basis Null und Prfung,
        // ob erstes Element der nchsten anzuzeigenden Seite (Basis 0) >= Anzahl Gesamt (Basis 1)
        return ((currentPage * itemsPerPage) >= itemCount);
    } // end method

    /**
     * holt die Element-Anzahl vom festen Session-Key {@link SessionKeys#CNC_LIST_KEY}
     */
    public static int getItemCountFromSession(HttpSession session)
    {
        int iAnzahl = 0;
        Object obj = session.getAttribute(SessionKeys.CNC_LIST_KEY);

        if (obj != null)
        {
            if (obj instanceof ArrayList)
            {
                ArrayList myArrayList = (ArrayList) obj;

                iAnzahl = myArrayList.size();
            }
        }
        return iAnzahl;
    } // end method getItemCountFromSession

    /**
     * holt das BrowseControl vom festen Session-Key {@link SessionKeys#CNC_BROWSE_CTRL_KEY}
     */
    public static final BrowseControl getBrowseControlFromSession(HttpSession session)
    {
        Object obj = session.getAttribute(SessionKeys.CNC_BROWSE_CTRL_KEY);

        if (obj != null)
        {
            if (obj instanceof BrowseControl)
            {
                return (BrowseControl) obj;
            }
        }
        return null;
    } // end method getBrowseControlFromSession

    /**
     * holt das BrowseControl vom festen Session-Key {@link SessionKeys#CNC_BROWSE_CTRL_KEY}
     * oder legt ein neues BrowseControl an, wenn keins vorhanden
     */
    private static final BrowseControl getOrNewBrowseControlFromSession(HttpSession session)
    {
        BrowseControl browseControl = getBrowseControlFromSession(session);
        //Log.debug( "BrowseControl#getOrNewBrowseControlFromSession browseControl" , browseControl );
        if (browseControl == null)
        {
            // neues BrowseControl anlegen
            browseControl = new BrowseControl();
            CncUtil.setAttribute(
                session,
                SessionKeys.CNC_BROWSE_CTRL_KEY,
                browseControl);
        }
        return browseControl;
    } // end method

    /**
     * Aufruf {@link #validate} auf BrowseControl auf festen Session-Key {@link SessionKeys#CNC_BROWSE_CTRL_KEY}
     */
    public static final void staticValidate(HttpSession session)
    {
        int iAnzahl = getItemCountFromSession(session);
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        browseControl.validate(iAnzahl);
    } // end method

    /**
     * Initialisieren des Browsed-Control auf festem Session-Key {@link SessionKeys#CNC_BROWSE_CTRL_KEY}. <br/>
     * Diese Methode sollte von jeder Action, die eine Bltter-Liste in die Session stellt, aufgerufen werden, <br/>
     * es sei denn, die Tabelle wurde verndert und soll wieder auf der selben Seite angezeigt werden. <br/>
     *
     * @param session die Session mu von der aufrufenden Action(Tag usw.) sichtbar gemacht werden
     */
    public static void resetBrowseControl(HttpSession session)
    {
        CncUtil.setAttribute(
            session,
            SessionKeys.CNC_BROWSE_CTRL_KEY,
            new BrowseControl());
        BrowseControl.staticValidate(session);
        // holt Elementanzahl aus CNC_LIST_KEY
    } // end method resetBrowseControl

    /**
     * Initialisieren des Browsed-Control auf festem Session-Key  {@link SessionKeys#CNC_BROWSE_CTRL_KEY}. <body><br/>
     * Diese Methode sollte von jeder Action, die eine Bltter-Liste in die Session stellt, aufgerufen werden.
     *
     * @param session die Session mu von der aufrufenden Action(Tag usw.) sichtbar gemacht werden
     * @param newItemsPerPage Anzahl Positionen je Seite
     */
    public static void resetBrowseControl(
        HttpSession session,
        int newItemsPerPage)
    {
        // neues BrowseControl mit nicht Default-ItemsPerPage
        BrowseControl browseControl = new BrowseControl(newItemsPerPage);

        CncUtil.setAttribute(
            session,
            SessionKeys.CNC_BROWSE_CTRL_KEY,
            browseControl);
        staticValidate(session); // holt Elementanzahl aus CNC_LIST_KEY
    } // end method resetBrowseControl(session,int)

    /**
     * Aufruf {@link #goNext} auf BrowseControl auf festen Session-Key {@link de.cnc.SessionKeys#CNC_BROWSE_CTRL_KEY}
     */
    public static final void staticGoNext(HttpSession session)
    {
        // int iAnzahl = getItemCountFromSession( session ) ;
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        // browseControl.validate( iAnzahl ) ;
        browseControl.goNext();

        // browseControl.validate( iAnzahl ) ;//sicher ist sicher
    } // end method

    /**
     * Aufruf {@link #goNextPos} auf BrowseControl auf festen Session-Key {@link de.cnc.SessionKeys#CNC_BROWSE_CTRL_KEY}
     */
    public static final void staticGoNextPos(HttpSession session)
    {
        // int iAnzahl = getItemCountFromSession( session ) ;
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);
        //Log.debug("browseControl", browseControl);

        // browseControl.validate( iAnzahl ) ;
        browseControl.goNextPos();

        // browseControl.validate( iAnzahl ) ;//sicher ist sicher
    } // end method

    /**
     * Aufruf {@link #goPrev} auf BrowseControl auf festen Session-Key {@link de.cnc.SessionKeys#CNC_BROWSE_CTRL_KEY}
     */
    public static final void staticGoPrev(HttpSession session)
    {
        // int iAnzahl = getItemCountFromSession( session ) ;
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        // browseControl.validate( iAnzahl ) ;
        browseControl.goPrev();

        // browseControl.validate( iAnzahl ) ;//sicher ist sicher
    } // end method

    /**
     * Aufruf {@link #goPrev} auf BrowseControl auf festen Session-Key {@link de.cnc.SessionKeys#CNC_BROWSE_CTRL_KEY}
     */
    public static final void staticGoPrevPos(HttpSession session)
    {
        // int iAnzahl = getItemCountFromSession( session ) ;
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        // browseControl.validate( iAnzahl ) ;
        browseControl.goPrevPos();

        // browseControl.validate( iAnzahl ) ;//sicher ist sicher
    } // end method

    /**
     * Aufruf {@link #goLast} auf BrowseControl auf festen Session-Key {@link de.cnc.SessionKeys#CNC_BROWSE_CTRL_KEY}
     */
    public static final void staticGoLast(HttpSession session)
    {
        int iAnzahl = getItemCountFromSession(session);
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        browseControl.validate(iAnzahl);
        browseControl.goLast();
        browseControl.validate(iAnzahl); // sicher ist sicher
    } // end method

    /**
     * Aufruf {@link #goFirst} auf BrowseControl auf festen Session-Key {@link de.cnc.SessionKeys#CNC_BROWSE_CTRL_KEY}
     */
    public static final void staticGoFirst(HttpSession session)
    {
        int iAnzahl = getItemCountFromSession(session);
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        browseControl.validate(iAnzahl);
        browseControl.goFirst();
        browseControl.validate(iAnzahl); // sicher ist sicher
    } // end method

    /**
     * Aufruf {@link #isLastPage} auf BrowseControl auf festen Session-Key {@link de.cnc.SessionKeys#CNC_BROWSE_CTRL_KEY}
     */
    public static final boolean staticIsLastPos(HttpSession session)
    {
        int iAnzahl = getItemCountFromSession(session);
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        browseControl.validate(iAnzahl);

        return browseControl.isLastPos();
    } // end method

    /**
     * Aufruf {@link #isLastPage} auf BrowseControl auf festen Session-Key {@link de.cnc.SessionKeys#CNC_BROWSE_CTRL_KEY}
     */
    public static final boolean staticIsLastPage(HttpSession session)
    {
        int iAnzahl = getItemCountFromSession(session);
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        browseControl.validate(iAnzahl);

        return browseControl.isLastPage();
    } // end method

    /**
     * Aufruf {@link #isFirstPage} auf BrowseControl auf festen Session-Key {@link de.cnc.SessionKeys#CNC_BROWSE_CTRL_KEY}
     */
    public static final boolean staticIsFirstPos(HttpSession session)
    {
        int iAnzahl = getItemCountFromSession(session);
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        browseControl.validate(iAnzahl);
        return browseControl.isFirstPos();
    } // end method

    /**
     * Aufruf {@link #isFirstPage} auf BrowseControl auf festen Session-Key {@link de.cnc.SessionKeys#CNC_BROWSE_CTRL_KEY}
     */
    public static final boolean staticIsFirstPage(HttpSession session)
    {
        int iAnzahl = getItemCountFromSession(session);
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        browseControl.validate(iAnzahl);
        return browseControl.isFirstPage();
    } // end method

    /**
     * Aufruf {@link #getCurrentPage} auf BrowseControl auf festen Session-Key {@link de.cnc.SessionKeys#CNC_BROWSE_CTRL_KEY}
     */
    public static final float staticGetCurrentPage(HttpSession session)
    {
        int iAnzahl = getItemCountFromSession(session);
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        browseControl.validate(iAnzahl);

        return browseControl.getCurrentPage();
    } // end method

    /**
     * Aufruf {@link #getPageCount} auf BrowseControl auf festen Session-Key {@link de.cnc.SessionKeys#CNC_BROWSE_CTRL_KEY}
     */
    public static final int staticGetPageCount(HttpSession session)
    {
        int iAnzahl = getItemCountFromSession(session);
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        browseControl.validate(iAnzahl);
        return browseControl.getPageCount();
    } // end method staticGetPageCount

    /**
     * Ausgabe ab welchem Element ausgegeben wird fr BrowseControl auf festen Session-Key {@link de.cnc.SessionKeys#CNC_BROWSE_CTRL_KEY}
     *
     * @return ab welchem Element angezeigt wird (Basis 1)
     */
    public static final int staticElementAb(HttpSession session)
    {
        int iAnzahl = getItemCountFromSession(session);
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        browseControl.validate(iAnzahl);

        return browseControl.getOffset() + 1;
    } // end method

    /**
     * Ausgabe bis zu welchem Element ausgegeben wird fr BrowseControl auf festen Session-Key {@link de.cnc.SessionKeys#CNC_BROWSE_CTRL_KEY}
     *
     * @return bis zu welchem Element angezeigt wird (Basis 1)
     */
    public static final int staticElementBis(HttpSession session)
    {
        int iAnzahl = getItemCountFromSession(session);
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        browseControl.validate(iAnzahl);

        return Math.min(
            iAnzahl,
            browseControl.getOffset() + browseControl.getItemsPerPage());
    } // end method

    /**
     * Ausgabe Anzahl Elemente auf festen Session-Key {@link de.cnc.SessionKeys#CNC_LIST_KEY}
     *
     * @return Anzahl Elemente (Basis 1)
     */
    public static final int staticElementAnz(HttpSession session)
    {
        return getItemCountFromSession(session);
    } // end method

    /**
     * Anzeigebereich (Fokus) auf ein bestimmtes Element setzen . <br/>
     * Aufruf {@link #getPageCount} auf BrowseControl auf festen Session-Key {@link SessionKeys#CNC_BROWSE_CTRL_KEY}
     *
     * @param iPaPosition auf dieses Listenelelement soll der Listen-Fokus gesetzt werden.
     */
    public static final void staticSetFocus(
        HttpSession session,
        int iPaPosition)
    {
        int iAnzahl = getItemCountFromSession(session);
        BrowseControl browseControl = getOrNewBrowseControlFromSession(session);

        browseControl.validate(iAnzahl);
        browseControl.setFocus(iPaPosition);
    } // end method

    public String toString()
    {
        return this.getClass().getName() //--
        +"[currentPage: " + currentPage //--
        +", itemsPerPage: " + itemsPerPage //--
        +", itemCount: " + itemCount + "]";
    }

} // end class BrowseControl
