package de.cnc.util.config;

import java.io.File;
import java.util.Hashtable;

import javax.servlet.http.HttpServlet;

import de.cnc.util.*;

/**
 * Singleton zum Lesen der Konfigurationsdateien . <br/>
 *
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner Kcker</a>
 * <br/>
 *
 * @author Heiner Kcker
 * @version $Id: $
 */
public class ConfigManager
{
  /**
   * Properties-Cache
   */
  private static Hashtable  propCache = new Hashtable();

  /**
   * Eltern-Verzeichniss der properties-Files . <br/><br/>
   * Typischerweise <b>web/WEB-INF/</b>.<br/><br/>
   * Unterhalb dieses Verzeichnisses gibt es ein Verzeichnis <b>config</b>,<br/>
   * in welchem die properties-Files liegen.<br/><br/>
   */
  private static String configRoot = null;

   /**
    * Default-Properties-File der jeweiligen CnC-Applikation . <br>
    * siehe <b>web.xml</b>-Init-Parameter <b>defaultconfigfile</b>
    */
   private static String defaultConfigFileName = null;

  /**
   * Erzeugen kompletten <b>Pfad+Dateinamen</b> fr ein properties-File
   *
   * @param configuration Dateiname ohne Pfad
   * @return kompletter <b>Pfad+Dateiname</b> fr ein properties-File
   */
  public static String getRealPath(String configuration)
  {
    return getConfigRoot() + "config/" + configuration; // .toLowerCase();
  }

  /**
   * Erzeugen kompletten <b>Pfad+Dateinamen+Endung</b> fr ein properties-File
   *
   * @param configuration Dateiname ohne Pfad und Endung
   * @return kompletter <b>Pfad+Dateiname+Endung</b> fr ein properties-File
   */
  public static String getRealPathAndExt(String configuration)
  {
    return getConfigRoot() + "config/" + configuration.toLowerCase() + ".properties";
  }

  /**
   * Setzen des Eltern-Verzeichnisses der properties-Files {@link #configRoot} . <br/><br/>
   * Typischerweise <b>web/WEB-INF/</b>.<br/><br/>
   * Unterhalb dieses Verzeichnisses gibt es ein Verzeichnis <b>config</b>,<br/>
   * in welchem die properties-Files liegen.<br/><br/>
   * Ein eventuell fehlendes Slash wird am Ende angehngt.<br/><br/>
   *
   * @param root zu setzender Pfad
   */
  synchronized static private void setConfigRoot(String root)
  {
    if (root != null)
    {
      configRoot = root;

      if (!root.endsWith("/") &&!root.endsWith("\\"))
      {
        configRoot += "/";
      }
    }
  }

  /**
   * Abfragen des mit {@link #setConfigRoot} gesetzten Pfades des  Eltern-Verzeichnisses der properties-Files {@link #configRoot} . <br/>
   *
   * @return  Eltern-Verzeichniss der properties-Files . <br/>
   */
  static public String getConfigRoot()
  {
    return configRoot;
  }

  /**
   * Initialisierung des Config-Managers . <br/>
   * Wird aufgerufen von {@link de.cnc.generic.CncServlet#init CncActionServlet#init}.<br/>
   *
   * @param servlet Referenz auf aufrufendes Servlet, wird bentigt um an die Configurationseinstellungen aus <b>web.xml</b> zu kommen
   */
  static public void init(HttpServlet servlet)
  {
    if (getConfigRoot() != null)
    {
      return;
    }

    // System.out.println( "ConfigManager#init" );
    synchronized (ConfigManager.class)
    {
      if (getConfigRoot() == null)
      {
        //Abfrage des Pfades der Properties-Files aus web.xml
        final String configrootParamStr = servlet.getServletConfig().getInitParameter("configroot");

        if (configrootParamStr == null)
        {
           throw new RuntimeException( "web.xml init parameter configroot missed" );
        }

        final String configrootStr = servlet.getServletConfig().getServletContext().getRealPath(configrootParamStr).trim();

        setConfigRoot(configrootStr);

        //Abfrage des Default-Properties-Files aus web.xml
        final String defaultConfigFileNameParamStr = servlet.getServletConfig().getInitParameter("defaultconfigfile");

        if (defaultConfigFileNameParamStr == null)
        {
           throw new RuntimeException( "web.xml init parameter defaultconfigfile missed" );
        }
        defaultConfigFileName = StringUtil.cutRight( defaultConfigFileNameParamStr , ".properties" );

      }
    }
  }

  /**
   * Holen eines erweiterten Properties-Objekts aus dem Cache . <br/>
   * Bei Bedarf Neuladen.<br/>
   * Enthlt eine Aktualisierungslsung.<br/>
   *
   * @param fileName Dateiname ohne Endung ".properties"
   * @return gewnschtes CncProperties-Objekt
   */
  public static CncProperties getProperties(String fileName)
  {
    CncProperties proper = (CncProperties) propCache.get(fileName);

    if (proper != null)
    {
      //Hier wird stndig auf die nderung des Verzeichnisses WEB-INF/config abgefragt
      //habe ich erst mal so bernommen, ist aber wahrscheinlich keine effektive Lsung
      //Besser Timeout-Mimik einbauen
      File  file = new File(ConfigManager.getRealPathAndExt(fileName));

      if (proper.getLastModified() != file.lastModified())
      {
        propCache.remove(fileName);

        proper = null;
      }
    }

    if (proper == null)
    {
      synchronized (ConfigManager.class)
      {
        if (proper == null)
        {
          try
          {
            proper = new CncProperties();

            // System.out.println("reading fileName: " + fileName);
            proper.load(ConfigManager.getRealPathAndExt(fileName));

            File  file = new File(ConfigManager.getRealPathAndExt(fileName));

            proper.setLastModified(file.lastModified());
            propCache.put(fileName, proper);
          }
          catch (Exception e)
          {
            Log.fatal( e );

            proper = null;
          }
        }
      }
    }

    return proper;
  }

  /**
   * Einlesen Properties-File in ein String-Array mit Caching . <br/>
   * wird benutzt in {@link de.cnc.server.PersistensHelper PersistensHelper}
   * zum Einlesen<br/> des in der Datei <b>&lt;default-app-props&gt;.properties</b> hinterlegten<br/>
   * Feld-Property-Mappings der Datenbank-Felder.<br/>
   * Benutzt {@link StringUtil#fileToStrArr StringUtil#fileToStrArr}
   *
   * @param fileName Datei-Name
   * @return String-Array mit den Zeilen des Properties-Files
   */
  public static String[] getPropertiesFileAsStringArr(String fileName)
  {
    // in Hashtable wird an den Filenamen zur Unterscheidung  + "PropertiesFileAsString[]" angehngt
    String    hashKey = fileName + "PropertiesFileAsString[]";
    String[]  strArr = (String[]) propCache.get(hashKey);

    if (strArr == null)
    {
      synchronized (ConfigManager.class)
      {
        if (strArr == null)
        {
          try
          {
            String realPathAndExt = ConfigManager.getRealPathAndExt(fileName);
//Log.debug( "realPathAndExt: " + realPathAndExt );
            strArr = StringUtil.fileToStrArr(realPathAndExt);

            propCache.put(hashKey, strArr);
          }
          catch (Exception e)
          {
            Log.fatal( e );
            strArr = null;
          }
        }
      }
    }

    return strArr;
  }

  /**
   * @return
   */
  public static CncProperties getDefaultProperties()
  {
     return getProperties( defaultConfigFileName );
  }

   /**
    * @return
    */
   public static String getDefaultConfigFileName()
   {
      return defaultConfigFileName;
   }

}
